# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

from mako.template import Template

from Tools.SourceStorage import FileType

class Template_FileType(FileType):
    """ Represents a file which contains a template 
        that it could be convert into a simple text.
    """
    def __init__(self):
        FileType.__init__(self, 'Template', '.tmpl')
            
    def pretty_print(self, text):
        """ Return a pretty representation of the file 
        
            :param text: The original text to be pretty printed
        """
        return text
    
    def render (self, text, **kwargs):
        """ Return a string after the renderization of the content
        
            :param text: The template 
            :param kwargs: The list of variables to be passed to the template
        """
        try:
            text = Template(text).render_unicode(**kwargs)
        except NameError:
            text = 'Error!! There is one or more variable in the template without its corresponding value'
        return text
        

class Makefile_FileType(FileType):
    """ Class to store makefile source code
    """
    def __init__(self):
        FileType.__init__(self, 'Makefile', '')

class MK_FileType(FileType):
    """ Class to store makefile source code
    """
    def __init__(self):
        FileType.__init__(self, 'makefile', '.mk')
