# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
Created on Apr 30, 2012

@author: Ruyman Reyes
'''
from MiddleEnd.Functions.Mutators.Outliner import block_is_SESE
from Backends.Common.Visitors.GenericVisitors import FuncDefOfNameFilter,\
    FuncCallFilter
from Tools.Tree import getFileNode, NodeNotFound

class Kernelize(object):
    def init(self,):
        pass

    @classmethod
    def loadTemplateFromFile(self, file_name):
        """ Loads a template from a file
        """
        with open(file_name, 'r') as myfile:
            data = myfile.read()
        return data

    def is_kernelizable(self, function, loop_parameters,):
        ### Disabled due to problem with kernelize assert block_is_SESE(function.body)
        # TODO
        #  1 - Check no pointers inside stmts
        return True
    
    def inline_called_functions(self, non_visited_funcalls, objetive, 
                    native_list, allow_native):
        """ Inline called functions 
        """
        undeclared_list = []
        visited = []
        called_list = []
        while len(non_visited_funcalls) > 0:
            try:
                elem = non_visited_funcalls.pop()
                if not allow_native or (allow_native and not elem.name in native_list):
                    #~ print " Extracting function named: " + str(elem.name)
                    funcDef = FuncDefOfNameFilter  (name = elem.name).apply(getFileNode(objetive))
                    called_list.append(funcDef) 
                    visited.append(elem.name)
                    # Check for nested calls
                    for elem in FuncCallFilter().iterate(funcDef.body):
                        if not elem in visited:
                            non_visited_funcalls.append(elem)
                elif allow_native and elem.name in native_list:
                    print " Native call to :" + str(elem.name)
                else:
                    print " Function " + elem.name + " is neither native or declared "
                    raise NotImplemented
            except NodeNotFound:
                # Declaration could not be found
                undeclared_list.append(elem)
                continue
        if len(undeclared_list) > 0:
            print " Following functions were called but not declared: "
            print [str(elem.name) for elem in undeclared_list]
	    print " Will assume they are native "
	# Reverse list to match declaration order
        called_list.reverse()
        return called_list
        
