# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------


import unittest

from Tools.Dump import Dump

#from yacf.Backends.Cuda.Mutators.CM_OmpParallelFor import CM_OmpParallelFor
#from yacf.Backends.Cuda.Mutators.CM_OmpParallel import FM_LlcRegion

# from yacf.Frontend.Parse import parse_source
import config
from Frontend.FrontendFactory import FrontendFactory

from Backends.C.Writers.OmpWriter import OmpWriter

from Frontend.InternalRepr import AstToIR


BACKEND_NAME = 'Common'
TEST_PATH = 'Backends/' + BACKEND_NAME + '/tests/'
CODE_PATH =  TEST_PATH + 'codes/'
TREE_PATH =  TEST_PATH + 'trees/'

class TestCase(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestCase, self).__init__(*args, **kwargs)

    def check_output(self, new_tree, good_tree):
        import difflib
        import pprint
        # Build differ object and calculate diferences
        d = difflib.Differ()
        diff = list(d.compare(str(good_tree).split('\n'), str(new_tree).split('\n')))
        # If differences exists, show them
        if len([elem for elem in diff if elem[0] != ' ']) != 0:
            pprint.pprint(diff)
        self.assertEqual(str(good_tree).split('\n'), str(new_tree).split('\n'))


config.CURRENT_LANGUAGES = ['C99',]

class TestCommonFunctions(TestCase):

     def setUp(self):
          pass

     def test_Tools(self):
          """ Test common tools """
          template_code = '''
              int main() {
                  int i = 0;
                  i = 7;
              }
          '''
          [ast, prepro_class, lexer_class, parser_class] = FrontendFactory().parse_with_language_list(template_code, 'Tool test 1',
                 config.CURRENT_LANGUAGES)

          # Transform the C ast into the internal representation
          # new_ast = AstToIR(Writer = OmpWriter).annotate(ast)
          migrator = AstToIR(Writer = OmpWriter, ast = ast)
          migrator.annotate()
          new_ast = migrator.ast

          template_code = """
              int main() {
                  int f = 0;
                  f = 7;
              }
          """

          [ast, prepro_class, lexer_class, parser_class] = FrontendFactory().parse_with_language_list(template_code, 'Tool test 2',
                 config.CURRENT_LANGUAGES)
#          tmp = AstToIR(Writer = OmpWriter).annotate(parse_source(template_code, "Tool test 2")).ext[-1].body
          migrator = AstToIR(Writer = OmpWriter, ast = ast)
          migrator.annotate()
          new_ast_2 = migrator.ast
          
          tmp = new_ast_2.ext[-1].body

          #declarations = tmp.dcls
          #statements = tmp.stmts

          from Tools.Tree import InsertTool, ReplaceTool, RemoveTool #, CloneTool
          InsertTool(subtree = tmp.block_items[0], position = "begin").apply(new_ast.ext[-1].body, 'block_items')

          #print declarations
          #print new_ast
          #from yacf.Tools.Debug import DotDebugTool
          #DotDebugTool(highlight = [declarations,]).apply(new_ast.ext[-1].body)

          self.assertEqual(new_ast.ext[-1].body.block_items[0].parent, new_ast.ext[-1].body)
          # Clone the last statement
          #CloneTool(original = new_ast.ext[-1].body.stmts[-1], position = 'begin').apply(new_ast.ext[-1].body, 'stmts')
          # Check replace tool
          ReplaceTool(new_node = tmp.block_items[-1], old_node = new_ast.ext[-1].body.block_items[-1]).\
                        apply(new_ast.ext[-1].body, 'block_items')
          # Check if parent link is preserved
          self.assertEqual(new_ast.ext[-1].body.block_items[-1].parent, new_ast.ext[-1].body)
          # Check remove tool
          RemoveTool(target_node = new_ast.ext[-1].body.block_items[-1]).apply(new_ast.ext[-1].body, 'block_items')
          # Check parent links
          self.assertEqual(new_ast.ext[-1].body.block_items[-1].parent, new_ast.ext[-1].body)

          # Tree must be the same as original
          #good_tree = Dump.load(TREE_PATH + '/insert_tool_tree')
          #self.check_output(new_ast, good_tree) 


if __name__ == '__main__':
     unittest.main()
