#include "llcomp_cuda.h"

/* Define correlations between host and device vars */
typedef struct {
	void * atHost;
	void * atCuda;
	int references;
	size_t size;
	Correspondant * prev;
} Correspondant;

typedef * Correspondant CorrPtr;

/* Global list with correlations */
CorrPtr cuda_relation_list = NULL;

/* Create a new device var which is a copy of the given host var */
void * cuda_cloneHostVar(void * var, size_t size) {
     void * newVar_atCuda = NULL;
	  cudaMalloc((void **) newVar_atCuda, size);
     cudaMemcpy(newVar_atCuda, &var, size, cudaMemcpyHostToDevice);
	  /* Create a correspondence relation */
	  CorrPtr relation = (CorrespondantPtr) malloc(sizeof(Correspondant));
	  relation->atCuda = newVar_atCuda;
	  relation->atHost = newVar;
	  relation->size = size;
	  relation->prev = cuda_relation_list;
	  return newVar_atCuda;
}

/* Find the Correspondence structure related to the given var */
CorrPtr findCorrPtr(void * var) {
	CorrPtr act = cuda_relation_list;
	while (act) {
		if (act->atHost == var) return act;
		act->prev;	
	}
	return NULL;
}

/* Update the host var with the values of the Device var */
int cuda_reconcileHostVar(void * var) {

	CorrPtr relation = findCorrPtr(var);
	if (!relation)
		return NULL;

	cudaMemcpy(relation->atHost, relation->atCuda, relation->size,
						cudaMemcpyDeviceToHost);
	return 0;

}

/* Remove the given host var from the device */
int cuda_removeVar(void * var) {
	CorrPtr relation = findCorrPtr(var);
	if (!relation)
		return NULL;

	cudaMemcpy(relation->atHost, relation->atCuda, relation->size, cudaMemcpyDeviceToHost);
   cudaMemcpy(newVar_atCuda, &var, size, cudaMemcpyHostToDevice);
   cudaFree(relation->atCuda);

	return 0;
}


