# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
Created on Feb 2, 2011

@author: rreyes

TODO: Move all functions to CudaPlatform class
'''
from Frontend.Shortcuts import getCurrentLanguageAst

c_ast = getCurrentLanguageAst()

from Frontend.FrontendFactory import partial_parse;
from Backends.Common.TemplateEngine.Functions import is_complex_type
from Backends.Common.Mutators.AstSupport import IDNameMutator, PointerMutator
from Backends.Common.Mutators.AbstractMutator import AbstractMutator
from Backends.Common.Visitors.GenericVisitors import GenericFilterVisitor
from Frontend.InternalRepr import AstToIR
from Frontend.IRTypes import DeclList
from Tools.Tree import getContainerAttribute, getFileNode, NotValidParent, ReplaceTool
from MiddleEnd.Optimizer.Mutators.Optimizer import VectDecl2Ptr, VectRef2Ptr
from Frontend.SymbolTable import getSymbolTableFromNode, SymbolTable, SymbolTableBuilder, Symbol

from MiddleEnd.Stats.Visitors.FlopCountVisitor import FlopCountVisitor

import re

class CudaPlatform(object):
    '''
    classdocs
    '''


    def __init__(self,):
        '''
        Constructor
        '''
        pass
    
   
# TODO: Clean    
def substract_decl_list(decl_list, decls_to_remove):
    new = []
    for elem in decl_list:
        for tmp in decls_to_remove:
            if tmp.name == elem.name:
                break
        else:
            new.append(elem)
    return new
    
def intersection_decl_list(decl_list, decls_to_intersect):
    new = []
    for elem in decl_list:
        for tmp in decls_to_intersect:
            if tmp.name == elem.name:
                new.append(elem)
    return new

def _implement_reduction(op, symbolList, stmts, kernel_parameters):
    for symbol in symbolList:
          elem = symbol.node
          new_name = elem.name + '_private'
          IDNameMutator(old = c_ast.ID(name = elem.name, parent = elem.parent), 
                              new = c_ast.ID(name = new_name, parent = elem.parent)).apply_all(stmts)
          # Kernel parameter must be changed to a global reduction array
          for symbol_kparam in kernel_parameters:
             kparam = symbol_kparam.node
             if kparam.name == elem.name:
                # Element must be a scalar
                if is_complex_type(kparam):
                   import pdb
                   pdb.set_trace()
                   print " Only scalars are supported as reduction vars "
                   raise NotImplemented
                new_param = kparam.__deepcopy__()
                # It is a scalar, pointerize declaration
                pointer_node = c_ast.PtrDecl(type = new_param.type, quals = [], parent = new_param)
                setattr(new_param, 'type', pointer_node)
                new_symbol = Symbol(name = symbol_kparam.node.name, node = new_param, 
                                    type  = new_param.type, btype = symbol_kparam.btype, 
                                    sizeExpression = symbol_kparam.sizeExpression )
                pos = kernel_parameters.index(symbol_kparam)
                kernel_parameters[pos] = new_symbol
                break
          else:
             print " Something weird happened, as a reduction variable was not a kernel parameter"
             raise NotImplemented



def kernel_optimizations(var, kernel_decls, nFAST, kernel_stmts, global_vars):
    if (len(global_vars) > 0): # Count the number of floating point operations over kernel parameters
        fpv = FlopCountVisitor(contableVars=global_vars)
        fpv.visit(kernel_stmts) #        _system_parameters['flops'] = c_ast.Constant(type=['integer'], value = str(fpv.getFlopCount()))
#        _system_parameters['memAccess'] = c_ast.Constant(type=['integer'], value = str(fpv.getMemAccess()))
        flops_subtree = fpv.getFlopCountSubtree()
        memAccess_subtree = fpv.getMemAccessSubtree()
        divergence_subtree = fpv.getIfCount()
    else:
        flops_subtree = c_ast.Constant(type=['integer'], value="10")
        memAccess_subtree = c_ast.Constant(type=['integer'], value="1")
        divergence_subtree = 0 # Not implemented
    return flops_subtree, memAccess_subtree, divergence_subtree

def build_cuda_system_parameters(_loop_parameters, stmts, _loop2_parameters = None, _loop3_parameters = None):
    stmts = AstToIR.deepcopySubtree(stmts)
    _system_parameters = {}

    kernel_parameters = []
    loop_vars = None
    if _loop3_parameters and _loop2_parameters:
        loop_vars = (_loop_parameters['loop_variable'].name, 
                     _loop2_parameters['loop_variable'].name, 
                     _loop3_parameters['loop_variable'].name)
    elif _loop2_parameters:
        loop_vars = (_loop_parameters['loop_variable'].name, 
                     _loop2_parameters['loop_variable'].name,)
    else:
        loop_vars = (_loop_parameters['loop_variable'].name,)
    for var in _loop_parameters['used_vars']:
        if not (var.name in loop_vars):
            kernel_parameters.append(var)
#    kernel_parameters = [ var for var in _loop_parameters['used_vars'] if var.name != _loop_parameters['loop_variable'].name]
    # Remove the inner loop variable if any
#    if _loop2_parameters:
#        kernel_parameters = [ var for var in kernel_parameters if var.name != _loop2_parameters['loop_variable'].name]
    # register_vars = local_vars - loop_vars
#    import pdb
#    pdb.set_trace()

    register_vars = []
    for var in _loop_parameters['local_vars']:
        if not (var.name in loop_vars):
            print " Adding " + str(var.name) + ' as a register'
            if var.extra.has_key('startAddr'):
                register_vars.append(var)

    _system_parameters['register_vars'] = register_vars
    # Private vars are not kernel parameters
    kernel_parameters = substract_decl_list(kernel_parameters, register_vars)
    kernel_decls = [ AstToIR.deepcopySubtree(elem.node) for elem in kernel_parameters ]
    original_kernel_parameters = substract_decl_list(kernel_parameters, _loop_parameters['local_vars'])
    # kernel_parameters = [ var.__deepcopy__() for var in kernel_parameters]
    # All declarations have to be part of a File. This is a temporary
    #  set of declarations, so we will create a temporary virtual file
    #  that will be used to handle this variables indepently from the rest
    #  of the code, avoiding future manipulations to alter the original vars
    # kernel_parameters = [ elem.node for elem in kernel_parameters]
    nFAST = c_ast.FileAST(ext = kernel_decls)
    # Replace access to reduction variables to reduction arrays 
    #for elem in _loop_parameters['reduction_vars']:
    #        ReplaceReductionArrayMutator(original_var = c_ast.ID(name = elem.name, parent = elem.parent), 
    #                                     array_index = c_ast.ID(name = elem.name, parent = elem.parent)).apply_all(stmts)
    # Reductions are implemented using a local private variable
    for op,symbolList in _loop_parameters['reduction_vars']:
            _implement_reduction(op, symbolList, stmts, kernel_parameters)
    # Code is to be held here
    if _loop2_parameters and _loop3_parameters:
        # Remove the three nested loops
        kernel_stmts = stmts.stmt.stmt
    elif _loop2_parameters:
        # Remove the inner loop
        #if hasattr(stmts, "stmt") and isinstance(stmts.stmt, c_ast.For):
        kernel_stmts = stmts.stmt
        #else:
        #    kernel_stmts = stmts.block_items[0].stmt.stmt.stmt
    else:
        kernel_stmts = stmts
        
    # This kernel will be inside the boundary condition check (i.e if ...)
    _system_parameters['code'] = kernel_stmts
    # No source mutators are required since we modified the loop here
    _system_parameters['source_mutators'] = []
    # Parameters to the kernel call
    _system_parameters['kernel_parameters'] = kernel_parameters
    _system_parameters['original_parameters'] = original_kernel_parameters
    # Identify global (i.e "far away from core memory", 
    # which is Global in CUDA and __global in OpenCL)
    global_vars = kernel_parameters # Reduction vars are not counted
    flops_subtree, memAccess_subtree, divergence_subtree = kernel_optimizations(var, kernel_decls, nFAST, kernel_stmts, global_vars)

    _system_parameters['flops'] = flops_subtree
    _system_parameters['memAccess'] = memAccess_subtree
    _system_parameters['divergenceFactor'] = divergence_subtree

    print "***** Kernel Statistics *******"
    print " Flop count: " + str(_system_parameters['flops'] )
    print " Mem Access: " + str( _system_parameters['memAccess'])
    print " Div. Factor: " + str( _system_parameters['divergenceFactor'])

    return _system_parameters



