/* The default behaviour of the CUDA Backend is to enable double precision floating point operations */
 
#define __YACF_CUDA_MAP_DIM(VAR, DIM) {\
   int val;\
   switch(DIM)\
   {\
      default:\
      case (0) : val = VAR.x;\
                 break;\
      case (1) : val = VAR.y;\
                 break;\
      case (2) : val = VAR.z;\
                 break;\
   }\
   return val;\
}

/* Mapping of thread indexing functions to CUDA */
__device__ int __yacf_get_num_groups(int dim)
{
   __YACF_CUDA_MAP_DIM(gridDim, dim)
}

__device__ int __yacf_get_local_size(int dim)
{
   __YACF_CUDA_MAP_DIM(blockDim, dim)
}

__device__ int __yacf_get_group_id(int dim)
{
   __YACF_CUDA_MAP_DIM(blockIdx, dim)
}

__device__ int __yacf_get_local_id(int dim)
{
   __YACF_CUDA_MAP_DIM(threadIdx, dim)
}

__device__ int __yacf_get_global_id(int dim)
{
   int val;
   switch(dim)
   {
      default  :
      case (0) : val = blockIdx.x * blockDim.x + threadIdx.x;
                 break;
      case (1) : val = blockIdx.y * blockDim.y + threadIdx.y;
                 break;
      case (2) : val = blockIdx.z * blockDim.z + threadIdx.z;
                 break;
   }
   return val;
}

__device__ int __yacf_get_global_size(int dim)
{
int val;
   switch(dim)
   {
      default  :
      case (0) : val = gridDim.x*blockDim.x;
                 break;
      case (1) : val = gridDim.y*blockDim.y;
                 break;
      case (2) : val = blockDim.z*blockDim.z;
                 break;
   }
   return val;
}
   
/* Reprint typedefs */
%for typedef in typedeflist:
 ${typedef};
%endfor

/* Functions called within the kernel */
%for func in funccalllist:
   __device__ ${func};
%endfor
 
/* Kernel functiokns */          
extern "C" {

   ${kernelProto}

}

${kernelFunction};
 

/*** End of CUDA Template header ***/
