# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

from Frontend.C99 import c99_ast as c_ast


from Backends.Cuda.Visitors.CM_Visitors import *

from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, RemoveTool, getContainerAttribute

from Tools import SourceStorage

from Frontend.SymbolTable import updateSymbolTableFromNode,\
    getSymbolTableFromNode

from Backends.Frangollo.Templates.FrangolloVarNode import createContextVarList,\
    createKernelParameterList

from Backends.Frangollo.Mutators.FM_PrivDecl import FM_PrivDecl

from Backends.Common.Mutators.AbstractMutator import AbstractMutator
from Backends.C99.Visitors.LlcFilters import AccLoopFilter, AccKernelsReverse,\
    AccScopeReverse, FuncDefReverse, AccHostDataFilter

# CUDA Code generation
from Backends.Cuda.Files.FileTypes import Cuda_FileType


# OpenCL Code generation
from Backends.OpenCl.Files.FileTypes import OpenCl_FileType

from Tools.Tree import getFileNode

from MiddleEnd.Loop.Analysis import AccLoopParametrize


from MiddleEnd.Functions.Mutators.Outliner import block_is_SESE, OutlineMutator,\
    ParameterStyle, c_outline_loop_stmts

from MiddleEnd.Loop.Mutators.LoopInvariant import LoopInvariantMutator
from Backends.C.Files.FileTypes import C_FileType
from Backends.Frangollo.Mutators.Common import AbstractFrangolloMutator


class FM_AccHostData(AbstractFrangolloMutator):
    instantiationNumber = 0

    def __init__(self, *args, **kwargs):
#        self._kernel_name = kwargs['kernel_name']
        self._clauses = kwargs['clauses']
        self._st = kwargs['source_storage']
        self._kernelList = []
        self._parentRegion = None
        if kwargs.has_key('xmlm'):
#            print " Has key, value: " + str(kwargs['xmlm'])
            self._xmlm = kwargs['xmlm']
        else:
            self._xmlm = None
        self._backend_name = "Frangollo"
        super(FM_AccHostData, self).__init__()

    def lookParentWithAttr(self, parent_node, prev, attr):
        while not hasattr(parent_node, attr):
            if isinstance(parent_node, c_ast.FileAST):
                raise MutationError
            prev = parent_node
            parent_node = parent_node.parent
        
        if prev == None:
            raise MutationError
        return prev, parent_node


    def getExtractedKernelList(self,):
        return self._kernelList
#------------------------------------------------------------------------------ 
    #--------------------------------- def  _get_scope_clause_dict(self, node,):
        #-------------------------- """ Builds the OpenMP + llc scope dictionary
        #---------------------- :param node: Node from which extract information
        #---------------------------------------- :return: Dictionary of clauses
        #------------------------------------------------------------------- """
        #-------------------------------- # from Tools.Debug import DotDebugTool
        # # DotDebugTool(highlight = [node, node.parent, node.parent.parent]).apply(getFileNode(node))
        #--------------- llcfilt = AccScopeReverse(rootNode = getFileNode(node))
        #--------------------------------------------------- llcfilt.apply(node)
        #------------------------------------------------ return llcfilt.clauses
#------------------------------------------------------------------------------ 

    def filter(self, ast):
        """ Filter definition
            Returns the first node matching with the filter
        """
        # Build a visitor , matching the OmpFor node of the AST
        f = AccHostDataFilter()
        node = f.apply(ast)
        self._parentRegion = f.get_parallel()
        return node

    def apply_all(self, parent_region_node, ast):
        """ Apply mutation to all matches """
        start_node = None
        self.ast = ast
        f = AccHostDataFilter()
        num = 0;
        try:
            for elem in f.iterate(ast):
                # Current scope variables
                self._parentRegion = f._parentRegion
                # If current node is not child of first parallel node, stop
                if self._parentRegion != parent_region_node:
                    # print " Different Parent region "
                    continue
                start_node = self.mutatorFunction(ast, elem)
                # Restore previous state
                # self._clauses = old_clauses
                num += 1;
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            return self._parentRegion
        return start_node


    def mutatorFunction(self, ast, accHostData_node):
        """ Replace Host ptr by device ptr
        """
        # General region parameters
        clauses = self._get_scope_clause_dict(accHostData_node.stmt);
        region = {}
        region['name'] = clauses['NAME'] or "random"
        region['vars'] = self.extract_directionality(ast, clauses) 

        # Check what to do with this particular llcFor
        # Default behavior 
        transform = True
        buildCuda = True
        buildCl = True
        collapse = True
        # Update the kernel name
        FM_AccHostData.instantiationNumber += 1


        # print "Cluster " + xmlm.cluster.get("name") + "."
        # Assuming the first cluster is the one you want
        if self._xmlm:
                print " XML Description file provided "
                xml_regions = self._xmlm.find_cluster("default")
                if self._xmlm.find_cluster("default"):
                    # If we could read the xml, lets get information from it
                    if self._xmlm.find_region(region['name']):
                        # If we found the region, then we have to look for this element
                        kname = region['name'] + "_" + str(FM_AccHostData.instantiationNumber)
                        if self._xmlm.find_element(kname):
                            # This is the element that we are processing now
                            pass
        else:
            print "No XML Description file provided "

 
        accHostData_code = """
        #include "openacc.h"
        typedef long CUdeviceptr; 
        // Context substitution
        int main() {
         {
            // Replace variables inside scope
            %for var in region['usedevice']:
                void * tmp_${var.name} =  FRG__getDevicePtr(FRG__getActiveContext(), ${var.name});
                /** To avoid compiler doing strange things **/
                CUdeviceptr ${var.name} = (*((CUdeviceptr *) tmp_${var.name}));
            %endfor
                 {
                     placeholder();
                 }
            
         }
        }
        """
        clauses = self._get_scope_clause_dict(accHostData_node.stmt);
        
        region = {}
        if clauses.has_key('NAME'):
            region['name'] = clauses['NAME'] or "random"
        else:
            region['name'] = "random"
        names, ast = self.extract_directionality(ast, clauses)
        region['vars'] = names
        st = getSymbolTableFromNode(accHostData_node)
        region['usedevice'] = [st.lookUp(id) for id in accHostData_node.varlist]

        accHostData_subtree = self.parse_snippet(accHostData_code, 
                                             {'region' : region}, 
                                             name = 'General Context', 
                                              show = False).ext[-1].body
 
        updateSymbolTableFromNode(accHostData_subtree)
        # Replace placeholder
#        import pdb
#        pdb.set_trace()
        ReplaceTool(new_node = accHostData_node.stmt, old_node = accHostData_subtree.block_items[-1].block_items[-1]).apply(accHostData_subtree.block_items[-1], 'block_items')
        # Replace Pragma
        ReplaceTool(new_node = accHostData_subtree.block_items[-1], 
                    old_node = accHostData_node.parent).apply(accHostData_node.parent.parent, 'block_items')
        
        updateSymbolTableFromNode(accHostData_node.parent.parent)
        # Update Tree (sequence, depth, etc)
       
        return accHostData_subtree
        
       
