# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

import config


from Frontend.Shortcuts import getCurrentLanguageAst
c_ast = getCurrentLanguageAst()

from Backends.C99.Visitors.LlcFilters import AccKernelsFilter, AccLoopFilter,\
    AccScopeReverse, AccDataFilter, AccParallelFilter

from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, getFileNode
from Tools.Tree import ReplaceTool
from Tools.Tree import getContainerAttribute
from Frontend.C99 import c99_ast
from Frontend.Acc import acc_ast
from Tools.Declarations import decl_of_id

# CUDA Code generation
from Backends.Cuda.Kernelize import CudaKernelize
from Backends.Cuda.Files.FileTypes import Cuda_FileType
from Backends.Cuda.Platform import build_cuda_system_parameters


# OpenCL Code generation
from Backends.OpenCl.Kernelize import  CLKernelize
from Backends.OpenCl.Files.FileTypes import OpenCl_FileType


from Frontend.InternalRepr import AstToIR
from Tools import SourceStorage

from Backends.C.Files.FileTypes import C_FileType
from Backends.Common.Mutators.AbstractMutator import AbstractMutator
from Backends.Frangollo.Templates.FrangolloVarNode import FrangolloVarNode
from Backends.Frangollo.Mutators.FM_LlcFor import FM_AccLoop
from Backends.Frangollo.Mutators.FM_PrivDecl import FM_PrivDecl
from Backends.Frangollo.Mutators.Common import AccParallelLoopMutator

from MiddleEnd.DataAnalysis.LlcScope import DNode,DGraph,BlockDepGraphBuilder
from Backends.Frangollo.Templates.FrangolloVarNode import createContextVarList,\
    createKernelParameterList


from Frontend.SymbolTable import getSymbolTableFromNode, updateSymbolTableFromNode
from Backends.Frangollo.Mutators.Common import AbstractFrangolloMutator

from Backends.Frangollo.Mutators.FM_AccHostData import FM_AccHostData

from Backends.Frangollo.Mutators.FM_AccUpdate import FM_AccUpdate

from Backends.Frangollo.Mutators.FM_AccRegions import FM_AccRegion, _merge_dict_per_key

from MiddleEnd.Functions.Mutators.Outliner import block_is_SESE, OutlineMutator,\
    ParameterStyle, c_outline_loop_stmts

from MiddleEnd.Loop.Common import PragmaFilter, IDMutator
    
from Backends.C99.Writers import C99Writer
from Backends.C99.Writers.AccWriter import AccWriter

from Backends.Cuda.Platform import _implement_reduction

class FM_AccParallel(FM_AccRegion):
    instantiationNumber = 1
    def __init__(self, *args, **kwargs):
        super(FM_AccParallel, self).__init__(*args, **kwargs)
        self._isDataDirective = False
        self._isParallelDirective = True
        self._isKernelsDirective = False
    
    def filter(self, ast):
        """ Filter definition
            Returns the first node matching with the filter
        
        """
        
        f = AccParallelFilter()
        node = f.apply(ast)
        self._parentRegion = node
        self._isDataDirective = False
        self._isParallelDirective = True
        return node

    def apply_all(self, ast):
        """ Apply mutation to all matches """
        self.ast = ast
        f = AccParallelFilter()
        final_kernel_list = []
        try:
            for elem in f.iterate(ast):
                self._kernelList = []
                self.mutatorFunction(ast, elem)
                final_kernel_list += self._kernelList
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            self._kernel_list = final_kernelList
            return self._parentRegion

        self._kernelList = final_kernel_list
        return ast

    def build_basic_kernel(self, clauses, region, node, \
                        buildCuda, buildCl, om, dGraph = None):                 
        FM_AccParallel.instantiationNumber += 1
	#if hasattr(node, 'stmt'):
	        #stmts = node.stmt
	#elif hasattr(node, 'block_items'):
		##if hasattr(node.block_items[0], 'name'):
			##if node.block_items[0].name == 'loop':
				##stmts = node.block_items[0].stmt
			##elif hasattr(node.block_items[0], 'stmt'):
                               ##if hasattr(node.block_items[0].stmt, 'name') and \
                                        ##node.block_items[0].stmt.name == 'loop':
                                    ##stmts = node.block_items[0].stmt.stmt
                               ##else:       
                                    ##stmts = node.block_items[0].stmt
			##else:
				##raise MutatorException("Subtree within AccParallel not recognised")
		##else:
			#stmts = node
	#else:
		#raise MutatorException("Subtree within AccParallel not recognised")
        stmts = node # .parent.parent

        # Extract loop contents as functions
        kernel_name = region['name'] + "_P_" + str(FM_AccParallel.instantiationNumber)
        om = OutlineMutator(name = kernel_name + "_support", replace = False, change_code_block = False, writer = AccWriter)        
        om.apply(stmts)
        
	# Replace the kernel call to a Frangollo kernel launch statement
        kernelCall_code = ""
        if config.FRANGOLLO_DEBUG:
            kernelCall_code += "#define DEBUG"

        kernelCall_code += """
        // Get Frangollo interface definitions
        #include "openacc.h"
        
        %for var in typedef:
            ${var};
        %endfor 
        
        int main() {
               if (FRG__offloadKernel(FRG__getActiveContext(), "${kernel['name']}") && (${cond_expr})) {
                  tparams *__FRG__parameter_list_head, * __FRG__parameter_list_prev;
			         var_layout_t var_layout, acc_layout;
		  __FRG__parameter_list_prev = NULL;
		  __FRG__parameter_list_head = NULL;

		 %for var in kernel['parameters']:

		  {
			void * ___ptr_to_var;
                  %if var.frangolloType == "VAR_SCALAR":
			___ptr_to_var = &${var};
                  %else:
			___ptr_to_var = ${var};
                  %endif 

		  %if var.force_register:
			{
			var_layout.x = ${var.size};
			var_layout.y = 0;
			var_layout.type = ${var.frangolloType};
			FRG__registerVar(___ptr_to_var, regionId, 0 | 
					    (VAR_INOUT * 0x10) | 
					    (1 * 0x100),
					    var_layout, &acc_layout);
                	}
		   %endif

		  %if var.reductionOp:
                        NEW_KERNEL_PARAM("${var.name}", ___ptr_to_var, 
					  ${var.frangolloType}, ${var.reductionOp}, 
					  ${var.reductionType}, "0", ${var.canChange}, ${var.size});
		  %else:
		        NEW_KERNEL_PARAM("${var.name}", ___ptr_to_var, 
					  ${var.frangolloType}, "", 
					  "", "0", ${var.canChange}, ${var.size});

		  %endif
                  
                  }     

                  %endfor
                     
                      FRG__submit_kernel_manual(FRG__getActiveContext(), "${kernel['name']}", 
                                                __FRG__parameter_list_head, ${kernel['blocks']},
					       ${kernel['threads']}, 
                                                  %if kernel['dimensions'] > 1:
                                                  ${kernel['blocks']}, ${kernel['threads']} 
                                                  %else:
                                                  -1, -1 
                                                  %endif
					       );

                    tparams * __FRG_tmp_parameter_list__;
                    while (__FRG__parameter_list_head) {
                        __FRG_tmp_parameter_list__ = __FRG__parameter_list_head;
                        __FRG__parameter_list_head = __FRG__parameter_list_head->next;
                        free(__FRG_tmp_parameter_list__);
                    }
                    %if kernel['async']:
                        FRG__barrier();
                    %endif

               } 
		else {
                   /* Call to the original code outlined to a external function */
                   /* ${callToOriginal}; */
                   ${original_code};
               }
        }
        """
	# Assuming LLC --> FOR --> STMT
        system_parameters = self.get_kernel_parameters_from_stmt(stmts, clauses, om)
        # Create kernel dictionary, wich is just a shortcut for the template
        kernel = {}
        kernel['name'] = region['name'] + "_P_" + str(FM_AccParallel.instantiationNumber)
        kernel['flops'] = 1
        kernel['memAccess'] = 1
        kernel['dimensions'] = system_parameters['dimensions'] if system_parameters['dimensions'] in range(1,3) else 1
        print " Parallel kernel " + str(kernel['dimensions']) + "D"

        self._kernelList.append(kernel)
         
        # Notice how kernel parameters are not the same as region variables. 
        kernel['parameters'] = createKernelParameterList(None, 
                                None, 
                                system_parameters['original_parameters'], 
                                system_parameters['reduction_vars'],
                                clauses['FIRSTPRIVATE'], node, dGraph)
        if clauses.has_key('ASYNC'):
            kernel['async'] = clauses['ASYNC']
        else:
            kernel['async'] = None

	if clauses.has_key('IF'):
	    if clauses['IF']:
	    	cond_expr = str(clauses['IF']);
	    else:
		cond_expr = "1"
	else:
	    cond_expr = "1"

        system_parameters['kernel_name'] = kernel['name']
        
        if (clauses.has_key('num_gangs')):
            kernel['blocks'] = clauses['num_gangs']
        else:
            kernel['blocks'] = 16;
            
        if (clauses.has_key('num_workers')):
            kernel['threads'] = clauses['num_workers']
        else:
            kernel['threads'] = 16;

        print " Gangs:", kernel['blocks']
        print " Workers:", kernel['threads']
 
        # CUDA kernel
        # Extract kernel from the loop
        kernel_source = CudaKernelize().cuda_outline_stmts(
                                                    system_parameters, 
                                                    stmts,
                                                    kernel['name'],
                                                    om = om) 
        
        # Add source file to source storage
        self._source_storage.addFile(kernel['name'], Cuda_FileType())
        # No particular include file is used here
        self._source_storage.append(kernel['name'], Cuda_FileType(), kernel_source)


        # OpenCl kernel
        # Extract kernel from the loop
        # TODO : Probably, system_parameters should be different for OpenCL than for CUDA
        #         as the transformation/parametrization phase is probably different
        kernel_source = CLKernelize().opencl_outline_stmts(
                                            system_parameters, 
                                            stmts, 
                                            kernel['name'],
                                            om = om) 
        
        # Add source file to source storage
        self._source_storage.addFile(kernel['name'], OpenCl_FileType())
        # No particular include file is used here
        self._source_storage.append(kernel['name'], OpenCl_FileType(), kernel_source)
        
        # TODO: c_outline does not properly tracks functions and types, this should be llcFor_node.stmt.stmt
        #        but for some reason it does not work
        kernel_source = c_outline_loop_stmts(stmts,om)
        
        # Add source file to source storage
        self._source_storage.addFile(kernel['name'], C_FileType())
        # No particular include file is used here
        if config.GENERATE_HOST_CODE:
            self._source_storage.append(kernel['name'], C_FileType(), kernel_source)
        else:
            self._source_storage.append(kernel['name'], C_FileType(), ";")
            
        #import pdb
        #pdb.set_trace()
        #####################################################################################            
        # Finally remove pragmas for original_code in template kernelCall_code
        try:
            #import pdb
            #pdb.set_trace()
            pragma_list = PragmaFilter()
            node_loop = stmts
            while True:
                # TODO: Check pragma_loop is a pragma acc loop
                pragma_loop = pragma_list.apply(node_loop)
                node_loop = ReplaceTool(new_node = pragma_loop.stmt.stmt, old_node = pragma_loop               
                                          ).apply(pragma_loop.parent, getContainerAttribute(pragma_loop))
        except NodeNotFound:
            pass
        
        # TODO: Fix double pragma bug to avoid this:
        if hasattr(stmts.parent, 'stmt'):
            stmts = stmts.parent.stmt
        
            
        kernelCall_subtree = self.parse_snippet(kernelCall_code, {'region' : region, 'kernel' : kernel, 
                                           'callToOriginal' : None, # om.getOutlinedFunctionCall(), 
                                           'original_code' : str(stmts),
                                           'cond_expr' : cond_expr ,
                                           'typedef' : om._typedefList}, 
                                                name = 'For kernel', show = False, writer = AccWriter).ext[-1].body    
            
        # --------------------------------------------------------------------------
        # If prev node can be parent_node in self.lookParentWithAttr then
        # --------------------------------------------------------------------------
        #parent_node = node.stmt
        #prev = None
        #attr = 'stmt'
        
        #prev, parent_node = self.lookParentWithAttr(parent_node, prev, attr)
        
        ## Replace call
        #ReplaceTool(new_node = kernelCall_subtree, old_node = prev).apply(parent_node, attr)
        
        # Else
        # --------------------------------------------------------------------------
	if hasattr(node, 'stmt'):
        	parent_node = node.parent # parent_node = node
        	prev = node # prev = node.stmt
        	attr = 'stmt'
	elif hasattr(node, 'block_items'):
        	# Look for the parent node which has the stmt attr
		parent_node = node
		prev = None
		attr = 'stmt'
		prev, parent_node = self.lookParentWithAttr(parent_node, prev, attr)
	else:
		raise MutatorException(" Cannot replace the parent node in the original code ")
        # Replace call
        ReplaceTool(new_node = kernelCall_subtree, old_node = prev).apply(parent_node, attr)


        # --------------------------------------------------------------------------
 
        # Update Tree (sequence, depth, etc)
        updateSymbolTableFromNode(node)


        return node

    def get_kernel_parameters_from_stmt(self ,stmts, clause_dict, om):
            stmts = AstToIR.deepcopySubtree(stmts)
            _system_parameters = {}

            kernel_parameters = []
            used_vars = self.extract_used_vars(stmts)
            local_vars = []
            if clause_dict.has_key('PRIVATE'):
               local_vars +=  clause_dict['PRIVATE'] 
           
            for var in used_vars:
                kernel_parameters.append(var)
            register_vars = []
            for var in local_vars:
                if var not in register_vars:
                    register_vars.append(var)

            _system_parameters['register_vars'] = register_vars
            original_kernel_decls = [ AstToIR.deepcopySubtree(elem.node) for elem in kernel_parameters ]
            
            # Private vars are not kernel parameters
            kernel_parameters = self.substract_decl_list(kernel_parameters, register_vars)            
            kernel_decls = [ AstToIR.deepcopySubtree(elem.node) for elem in kernel_parameters ]
            original_kernel_parameters = self.substract_decl_list(kernel_parameters, local_vars)
            # All declarations have to be part of a File. This is a temporary
            #  set of declarations, so we will create a temporary virtual file
            #  that will be used to handle this variables indepently from the rest
            #  of the code, avoiding future manipulations to alter the original vars
            # kernel_parameters = [ elem.node for elem in kernel_parameters]
            kernel_stmts = AstToIR(Writer = AccWriter, ast = stmts).deepcopySubtree(stmts,)
	    AstToIR.link_subtree_parents(kernel_stmts)
            # TODO: Replace this by a node construct AST. May be Outliner without parse_snippet
            new_stmts_template = """
            %for var in typedef:
                ${var};
            %endfor 
            void fake() { ${stmts}; }"""
            fake_AST = self.parse_snippet(new_stmts_template, {'stmts' : stmts, 'typedef': om._typedefList}, 
                              name = 'new_stmts_template', show = False, writer = AccWriter)
            kernel_stmts = fake_AST.children()[-1]
            nFAST = c_ast.FileAST(ext = original_kernel_decls + [kernel_stmts])
            kernel_stmts.parent = nFAST
            for decl in original_kernel_decls: decl.parent = nFAST 
	    updateSymbolTableFromNode(nFAST)

            # Apply acc parallel loop mutator
            parallelLoop = AccParallelLoopMutator(om)
            # BUG: two nested acc loop before one acc loop inside a parallel region
            #           give incorrect maxLevel
            # TODO: iterate parallelLoop and get max(maxLevels) in each apply
            loop_node = parallelLoop.apply_all(kernel_stmts)
            _system_parameters['dimensions'] = parallelLoop.maxLevel #if loop_node else 1
            _system_parameters['reduction_vars'] = parallelLoop.reduction_parameters
            # Private reduction vars are not reduced register vars
            reduced_reg_vars = register_vars
            for op, var_list in parallelLoop.reduction_parameters:
                reduced_reg_vars = self.substract_decl_list(reduced_reg_vars, var_list)
            _system_parameters['reduced_reg_vars'] = reduced_reg_vars
            # Save number of iterations for distributed loops
            #_system_parameters['__noi__'] = ''.join([str(x)+'*' for x in parallelLoop.__noi__[0:-1]]) + str(parallelLoop.__noi__[-1])
            # This kernel will be inside the boundary condition check (i.e if ...)
            #_system_parameters['code'] = kernel_stmts.children()[0].body #if not loop_node else loop_node
            code = kernel_stmts.body #if not loop_node else loop_node
            # Translate reduction_vars
            reduced_reduction_vars = []
            for op,symbolList in parallelLoop.reduction_parameters:
                # Private reduction vars are private vars, we do not apply true reduction
                reducedSymbolList = self.substract_decl_list(symbolList, local_vars)
                if reducedSymbolList:
                    reduced_reduction_vars += [(op, reducedSymbolList)]
                    # Traslate reduction_vars IDs & update kernel_parameters with reduction_vars types and 
                    _implement_reduction(op, reducedSymbolList, code, kernel_parameters)
            _system_parameters['code'] = code
            # No source mutators are required since we modified the loop here
            _system_parameters['source_mutators'] = []
            # Parameters to the kernel call
            _system_parameters['kernel_parameters'] = kernel_parameters
            _system_parameters['original_parameters'] = original_kernel_parameters
            # Identify global (i.e "far away from core memory", 
            # which is Global in CUDA and __global in OpenCL)
            global_vars = kernel_parameters 

            _system_parameters['flops'] = 1
            _system_parameters['memAccess'] = 1
            _system_parameters['divergenceFactor'] = 0

            return _system_parameters


