# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

from Frontend.C99 import c99_ast as c_ast
from Backends.Cuda.Visitors.CM_Visitors import *
from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, RemoveTool, getContainerAttribute
from Tools import SourceStorage
from Frontend.SymbolTable import updateSymbolTableFromNode,\
    getSymbolTableFromNode
from Backends.Frangollo.Templates.FrangolloVarNode import FrangolloVarNode

from Backends.Frangollo.Mutators.FM_PrivDecl import FM_PrivDecl
from Backends.Common.Mutators.AbstractMutator import AbstractMutator
from Backends.C99.Visitors.LlcFilters import AccLoopFilter, AccKernelsReverse,\
    AccScopeReverse, FuncDefReverse, AccHostDataFilter, AccWaitFilter
from Tools.Tree import getFileNode
from Backends.Frangollo.Mutators.Common import AbstractFrangolloMutator


class FM_AccWait(AbstractFrangolloMutator):
    instantiationNumber = 0

    def __init__(self, *args, **kwargs):
        if kwargs.has_key('clauses'):
            self._clauses = kwargs['clauses']
        self._st = kwargs['source_storage']
        self._kernelList = []
        self._parentRegion = None
        if kwargs.has_key('xmlm'):
            self._xmlm = kwargs['xmlm']
        else:
            self._xmlm = None
        self._backend_name = "Frangollo"
        super(FM_AccWait, self).__init__()

    def getExtractedKernelList(self,):
        return self._kernelList

    def filter(self, ast):
        """ Filter definition
            Returns the first node matching with the filter
        """
        # Build a visitor , matching the OmpFor node of the AST
        f = AccWaitFilter()
        node = f.apply(ast)
        self._parentRegion = f.get_parallel()
        return node

    def apply_all(self, ast):
        """ Apply mutation to all matches """
        start_node = None
        self.ast = ast
        f = AccWaitFilter()
        print " - * - "
        num = 0;
        try:
            for elem in f.iterate(ast):
                # Current scope variables
                # If current node is not child of first parallel node, stop
                start_node = self.mutatorFunction(ast, elem)
                # Restore previous state
                # self._clauses = old_clauses
                num += 1;
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            return self._parentRegion
        return start_node


    def mutatorFunction(self, ast, accWait_node):
        """ Insert Frangollo update calls
        """
        # General region parameters
        # clauses = self._get_scope_clause_dict(accWait_node.stmt);
        region = {}

        accWait_code = """
        #include "openacc.h"
        
        // Context substitution
        int main() {
         {  
            FRG__barrier();
         }
        }
        """

        accWait_subtree = self.parse_snippet(accWait_code, 
                                             {'region' : region}, 
                                             name = 'Barrier node', 
                                              show = False).ext[-1].body
        ReplaceTool(new_node = accWait_subtree.block_items[-1], 
                    old_node = accWait_node.parent).apply(accWait_node.parent.parent, 'block_items')
        
        updateSymbolTableFromNode(accWait_node.parent.parent)
        return accWait_subtree
        
       
