# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

import config

from Frontend.C99 import c99_ast as c_ast


from Tools.Tree import getContainerAttribute
from Backends.Cuda.Visitors.CM_Visitors import *

from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, RemoveTool

from Tools import SourceStorage

from Frontend.SymbolTable import updateSymbolTableFromNode,\
    getSymbolTableFromNode

from Backends.Frangollo.Templates.FrangolloVarNode import createContextVarList,\
    createKernelParameterList

from Backends.Frangollo.Mutators.FM_PrivDecl import FM_PrivDecl

from Backends.Common.Mutators.AbstractMutator import AbstractMutator
from Backends.C99.Visitors.LlcFilters import AccLoopFilter, AccKernelsReverse,\
    AccScopeReverse, FuncDefReverse

# CUDA Code generation
from Backends.Cuda.Kernelize import CudaKernelize
from Backends.Cuda.Files.FileTypes import Cuda_FileType
from Backends.Cuda.Platform import build_cuda_system_parameters


# OpenCL Code generation
from Backends.OpenCl.Kernelize import  CLKernelize
from Backends.OpenCl.Files.FileTypes import OpenCl_FileType
#from Backends.OpenCl.Platform import build_opencl_system_parameters

from Tools.Tree import getFileNode

from MiddleEnd.Loop.Analysis import AccLoopParametrize


from MiddleEnd.Functions.Mutators.Outliner import block_is_SESE, OutlineMutator,\
    ParameterStyle, c_outline_loop_stmts

from MiddleEnd.Loop.Mutators.LoopInvariant import LoopInvariantMutator
from Backends.C.Files.FileTypes import C_FileType
from Backends.Frangollo.Mutators.Common import AbstractFrangolloMutator


class FM_AccLoop(AbstractFrangolloMutator):
    instantiationNumber = 0

    def __init__(self, *args, **kwargs):
        self._kernel_name = kwargs['kernel_name']
        self._clauses = kwargs['clauses']
        self._source_storage = kwargs['source_storage']
        self._kernelList = []
        self._parentRegion = None
        if kwargs.has_key('symbol_table'):
        	self._st = kwargs['symbol_table']
	else:
        	self._st = None
        if kwargs.has_key('xmlm'):
#            print " Has key, value: " + str(kwargs['xmlm'])
            self._xmlm = kwargs['xmlm']
        else:
            self._xmlm = None
        self._backend_name = "Frangollo"
        super(FM_AccLoop, self).__init__(*args, **kwargs)


    def getExtractedKernelList(self,):
        return self._kernelList
#------------------------------------------------------------------------------ 
    #--------------------------------- def  _get_scope_clause_dict(self, node,):
        #-------------------------- """ Builds the OpenMP + llc scope dictionary
        #---------------------- :param node: Node from which extract information
        #---------------------------------------- :return: Dictionary of clauses
        #------------------------------------------------------------------- """
        #-------------------------------- # from Tools.Debug import DotDebugTool
        # # DotDebugTool(highlight = [node, node.parent, node.parent.parent]).apply(getFileNode(node))
        #--------------- llcfilt = AccScopeReverse(rootNode = getFileNode(node))
        #--------------------------------------------------- llcfilt.apply(node)
        #------------------------------------------------ return llcfilt.clauses
#------------------------------------------------------------------------------ 

    def filter(self, ast):
        """ Filter definition
            Returns the first node matching with the filter
        """
        # Build a visitor , matching the OmpFor node of the AST
        f = AccLoopFilter()
        node = f.apply(ast)
        self._parentRegion = f.get_parallel()
        return node

    def apply_all(self, parent_region_node, ast):
        """ Apply mutation to all matches """
        start_node = None
        self.ast = ast
        f = AccLoopFilter()
        num = 0;
        try:
            for elem in f.iterate(ast):
                # Save previous state
                old_name = self._kernel_name
                # old_clauses = self._clauses
                self._kernel_name = self._kernel_name + str(num)
                # Current scope variables
                self._parentRegion = f._parentRegion
               # If current node is not child of first parallel node, stop
                if self._parentRegion != parent_region_node:
                    # print " Different Parent region "
                    continue
                start_node = self.mutatorFunction(ast, elem)
                # Restore previous state
                self._kernel_name = old_name
                # self._clauses = old_clauses
                num += 1;
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            return self._parentRegion
        return start_node


    def generic_optimizations(self, subtree,):
        dGraph = self.compute_dgraph(subtree)
#        from Tools.Debug import DGraphDebug
#        DGraphDebug().apply(dGraph)
        # Loop Invariant
        if "invariant" in config.kernel_optimizations:
            try:
                LoopInvariantMutator(self._st, dGraph).apply(subtree)
            except NodeNotFound:
                print " Could not apply loop invariant optimization "
                print " --> Translation may be broken "
        return dGraph

    def mutatorFunction(self, ast, accLoop_node):
        """ CUDA mutator, writes memory transfer operations for a parallel region
        """
        if not self._st:
            self._st = getSymbolTableFromNode(accLoop_node)
        # General region parameters
        clauses = self._get_scope_clause_dict(accLoop_node.stmt);

        region = {}
        region['name'] = self.fname + '_' + (clauses['NAME'] or "random")
        region['vars'] = self.extract_directionality(ast, clauses) 

        # Check what to do with this particular acc loop
        # Default behavior 
        transform = True
        buildCuda = True
        buildCl = True
        collapse = True
        # Update the kernel name
        FM_AccLoop.instantiationNumber += 1

        # print "Cluster " + xmlm.cluster.get("name") + "."
        # Assuming the first cluster is the one you want
        if self._xmlm:
                print " XML Description file provided "
                xml_regions = self._xmlm.find_cluster("default")
                if self._xmlm.find_cluster("default"):
                    # If we could read the xml, lets get information from it
                    if self._xmlm.find_region(region['name']):
                        # If we found the region, then we have to look for this element
                        kname = region['name'] + "_" + str(FM_AccLoop.instantiationNumber)
                        if self._xmlm.find_element(kname):
                            # This is the element that we are processing now
                            pass
        else:
            #~ print "No XML Description file provided "
            pass

       
        if not transform:
            # Remove the llc for pragma and quit
            rt = ReplaceTool(new_node = accLoop_node.stmt, 
                            old_node = accLoop_node.parent)
            rt.apply(accLoop_node.parent.parent, 'block_items')
            return accLoop_node.stmt
        


        # For/Kernel parameters
        loop_parameters = AccLoopParametrize().apply(accLoop_node.stmt,)
        #~ print " Loop parametrize OK "
        loop2D_parameters = None
        loop3D_parameters = None
        ncollapse = 0
        if clauses.has_key('COLLAPSE') and type(clauses['COLLAPSE']) != type([]):
            ncollapse = clauses['COLLAPSE']
        else:
            collapse = False

        if (clauses.has_key('COLLAPSE') and (ncollapse <= 1)):
            #~ print " Collapse have been disabled "
            collapse = False

	
        inner_node = accLoop_node.stmt

        # Look for nested loop clauses
        for pragma_acc in AccLoopFilter().iterate(inner_node):
            # print ">>> Found nested loop clause (which is not implemented!) <<<<"
            #loop = pragma_acc.stmt
            # ReplaceTool(new_node = loop, old_node = pragma_acc.parent).apply(pragma_acc.parent.parent, getContainerAttribute(pragma_acc.parent))
            #loop.parent = pragma_acc.parent.parent
            # print "~~~~~~ The clause was removed ~~~~~~"
            collapse = 2 # TODO: Implement collapse 3
    
        inner_node = None
        # TODO: Move this stuff to a function 
        if collapse:
            from Frontend.Acc.acc_ast import AccLoop
            if isinstance(accLoop_node.stmt.stmt, c_ast.For):
               #~ print " Simple for "
               inner_node = accLoop_node.stmt.stmt
               inner_clause = None
            elif isinstance(accLoop_node.stmt.stmt.block_items[0], c_ast.For):
               #~ print " For inside a Compound "
               inner_node = accLoop_node.stmt.stmt.block_items[0]
               inner_clause = None
            elif hasattr(accLoop_node.stmt.stmt.block_items[0], 'stmt') and isinstance(accLoop_node.stmt.stmt.block_items[0].stmt, AccLoop):
               #~ print " Loop clause inside a compound "
               inner_node = accLoop_node.stmt.stmt.block_items[0].stmt.stmt
               inner_clause = accLoop_node.stmt.stmt.block_items[0] 
            else:
                print " Cannot create a nested loop because inner loop was not found "
                #~import pdb
                #~pdb.set_trace()
                raise NotImplemented
            loop2D_parameters = AccLoopParametrize().apply(inner_node,)
            if inner_clause:
                #~ from Tools.Debug import DotDebugTool
                #~ DotDebugTool(highlight = [inner_node, inner_clause]).apply(inner_node.parent.parent.parent.parent)
                ReplaceTool(new_node = inner_node, old_node = inner_clause.parent).apply(inner_clause.parent.parent,getContainerAttribute(inner_clause.parent))  
                #~ DotDebugTool(highlight = [inner_node, inner_clause]).apply(inner_node.parent.parent.parent.parent)
            if (ncollapse == 3):
                print " 3Dimensional nested loop "
                loop3D_parameters = AccLoopParametrize().apply(accLoop_node.stmt.stmt.stmt,)
                inner_node = accLoop_node.stmt.stmt.stmt
            else:
                print " 2D nested loop " + str(ncollapse)
        else:
           inner_node = accLoop_node.stmt.stmt
		
        if not inner_node:
            print " Inner node is None, something is wrong "
            raise NotImplemented
    
        # Extract loop contents as functions
        kernel_name = region['name'] + "_" + str(FM_AccLoop.instantiationNumber)
        from Backends.C99.Writers.AccWriter import AccWriter
        om = OutlineMutator(name = kernel_name + "_support", replace = False, change_code_block = False, writer = AccWriter) 
        # om.add_additional_decls([elem.node for elem in loop_parameters['global_vars']])
	om.add_additional_decls(loop_parameters['global_vars']);
        om.apply(accLoop_node.stmt,)
        #import pdb
	#pdb.set_trace()
        # Apply platform-independent optimizations
        subtree = inner_node
        
        optimize = True
        dGraph = None
        if optimize:
            dGraph = self.generic_optimizations(subtree)
        # om.setOutlinedFunction(subtree)
        accFor_subtree = self.build_kernel(clauses, region, loop_parameters, 
                                            accLoop_node, ast, buildCuda, buildCl,
                                            om, collapse, loop2D_parameters, 
                                            loop3D_parameters, dGraph)
                                            
        #print "accFor_subtree", type(accFor_subtree)
        
        parent_node = accLoop_node
        prev = None
        attr = 'block_items'
        
        prev, parent_node = self.lookParentWithAttr(parent_node, prev, attr)
        
        #print "accLoop_node", type(accLoop_node), accLoop_node
        
        # In those cases where a pragma loop precedes immediatly another pragma
        if not prev.children().__contains__(accLoop_node) and \
          isinstance(accLoop_node.parent, c_ast.Pragma):
                #print "***********************"
                parent_node = accLoop_node.parent
                prev = accLoop_node
                attr = 'stmt'
        
        #~ print "parent_node", type(parent_node), parent_node
        #~ print "prev", type(prev)

        # Replace call
        ReplaceTool(new_node = accFor_subtree, old_node = prev).apply(parent_node, attr)
 
        # Update Tree (sequence, depth, etc)
        self._st = updateSymbolTableFromNode(accLoop_node)
       
        return accFor_subtree
        
        # Replace the entire pragma by a CompoundStatement with all the new statements
        #        DotDebugTool(highlight = [ompFor_node]).apply(self._parentRegion)
        #if isinstance(ompFor_node.parent.parent, c_ast.Compound) :
            #ReplaceTool(new_node = cuda_stmts, old_node = ompFor_node.parent).apply(ompFor_node.parent.parent, 'stmts')
        #else:
            # Maybe we have an error here
#            ompFor_node.parent.parent.stmt = cuda_stmts



    def build_kernel(self, clauses, region, loop_parameters, llcFor_node, ast, \
                        buildCuda, buildCl, om, collapse, loop2D_parameters, loop3D_parameters, dGraph = None):
        # Replace the kernel call to a Frangollo kernel launch statement
        llcFor_code = ""
        if config.FRANGOLLO_DEBUG:
            llcFor_code += "#define DEBUG"

        llcFor_code += """
         <%!
           from Backends.C99.Writers.AccWriter import AccWriter
           from Frontend.InternalRepr import AstToIR
        %>
        
        // Get Frangollo interface definitions
        #include "openacc.h"
        
        %for typedef in typedeflist:
            ${typedef};
        %endfor        
        
        int main() {
               if (FRG__offloadKernel(FRG__getActiveContext(), "${kernel['name']}") && (${cond_expr})) 
	       {
                  tparams *__FRG__parameter_list_head, * __FRG__parameter_list_prev;
	         		var_layout_t var_layout, acc_layout;

		  __FRG__parameter_list_prev = NULL;
		  __FRG__parameter_list_head = NULL;
			void * ___ptr_to_var;

                  %for var in kernel['parameters']:

	         {
                  %if var.frangolloType == "VAR_SCALAR":
			___ptr_to_var = &${var};
                  %else:
			___ptr_to_var = ${var};
                  %endif 

		   %if var.force_register:
			{
			var_layout.x = ${var.size};
			var_layout.y = 0;
			var_layout.type = ${var.frangolloType};
			FRG__registerVar(___ptr_to_var, regionId, 0 | 
					    (VAR_INOUT * 0x10) | 
					    (1 * 0x100),
					    var_layout, &acc_layout);
                	}
		    %endif

		  %if var.reductionOp:
                        NEW_KERNEL_PARAM("${var.name}", ___ptr_to_var, 
					  ${var.frangolloType}, ${var.reductionOp}, 
					  ${var.reductionType}, "0", ${var.canChange}, ${var.size});
		  %else:
		        NEW_KERNEL_PARAM("${var.name}", ___ptr_to_var, 
					  ${var.frangolloType}, "", 
					  "", "0", ${var.canChange}, ${var.size});

		  %endif
                  
                 }  

                  %endfor

                      FRG__submit_kernel_auto(FRG__getActiveContext(), 
						"${kernel['name']}", __FRG__parameter_list_head, 
						${kernel['iterations'][0]}, 
                                                ${kernel['iterations'][1]}, ${kernel['iterations'][2]}, 
                                                ${kernel['flops']}, ${kernel['memAccess']});
                    tparams * __FRG_tmp_parameter_list__;
                    while (__FRG__parameter_list_head) {
                        __FRG_tmp_parameter_list__ = __FRG__parameter_list_head;
                        __FRG__parameter_list_head = __FRG__parameter_list_head->next;
                        free(__FRG_tmp_parameter_list__);
                    }
                    %if kernel['async']:
                        FRG__barrier();
                    %endif

          } 
	       else 
	       {
                   /* Call to the original code outlined to a external function */
                   /* ${callToOriginal}; */
                   ${AstToIR(Writer=AccWriter, ast=original_code).ast}
               }
        }
        """
	# Assuming LLC --> FOR --> STMT
        system_parameters = build_cuda_system_parameters(loop_parameters, 
                                                          llcFor_node.stmt.stmt, 
                                                          loop2D_parameters, 
                                                          loop3D_parameters,)    
        # Create kernel dictionary, wich is just a shortcut for the template
        kernel = {}
        if collapse and loop3D_parameters:
            kernel['iterations'] = (loop_parameters['number_of_iterations'], 
                                    loop2D_parameters['number_of_iterations'],
                                    loop3D_parameters['number_of_iterations'],
                                   )
        elif collapse:
            kernel['iterations'] = (loop_parameters['number_of_iterations'], 
                                    loop2D_parameters['number_of_iterations'], "-1")
        else:
            kernel['iterations'] = (loop_parameters['number_of_iterations'],
                                        "-1",
                                        "-1",)
        kernel['name'] = region['name'] + "_" + str(FM_AccLoop.instantiationNumber)
        kernel['flops'] = system_parameters['flops']
        kernel['memAccess'] = system_parameters['memAccess']

        self._kernelList.append(kernel)
        # Notice how kernel parameters are not the same as region variables. 
        kernel['parameters'] = createKernelParameterList(None, 
                                None, 
                                system_parameters['original_parameters'], 
                                clauses['REDUCTION'], clauses['FIRSTPRIVATE'], ast, dGraph)
        if clauses.has_key('ASYNC'):
            kernel['async'] = clauses['ASYNC']
        else:
            kernel['async'] = None

        if clauses.has_key('IF'):
            if clauses['IF']:
                cond_expr = str(clauses['IF']);
            else:
                cond_expr = "1"
        else:
            cond_expr = "1"

        system_parameters['kernel_name'] = kernel['name']
        llcFor_subtree = self.parse_snippet(llcFor_code, {'region' : region, 'kernel' : kernel, 
                                           'callToOriginal' : None, # om.getOutlinedFunctionCall(), 
					   'original_code' : llcFor_node.stmt,
					   'cond_expr' : cond_expr,
					   'typedeflist' : om._typedefList}, 
                                                name = 'For kernel', show = False).ext[-1].body        
 
        # CUDA kernel
        # Extract kernel from the loop
        kernel_source = CudaKernelize().cuda_outline_loop_stmts(loop_parameters, 
                                                    system_parameters, 
                                                    llcFor_node.stmt.stmt,
                                                    kernel['name'],
                                                    loop2D_parameters,
                                                    loop3D_parameters,
                                                    om = om) 
        
        # Add source file to source storage
        self._source_storage.addFile(kernel['name'], Cuda_FileType())
        # No particular include file is used here
        self._source_storage.append(kernel['name'], Cuda_FileType(), kernel_source)


        # OpenCl kernel
        # Extract kernel from the loop
        # TODO : Probably, system_parameters should be different for OpenCL than for CUDA
        #         as the transformation/parametrization phase is probably different
        kernel_source = CLKernelize().opencl_outline_loop_stmts(loop_parameters, 
                                            system_parameters, 
                                            llcFor_node.stmt.stmt, 
                                            kernel['name'],
                                            loop2D_parameters,
                                            loop3D_parameters,
                                            om = om) 
        
        # Add source file to source storage
        self._source_storage.addFile(kernel['name'], OpenCl_FileType())
        # No particular include file is used here
        self._source_storage.append(kernel['name'], OpenCl_FileType(), kernel_source)
        
        # TODO: c_outline does not properly tracks functions and types, this should be llcFor_node.stmt.stmt
        #        but for some reason it does not work
        kernel_source = c_outline_loop_stmts(llcFor_node.stmt.stmt,om)        
        # Add source file to source storage
        self._source_storage.addFile(kernel['name'], C_FileType())
        # No particular include file is used here
        if config.GENERATE_HOST_CODE:
            self._source_storage.append(kernel['name'], C_FileType(), kernel_source)
        else:
            self._source_storage.append(kernel['name'], C_FileType(), ";")
        
        return llcFor_subtree


