# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

from Frontend.C99 import c99_ast as c_ast

from Frontend.Acc import acc_ast


from Backends.Cuda.Visitors.CM_Visitors import *

from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, RemoveTool,\
    getContainerAttribute

from Tools import SourceStorage

from Backends.Frangollo.Templates.FrangolloVarNode import createContextVarList, FRANGOLLO_VAR_DIRECTIONS
from Backends.Common.Mutators.AbstractMutator import AbstractMutator
from Backends.C99.Visitors.LlcFilters import AccKernelsReverse, AccScopeReverse, AccAnyRegionReverse

from Tools.Tree import getFileNode

from Backends.Common.Visitors.GenericVisitors import GenericFilterVisitor
from Backends.Frangollo.Templates.FrangolloVarNode import FrangolloVarNode
from Frontend.SymbolTable import getSymbolTableFromNode

class ContextDeclFilter(GenericFilterVisitor):
    """ Returns the first node with a Decl
    """
                #self._parentRegion = AccKernelsReverse().apply(elem)

    def __init__(self, parent_region, prev_brother = None):
        def condition(node):
            return type(node) == c_ast.Decl and (parent_region == AccAnyRegionReverse().apply(node))
        super(ContextDeclFilter, self).__init__(condition_func = condition, prev_brother = prev_brother)
    
    def visit_AccLoop(self, node, offset = 1, ignore = []):
        # print " Loop: Stop search"
        # Do not traverse this subtree (will be converted to a kernel)
        return None # self.generic_visit(node, offset, ignore)        

    def visit_AccParallel(self, node,  offset = 1, ignore = []):
        # print " Visit Acc Parallels "
        if isinstance(node.stmt, acc_ast.AccLoop):
            # print " For: Stop search"
            # Do not traverse this subtree (will be converted to a kernel)
            return None # self.generic_visit(node, offset, ignore)        
        return self.generic_visit(node,  offset, ignore)


    def visit_AccKernels(self, node,  offset = 1, ignore = []):
        print " Visiting AccKernels "
        if isinstance(node.stmt, acc_ast.AccLoop):
            # print " For: Stop search"
            # Do not traverse this subtree (will be converted to a kernel)
            return None # self.generic_visit(node, offset, ignore)        
        return self.generic_visit(node,  offset, ignore)





class FM_PrivDecl(AbstractMutator):
    instantiationNumber = 0

    def __init__(self, *args, **kwargs):
        self._clauses = kwargs['clauses']
        self._st = kwargs['source_storage']
        if kwargs.has_key('xmlm'):
#            print " Has key, value: " + str(kwargs['xmlm'])
            self._xmlm = kwargs['xmlm']
        else:
            self._xmlm = None
        self._backend_name = "Frangollo"

        super(FM_PrivDecl, self).__init__(backend_name = self._backend_name)

    def  _get_scope_clause_dict(self, node,):
        """ Builds the OpenMP + llc scope dictionary 
        :param node: Node from which extract information
        :return: Dictionary of clauses
        """
        # from Tools.Debug import DotDebugTool
        # DotDebugTool(highlight = [node, node.parent, node.parent.parent]).apply(getFileNode(node))
        llcfilt = AccScopeReverse(rootNode = getFileNode(node))
        llcfilt.apply(node)
        return llcfilt.clauses
    

    def filter(self, ast):
        """ Filter definition
            Returns the first node matching with the filter
        """
        # Build a visitor , matching the OmpFor node of the AST
        f = ContextDeclFilter()
        node = f.apply(ast)
#        self._parentRegion = f.get_parallel()
        return node

    def apply_all(self, region_node, fAST):
        """ Apply mutation to all matches """
        start_node = None
        self.ast = fAST
        f = ContextDeclFilter(region_node)
        num = 0;
        try:
            for elem in f.iterate(region_node):
                #self._parentRegion = AccKernelsReverse().apply(elem)
                # If current node is not child of first parallel node, stop
                #if self._parentRegion != parent_parallel_node:
                #    print "Not parent"
                #    continue
                #self.source_storage.addFile(self._kernel_name, SourceStorage.CUDA_SOURCE)
                #self.source_storage.append(self._kernel_name, SourceStorage.CUDA_SOURCE, '#include "llcomp_cuda.h"\n')
                start_node = self.mutatorFunction(fAST, elem)
                # self._clauses = old_clauses
                num += 1;
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            return None # self._parentRegion
        return start_node

    def mutatorFunction(self, fAST, ctxtDecl_node):
        """ Add a CreateVar statement for each private declaration available
        """
        createVar_code = """int main() {
                    FRG__registerVar(&(${var.name}), getActiveContext(), ${var.size}, ${var.direction});
                    } 
            """
        st = getSymbolTableFromNode(ctxtDecl_node)
        symbol = st.lookUpDecl(ctxtDecl_node)
        fvn = FrangolloVarNode(symbol = symbol, ast = self.ast, direction = FRANGOLLO_VAR_DIRECTIONS["PRIVATE"])
        createVar_subtree = self.parse_snippet(createVar_code, {'var' : fvn }, name = 'Private var declaration',
                 show = False).ext[-1].body.block_items[-1]
#        print " *** PrivDecl " 
#        createVar.show()        

        itool = InsertTool(subtree = createVar_subtree, 
                        position = "after", 
                       prev = ctxtDecl_node)
        itool.apply(ctxtDecl_node.parent, getContainerAttribute(ctxtDecl_node))

        return ctxtDecl_node
        
