# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
Created on Jun 24, 2011

@author: rreyes
'''
from Backends.Common.TemplateEngine.TemplateParser import TemplateVarNode
from Tools.Tree import getFileNode
from Frontend.SymbolTable import getSymbolTableFromNode
from Backends.Common.TemplateEngine.Functions import is_complex_type

class FrangolloVarNode(TemplateVarNode):
    '''
    @author: Ruyman
    '''


    @property
    def direction(self):
        return self._directionality
    
    def __init__(self, symbol, ast, direction, present, canChange = True, force_register = False):
        """ Build the TemplateVarNode making a clausure
                with the modifier functions.
        """
        self._ast = ast
        self._type = symbol.node.type 
        self._type_func = lambda elem: elem.type  #~~~ DEPRECATED
        self._name_func = lambda elem : elem.name #~~~ DEPRECATED
        
        # Get the symbol table from any node (they share the same FileAST)
#        self._st = getSymbolTableFromNode(getFileNode(var))
#        if var.name in ['pos', 'vel',]:
#            import pdb
#            pdb.set_trace()
        self._symbol = symbol 
        if symbol.extra.has_key('matrixInfo'):
            self._matrixInfo = symbol.extra['matrixInfo']
        else:
            self._matrixInfo = None
        self._var =  self._symbol.node  # This node is the declaration
        self._typedef = self._get_typedef_from_node(self._var, ast)
        self._structDecl = self._get_structdecl_from_node(self._var, ast)
        self._directionality = direction
        self._reductionOp = None 
        self._reductionType = None 
        self._canChange = canChange
        self._present = present
	self._force_register = force_register
    
    @property
    def frangolloType(self,):
        if is_complex_type(self._var) and not self._matrixInfo:
            return "VAR_ARRAY"
        elif self._matrixInfo and is_complex_type(self._var):
            return "VAR_ARRAY2D"
        else:
            return "VAR_SCALAR"

    @property
    def canChange(self,):
        if self._canChange:
            return 1
        else:
            return 0

    @canChange.setter
    def canChange(self, value):
        self._canChange = value

    @property
    def force_register(self,):
	return self._force_register

    @force_register.setter
    def force_register(self, value):
	self._force_register = value

    @property
    def present(self,):
        if self._present:
            return 1 # "COPY_FIRST_ONLY"
        else:
            return 0 # "FORCE_COPY_IN"

    @present.setter
    def present(self, value):
        self._present = value

    @property
    def reductionOp(self,):
        """ If a reduction operation is to be applied within this variable
                           this property specifies the reduction operation.
        """
        return self._reductionOp


    @reductionOp.setter
    def reductionOp(self, value):
        """ If a reduction operation is to be applied within this variable
                           this property specifies the reduction operation.
        """
        self._reductionOp = value

    @property
    def reductionType(self,):
        """ If a reduction operation is to be applied within this variable
                           this property specifies the type of the reduction operation.
        """
        return self._reductionType


    @reductionType.setter
    def reductionType(self, value):
        """ If a reduction operation is to be applied within this variable
                           this property specifies the type reduction operation.
        """
        self._reductionType = value

    def isReduction(self):
	return self._reductionOp == None


FRANGOLLO_VAR_DIRECTIONS = {"IN" : 'VAR_IN', 
                            "OUT" : 'VAR_OUT', 
                            "INOUT" : 'VAR_INOUT', 
                            "PRIVATE": 'VAR_PRIVATE', 
                            "NONE"   : 'VAR_NONE',
                            "UNKNOWN" : '',
                            }


def createContextVarList(copyIn_clause, copyOut_clause, ast,):
    """ Convert an array list to a template array 
    
        :param var_list: List with Decl elements
        :param ast: Subtree
        :param func: Type of element from var_list which is added to the template array
        :param name_func: Function to get the name of each element (default elem.name)
        :param type_func: Function to get the type of each element (default elem.type)
        :return: List with TemplateVarNode elements
    
    """
    names = []
    onlyOut = copyOut_clause[:]

    for in_symbol in copyIn_clause:
        isIn = True
        isOut = False
        for out_symbol in copyOut_clause:
            if out_symbol.name == in_symbol.name:
                isOut = True
                onlyOut = [ elem for elem in onlyOut if elem.name != out_symbol.name]
                break
        if isIn and isOut:
            direction = 'INOUT'
        elif isIn and not isOut:
            direction = 'IN'
        elif not isIn and isOut:
            direction = 'OUT'
        else:
            direction = 'INOUT'
           
        names.append(FrangolloVarNode(symbol = in_symbol, ast = ast, direction = FRANGOLLO_VAR_DIRECTIONS[direction]), None)

    for out_symbol in onlyOut:
        names.append(FrangolloVarNode(symbol = out_symbol, ast = ast, direction = FRANGOLLO_VAR_DIRECTIONS['OUT']), None)


           
    return names 
 
def createKernelParameterList(copyIn_clause, copyOut_clause, kernel_parameters, reduction_clause, 
			      firstprivate_clause, ast,dGraph = None):
    """ Convert an array list to a template array 
    
    """
    # Kernel parameters are : (shared + reduction - local_vars)
    names = []
    if copyIn_clause and copyOut_clause:
            for symbol in kernel_parameters:
                isIn = False
                isOut = False
                for in_symbol in copyIn_clause:
                    if in_symbol.name == symbol.name:
                        isIn = True
                        break
                for out_symbol in copyOut_clause:
                    if out_symbol.name == symbol.name:
                        isOut = True
                        break
                if isIn and isOut:
                    direction = 'INOUT'
                elif isIn and not isOut:
                    direction = 'IN'
                elif not isIn and isOut:
                    direction = 'OUT'
                else:
                    direction = 'INOUT'
		force_register = False
		for fp_symbol in firstprivate_clause:
			if fp_symbol.name == symbol.name:
				force_register = True
				direction = 'IN'
				break

                names.append(FrangolloVarNode(symbol = symbol, ast = ast, 
                            direction = FRANGOLLO_VAR_DIRECTIONS[direction], 
                            present = None, canChange = True, force_register = force_register))
    else:
        direction = 'UNKNOWN'
        for symbol in kernel_parameters:
            canChange = False
            try:
                if dGraph:
                  canChange = dGraph[symbol].isWrite
                  #~ print " Symbol : " + str(symbol) + " can change? " + str(canChange)
                else:
                  canChange = True
            except KeyError:
                canChange = True
                #~ print " Symbol : " + str(symbol) + " can change because not found  "
	    force_register = False
	    for fp_symbol in firstprivate_clause:
		if fp_symbol.name == symbol.name:
			force_register = True
			direction = 'IN'
			break
            names.append(FrangolloVarNode(symbol = symbol, ast = ast,
                                          direction = FRANGOLLO_VAR_DIRECTIONS[direction], 
                                          present = None, canChange = canChange, 
					  force_register = force_register))
   
    # Add which reduction operation (if any) is going to be applied to the FrangolloVar
    for op,varlist in reduction_clause:
        for elem in varlist:
           for fvn in names:
                if elem.name == fvn.ptr.name:
                    fvn.reductionOp = "\"a" + op + "b\""
                    fvn.reductionType = "\"" + elem.btype + "\""
		    fvn.force_register = True

    return names    
