# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

from Frontend.C99 import c99_ast as c_ast


from Backends.Cuda.Visitors.CM_Visitors import *

from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, RemoveTool

from Tools import SourceStorage

from Backends.GMAC.Templates.GMACVarNode import createRegionVarList,\
    createKernelParameterList
from Backends.Common.Mutators.AbstractMutator import AbstractMutator
from Backends.C99.Visitors.LlcFilters import AccLoopFilter, AccKernelsReverse,\
    AccScopeReverse

# CUDA Code generation
from Backends.Cuda.Kernelize import cuda_outline_loop_stmts
from Backends.Cuda.Files.FileTypes import Cuda_FileType
from Backends.Cuda.Platform import build_cuda_system_parameters


# OpenCL Code generation
from Backends.OpenCl.Kernelize import opencl_outline_loop_stmts
from Backends.OpenCl.Files.FileTypes import OpenCl_FileType
#from Backends.OpenCl.Platform import build_opencl_system_parameters


from MiddleEnd.Loop.Analysis import AccLoopParametrize


from Backends.Cuda.Writers.CUDAWriter import CUDAWriter

from Backends.Common.TemplateEngine.TemplateParser import TemplateParser


from Frontend.FrontendFactory import FrontendFactory

import config

class FM_AccLoop(AbstractMutator):
    instantiationNumber = 0

#    def parse_snippet(self, template_code, subs_dir, name, show = False, languages = None):
#        """ Transform a template code snippet to the IR
#
#            :param template_code: String with the source code
#            :param subs_dir: Dictionary with substitutions
#            :param name: Name of the template to be parsed (debugging)
#            :param show: Show the template after variable substitution (debugging)
#            :return: IR-1 representation of the snippet
#
#        """
#        subtree = None
#        if subs_dir:
#            template_code = TemplateParser(self._backend_name, template_code).render(**subs_dir)
#            if show:
#                print " Template " + str(template_code)
#        try:
#            languages = ['C99', 'Omp', ]
#            [subtree_ast, prepro_class, lexer_class, parser_class,] = \
#                FrontendFactory().parse_with_language_list(template_code, name, languages, includes = [config.WORKDIR + "/Backends/" + self._backend_name + "/Templates/include/",] )
#            # Transform the C99 ast into the internal representation
#            # subtree_ast.show()
#            migrator = C99AstToIR(Writer = CudaWriter, ast = subtree_ast)
#            migrator.annotate()
#            subtree_IR = migrator.ast
#            #subtree = parse_source(template_code, name)
#        except IOError:
#                print "Pipe Error"
#                return None
#        #--------------------------------------------------------------- except:
#            #------------------------------------- print "Error parsing snippet"
#            #------------------------ print "Snippet was: " + str(template_code)
#            #-------------------------------------- raise AbortMutationException
#
#        return subtree_IR;
#
#

    def __init__(self, *args, **kwargs):
        self._kernel_name = kwargs['kernel_name']
        self._clauses = kwargs['clauses']
        self._st = kwargs['source_storage']
        if kwargs.has_key('xmlm'):
#            print " Has key, value: " + str(kwargs['xmlm'])
            self._xmlm = kwargs['xmlm']
        else:
            self._xmlm = None
        self._backend_name = "GMAC"
        super(FM_AccLoop, self).__init__(backend_name = self._backend_name)

    def  _get_scope_clause_dict(self, node,):
        """ Builds the OpenMP + llc scope dictionary 
        :param node: Node from which extract information
        :return: Dictionary of clauses
        """
        # from Tools.Debug import DotDebugTool
        # DotDebugTool(highlight = [node, node.parent, node.parent.parent]).apply(getFileNode(node))
        llcfilt = AccScopeReverse()
        llcfilt.apply(node)
        return llcfilt.clauses
    

    def filter(self, ast):
        """ Filter definition
            Returns the first node matching with the filter
        """
        # Build a visitor , matching the OmpFor node of the AST
        f = AccLoopFilter()
        node = f.apply(ast)
        self._parentRegion = f.get_parallel()
        return node

    def apply_all(self, parent_parallel_node, ast):
        """ Apply mutation to all matches """
        start_node = None
        self.ast = ast
        f = AccLoopFilter()
        num = 0;
        try:
            for elem in f.iterate(ast):
                # Save previous state
                old_name = self._kernel_name
                # old_clauses = self._clauses
                self._kernel_name = self._kernel_name + str(num)
                # Current scope variables
                self._parentRegion = AccKernelsReverse().apply(elem)
                # If current node is not child of first parallel node, stop
                if self._parentRegion != parent_parallel_node:
                    raise StopIteration
                #self.source_storage.addFile(self._kernel_name, SourceStorage.CUDA_SOURCE)
                #self.source_storage.append(self._kernel_name, SourceStorage.CUDA_SOURCE, '#include "llcomp_cuda.h"\n')
                start_node = self.mutatorFunction(ast, elem)
                # Restore previous state
                self._kernel_name = old_name
                # self._clauses = old_clauses
                num += 1;
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            return self._parentRegion
        return start_node

    def mutatorFunction(self, ast, accLoop_node):
        """ CUDA mutator, writes memory transfer operations for a parallel region
        """
        # General region parameters
        clauses = self._get_scope_clause_dict(accLoop_node.stmt);
        region = {}
        region['name'] = clauses['NAME'] or "random"
        region['vars'] = createRegionVarList(clauses['COPY_IN'], clauses['COPY_OUT'], ast,) 

        # Check what to do with this particular llcFor
        # Default behavior 
        transform = True
        buildCuda = True
        buildCl = True
        # Update the kernel name
        FM_AccLoop.instantiationNumber += 1


        # print "Cluster " + xmlm.cluster.get("name") + "."
        # Assuming the first cluster is the one you want
        if self._xmlm:
                print " XML Description file provided "
                xml_regions = self._xmlm.find_cluster("default")
                if self._xmlm.find_cluster("default"):
                    # If we could read the xml, lets get information from it
                    if self._xmlm.find_region(region['name']):
                        # If we found the region, then we have to look for this element
                        kname = region['name'] + "_" + str(FM_AccLoop.instantiationNumber)
                        if self._xmlm.find_element(kname):
                            # This is the element that we are processing now
                            pass
        else:
            print "No XML Description file provided "

       
        if not transform:
            # Remove the llc for pragma and quit
            ReplaceTool(new_node = accLoop_node.stmt, old_node = accLoop_node.parent).apply(accLoop_node.parent.parent, 'block_items')
            return accLoop_node.stmt
        


        # For/Kernel parameters
        loop_parameters = AccLoopParametrize().apply(accLoop_node.stmt,)
        llcFor_subtree = None

        if (loop_parameters['ndims'] == 2):
            # Build two dimensional kernel
            print " Generating a two dimensional kernel for nested loop (name: " + str(region['name']) + ")"
            llcFor_subtree = self.build_2d_kernel(clauses, region, loop_parameters, accLoop_node, ast, buildCuda, buildCl)
        else:
            llcFor_subtree = self.build_1d_kernel(clauses, region, loop_parameters, accLoop_node, ast, buildCuda, buildCl)



        # Replace call
        ReplaceTool(new_node = llcFor_subtree, old_node = accLoop_node.parent).apply(accLoop_node.parent.parent, 'block_items')
        
        return llcFor_subtree
        
        # Replace the entire pragma by a CompoundStatement with all the new statements
        #        DotDebugTool(highlight = [ompFor_node]).apply(self._parentRegion)
        #if isinstance(ompFor_node.parent.parent, c_ast.Compound) :
            #ReplaceTool(new_node = cuda_stmts, old_node = ompFor_node.parent).apply(ompFor_node.parent.parent, 'stmts')
        #else:
            # Maybe we have an error here
#            ompFor_node.parent.parent.stmt = cuda_stmts



    def build_1d_kernel(self, clauses, region, loop_parameters, llcFor_node, ast, buildCuda, buildCl):
        # Replace the kernel call to a frangollo kernel launch statement
        llcFor_code = """
        // For substitution
        #include "llcFunctions.h"
            <%!
       from Backends.Common.TemplateEngine.Functions import varlist_to_funcproto, var_to_declaration, id_to_var_node, array_to_ptr, write_gmacKernelArgs
    %>
        int main() {
               {
                  // Kernel name is appended to region name
                  // launch_kernel(regionId, "${kernel['name']}", head, ${kernel['iterations']});
                  ${kernel['name']} <<< 1,1 >>>(
                   ${",".join(write_gmacKernelArgs(kernel['parameters'])) }                 );
               }
        }
        """

	    # Assuming LLC --> FOR --> STMT
        system_parameters = build_cuda_system_parameters(loop_parameters, llcFor_node.stmt.stmt)    
               # Create kernel dictionary, wich is just a shortcut for the template
        kernel = {}
        kernel['iterations'] = loop_parameters['last_iteration']
        kernel['name'] = region['name'] + "_" + str(FM_AccLoop.instantiationNumber)
        # Notice how kernel parameters are not the same as region variables. 
        kernel['parameters'] = createKernelParameterList(clauses['COPY_IN'], clauses['COPY_OUT'], system_parameters['original_parameters'], clauses['REDUCTION'], ast)
        system_parameters['kernel_name'] = kernel['name']

        llcFor_subtree = self.parse_snippet(llcFor_code, {'region' : region, 'kernel' : kernel}, name = 'For kernel', show = True, languages = ['C99', 'Omp',]).ext[-1].body        
 

        # CUDA kernel
        # Extract kernel from the loop
        kernel_source = cuda_outline_loop_stmts(loop_parameters, system_parameters, llcFor_node.stmt.stmt) 
        
        # Add source file to source storage
        self._st.addFile(kernel['name'], Cuda_FileType())
        # No particular include file is used here
        self._st.append(kernel['name'], Cuda_FileType(), kernel_source)


        # OpenCl kernel
        # Extract kernel from the loop
        # TODO : Probably, system_parameters should be different for OpenCL than for CUDA
        #         as the transformation/parametrization phase is probably different
        kernel_source = opencl_outline_loop_stmts(loop_parameters, system_parameters, llcFor_node.stmt.stmt) 
        
        # Add source file to source storage
        self._st.addFile(kernel['name'], OpenCl_FileType())
        # No particular include file is used here
        self._st.append(kernel['name'], OpenCl_FileType(), kernel_source)
        return llcFor_subtree

    def build_2d_kernel(self, clauses, region, loop_parameters, llcFor_node, ast, buildCuda, buildCl):
        # Get 2nd loop parameter
        assert isinstance(llcFor_node.stmt.stmt, c_ast.For)
        secondloop_parameters = AccLoopParametrize().apply(llcFor_node.stmt.stmt,)
        # Replace the kernel call to a frangollo kernel launch statement
        llcFor_code = """
        // For substitution
        #include "llcFunctions.h"
        
        int main() {
               {
                  tparams *head, * params, * prev;
		  prev = NULL;
		  head = NULL;
                  %for var in kernel['parameters']:
                  params = (tparams *) malloc(sizeof(tparams));
                  if (!head) {
                          head = params;
                  }
                  if (prev) {
                          prev->next = params;
                  }
                  prev = params;
                  params->type = ${var.frangolloType};
                  params->reduction = ${var.reductionOp}; /* 0x0 '+' '*'  */
                  // Base address is used to identify the correspondant HostVar structure
                  %if var.frangolloType == "VAR_SCALAR":
                          params->_base_ptr = &${var};                    
                  %else:
                          params->_base_ptr = ${var}; 
                  %endif 
		      params->next = NULL;
                  %endfor
                  // Kernel name is appended to region name
                  launch_kernel2d(regionId, "${kernel['name']}", head, ${kernel['iterations'][0]}, ${kernel['iterations'][1]});
               }
        }
        """

	    # Assuming LLC --> FOR --> STMT
        system_parameters = build_cuda_system_parameters(loop_parameters, llcFor_node.stmt.stmt, secondloop_parameters)    
        # Update the kernel name
        FM_AccLoop.instantiationNumber += 1
        # Create kernel dictionary, wich is just a shortcut for the template
        kernel = {}
        kernel['iterations'] = (loop_parameters['last_iteration'], secondloop_parameters['last_iteration'],)
        kernel['name'] = region['name'] + "_" + str(FM_AccLoop.instantiationNumber)
        # Notice how kernel parameters are not the same as region variables. 
        kernel['parameters'] = createKernelParameterList(clauses['COPY_IN'], clauses['COPY_OUT'], system_parameters['original_parameters'], clauses['REDUCTION'], ast)
        system_parameters['kernel_name'] = kernel['name']

        llcFor_subtree = self.parse_snippet(llcFor_code, {'region' : region, 'kernel' : kernel}, name = 'For kernel', show = True).ext[-1].body        
 

        # CUDA kernel
        # Extract kernel from the loop
        kernel_source = cuda_outline_loop_stmts(loop_parameters, system_parameters, llcFor_node.stmt.stmt, kernel['name'], secondloop_parameters) 
        
        # Add source file to source storage
        self._st.addFile(kernel['name'], Cuda_FileType())
        # No particular include file is used here
        self._st.append(kernel['name'], Cuda_FileType(), kernel_source)


        # OpenCl kernel
        # Extract kernel from the loop
        # TODO : Probably, system_parameters should be different for OpenCL than for CUDA
        #         as the transformation/parametrization phase is probably different
        kernel_source = opencl_outline_loop_stmts(loop_parameters, system_parameters, llcFor_node.stmt.stmt) 
        
        # Add source file to source storage
        self._st.addFile(kernel['name'], OpenCl_FileType())
        # No particular include file is used here
        self._st.append(kernel['name'], OpenCl_FileType(), kernel_source)
        return llcFor_subtree
