# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------


from Backends.C99.Visitors.LlcFilters import AccKernelsFilter, AccLoopFilter,\
    AccScopeReverse

from Tools.Tree import InsertTool, NodeNotFound, ReplaceTool, getFileNode
from Frontend.C99 import c99_ast
from Tools.Declarations import decl_of_id

# from Backends.Cuda.Mutators.Common import AbstractCudaMutator


from Tools import SourceStorage
from Backends.Common.Mutators.AbstractMutator import AbstractMutator
from Backends.GMAC.Templates.GMACVarNode import createRegionVarList
from Backends.GMAC.Mutators.FM_LlcFor import FM_AccLoop

class FM_LlcRegion(AbstractMutator):
    def __init__(self, *args, **kwargs):
        self._parentRegion = None
        self._source_storage = kwargs['source_storage']
        self._backend_name = "GMAC"
        if kwargs.has_key('xmlm'):
#            print " Has key, value: " + str(kwargs['xmlm'])
            self._xmlm = kwargs['xmlm']
        else:
            self._xmlm = None
        super(FM_LlcRegion, self).__init__(backend_name = self._backend_name)
    
    def  _get_scope_clause_dict(self, node,):
        """ Builds the OpenMP + llc scope dictionary 
        :param node: Node from which extract information
        :return: Dictionary of clauses
        """
        # from Tools.Debug import DotDebugTool
        # DotDebugTool(highlight = [node, node.parent, node.parent.parent]).apply(getFileNode(node))
        llcfilt = AccScopeReverse()
        llcfilt.apply(node)
        return llcfilt.clauses
    
    def filter(self, ast):
        """ Filter definition
            Returns the first node matching with the filter
        
        """
        # Build a visitor , matching the OmpFor node of the AST
        # f = OmpParallelFilter(device = self.device)
        f = AccKernelsFilter()
        node = f.apply(ast)
        self._parentRegion = node
        return node

    def apply_all(self, ast):
        """ Apply mutation to all matches """
        self.ast = ast
        f = AccKernelsFilter()
        try:
            for elem in f.iterate(ast):
                self.mutatorFunction(ast, elem)
        except NodeNotFound as nf:
            print str(nf)
        except StopIteration:
            return self._parentRegion
        return ast

    def mutatorFunction(self, ast, llcRegion_node):
        """ CUDA mutator, writes memory transfer operations for a parallel region
        """
        ##################### Statement for GMAC region
        
        
        llcRegion_code = """
        #include "llcFunctions.h"
            <%!
       from Backends.Common.TemplateEngine.Functions import varlist_to_funcproto, var_to_declaration, id_to_var_node, array_to_ptr
    %>
        // Region substitution
        int main() {
         // Variable declaration and allocation
         {
         %for var in region['vars']:
             %if not var.frangolloType == "VAR_SCALAR":
                 gmacMalloc(&${var.name}, ${var.size});
             %endif
         %endfor
         }
         {
             placeholder();
         }
         {
         %for var in region['vars']:
             %if not var.frangolloType == "VAR_SCALAR":
                 gmacFree(&${var.name});
             %endif
         %endfor
         }
        }
        
        """

        clauses = self._get_scope_clause_dict(llcRegion_node.stmt);
        region = {}
        region['name'] = clauses['NAME'] or "random"
        region['vars'] = createRegionVarList(clauses['COPY_IN'], clauses['COPY_OUT'], ast,) 
 
        # Translate all constructs within the region
        # 1. Search for llcFor constructs
        # 2. Search for llcSection construct

        FM_AccLoop(clauses = clauses, kernel_name = region['name'] + "_loopKernel", source_storage = self._source_storage, xmlm = self._xmlm).apply_all(llcRegion_node, ast)


        # import pdb
        # pdb.set_trace()
        llcRegion_subtree = self.parse_snippet(llcRegion_code, {'region' : region}, name = 'General Region', show = True).ext[-1].body
        
        
        # llcRegion_subtree.block_items[-2] = llcFor_node.stmt
        ReplaceTool(new_node = llcRegion_node.stmt, old_node = llcRegion_subtree.block_items[-2]).apply(llcRegion_subtree, 'block_items')
        
        
        ##################### Loop substitution 
        # Replace the entire pragma by a CompoundStatement with all the new statements
        # Note: The parent of Parallel is always a Pragma node
        #~ from Tools.Debug import DotDebugTool
        #~ DotDebugTool(llcFor_node).apply(llcFor_node.parent.parent)
        ReplaceTool(new_node = llcRegion_subtree, old_node = llcRegion_node.parent).apply(llcRegion_node.parent.parent, 'block_items')


