# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

# -*- coding: utf-8 -*-

'''
 :author: Rayco Abad-Martín <rayco.abad@gmail.com>
 :author: Elena Marrero-Méndez <elemarez@gmail.com>
'''

from Backends.Common.Visitors.GenericVisitors import * 
from Backends.MPI.Mutators.Common import AbstractMPIMutator
from Frontend.Shortcuts import getCurrentLanguageAst
from Frontend.SymbolTable import updateSymbolTableFromNode
from Tools.Tree import InsertTool, RemoveTool, ReplaceTool
import config

c_ast = getCurrentLanguageAst()

class MPI_MainCreatorFilter(GenericFilterVisitor):
    """ Returns the first node with the given attribute
    """
    def __init__(self):
        def condition(node):
            if type(node) == c_ast.FuncDef:
               if node.decl.name == 'main':
                  return True
            return False
        super(MPI_MainCreatorFilter, self).__init__(condition_func = condition)

class MPI_MainCreator(AbstractMPIMutator):
    """ Apply MPI Main """ 
    def __init__(self, *args, **kwargs):
        super(MPI_MainCreator, self).__init__()

    def filter(self, ast):
        """ Filter """
        raise NotImplemented

    def filter_iterator(self, ast):
        """ Filter iterator """
        return NotImplemented

    def fast_filter(self, ast):
        """ Fast filter , looking for binary expressions """
        return MPI_MainCreatorFilter().dfs_iter(ast)

    def mutatorFunction(self, ast):
        """ Mutator code 
            :param ast: node returned by the filter
        """
        ast.decl.name = 'llc_main'
        template_code = open(config.WORKDIR + '/Backends/MPI/Templates/main.tmpl', 'r').read()
        main = self.parse_snippet(template_code, {}, name = 'template main', show = False)
        InsertTool(subtree = main, position = "end").apply(ast.parent, 'ext')
        updateSymbolTableFromNode(ast)
        return ast
