# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

"""
    Author: Ruyman Reyes 
    Author: Juan Jose Fumero 
"""

from Frontend.C99 import c99_lexer 
from ply.lex import TOKEN

class AccLexer(object):
   
    @staticmethod
    def get_reserved():
        reserved = {
                'pragma'     : 'PRAGMA',
                'name'       : 'ACC_NAME', # ULL Addon to Acc 
		# ################################################
		#                OpenACC
		# ################################################
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #           Standard OpenACC V 1.0  
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                'acc'        : 'ACC_ACC',
                'host'       : 'ACCHOST',
                'device'     : 'ACCDEVICE',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #           Region clauses 
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                'parallel'   : 'ACC_PARALLEL',
                'kernels'    : 'ACC_KERNELS',
                'data'       : 'ACC_DATA',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #           Loop  clauses 
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                'loop'       : 'ACC_LOOP',
                'collapse'   : 'ACC_COLLAPSE',
                'num_gangs'  : 'ACC_NUM_GANGS',
                'gang'       : 'ACC_GANG',
                'num_workers': 'ACC_NUM_WORKERS',
                'worker'     : 'ACC_WORKER',
                'vector_length': 'ACC_VECTOR_LENGTH',
                'vector'     : 'ACC_VECTOR',
     	        'independent': 'ACC_INDEPENDENT',
                'if'         : 'ACC_IF',
                'seq'        : 'ACC_SEQ',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #           Data movement clauses 
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                'host_data'  : 'ACC_HOSTDATA',
                'use_device' : 'ACC_USEDEVICE',
                'update'     : 'ACCUPDATE',
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        #            Syncronization
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                'async'      : 'ACC_ASYNC',
                'wait'       : 'ACC_WAIT',

		# ################################################
        #                llc v2 syntax
        # ################################################
				'llc'   	: 'LLC',
				# llc yacf-frangollo
#				'context'	: 'REGION',
				'task'  	: 'TASK',
#				'name'		: 'NAME',
				'concycle' 	: 'CONCYCLE',
#				'in' 		: 'IN',
#				'out'		: 'OUT',
#				'inout'		: 'INOUT',
			# ----- Loop Transformations ------
				'nested_for': 'NESTED_FOR',
				'interchange': 'INTERCHANGE',
				'unroll'	: 'UNROLL',
				'unswitch'	: 'UNSWITCH',
				'tiling'	: 'TILING',
				'region'    : 'REGION',

		# ################################################
        #                OpenMP
        # ################################################
    		# -------- omp ---------
	    	'omp'		: 'OMP_OMP',
#		    'parallel'	: 'OMP_PARALLEL',
    		'sections'	: 'OMP_SECTIONS',
#			'name'		: 'OMP_NAME',
	    	# ------ clauses --------
#   		'shared'	: 'OMPSHARED',
	    	'private'	: 'OMP_PRIVATE',
#    		'reduction'	: 'OMP_REDUCTION',
    		'nowait'	: 'OMP_NOWAIT',
    		'threadprivate' : 'OMP_THREADPRIVATE',
#    		'copyin'	: 'OMP_COPYIN',
    		'schedule'	: 'OMP_SCHEDULE',
    		'omp_schedule_kind' : 'OMP_SCHEDULE_KIND',
    		'master' 	: 'OMP_MASTER',
    		'critical'	: 'OMP_CRITICAL',
    		'atomic'	: 'OMP_ATOMIC',
    		'barrier'	: 'OMP_BARRIER',
    		'flush'		: 'OMP_FLUSH',
    		'single'	: 'OMP_SINGLE',
#		'firstprivate'  : 'OMP_FIRSTPRIVATE',
#    		'collapse'	: 'OMP_COLLAPSE',
    		'none'		: 'OMP_NONE',
    		'result'	: 'OMP_RESULT',
    		'nc_result'     : 'OMP_NCRESULT',
    		'num_threads'   : 'OMP_NUM_THREADS',
#		'lastprivate'   : 'OMP_LASTPRIVATE',
		# ------- new clauses --------
		# From  "A proposal to extend the OpenMP Tasking Model for Heterogeneous Architectures" (2009)
#    		'device'	: 'DEVICE',
    		'target'	: 'TARGET',
#    		'copy_in'	: 'OMP_COPY_IN',
 #   		'copy_out'	: 'OMP_COPY_OUT',
    		# ----- CELL syntax ---------
    		'cell'		: 'CELL',
    		# --------- CUDA -----------
    		'cuda'		: 'CUDA',
    		# --------- FPGA -----------
    		'fpga'		: 'FPGA',
    		# --------- MPI -----------
    		'mpi'		: 'MPI',
    		'hybrid'	: 'HYBRID',
    		'smp' 		: 'SMP',
         # -------- QUICK HACK FOR CUDA/GMAC, CLEAN ME ------
          'default'     : 'OMP_DEFAULT',
          'for'		    : 'OMP_FOR',
#          'if'		    : 'OMP_IF',

            }
        return reserved
  
    @staticmethod
    def get_tokens():
            tokens = list(AccLexer.get_reserved().values()) + [
             'ACCIGNORE', 'ACCCOPYIN', 'ACCCOPYOUT', 'ACCCOPY', 'ACCPRIVATE', 
             'ACCFIRSTPRIVATE', 'ACCCREATE', 'ACCPCOPYIN', 'ACCPCOPYOUT', 'ACCPCOPY', 
             'ACCPCREATE', 'ACCREDUCTION', 'OMPSHARED', 'OMPLASTPRIVATE', 'ACCPRESENT',
            ]
            return tokens

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    #            Lexical states
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def t_OMPSHARED(self, t):
        r'shared'
        t.lexer.push_state('OMPSHARED')
        return t

    def t_OMPLASTPRIVATE(self, t):
        r'lastprivate'
        t.lexer.push_state('OMPLASTPRIVATE')
        return t

    def t_ACCCOPYOUT(self, t):
        r'copyout'
        t.lexer.push_state('ACCCOPYOUT')
        return t

    def t_ACCCOPYIN(self, t):
        r'copyin'
        t.lexer.push_state('ACCCOPYIN')
        return t

    def t_ACCCOPY(self, t):
        r'copy'
        t.lexer.push_state('ACCCOPY')
        return t

    def t_ACCPRIVATE(self, t):
        r'private'
        t.lexer.push_state('ACCPRIVATE')
        return t

    def t_ACCFIRSTPRIVATE(self, t):
        r'firstprivate'
        t.lexer.push_state('ACCFIRSTPRIVATE')
        return t

    def t_ACCCREATE(self, t):
        r'create'
        t.lexer.push_state('ACCCREATE')
        return t
    
    def t_ACCPCOPYIN(self, t):
        r'pcopyin'
        t.lexer.push_state('ACCPCOPYIN')
        return t

    def t_ACCPCOPYOUT(self, t):
        r'pcopyout'
        t.lexer.push_state('ACCPCOPYOUT')
        return t

    def t_ACCPCOPY(self, t):
        r'pcopy'
        t.lexer.push_state('ACCPCOPY')
        return t

    def t_ACCPRESENT(self, t):
        r'present'
        t.lexer.push_state('ACCPRESENT')
        return t


    def t_ACCPCREATE(self, t):
        r'pcreate'
        t.lexer.push_state('ACCPCREATE')
        return t
 
#    def t_ACCUPDATE(self, t):
#        r'update'
#        print " ACC UPDATE token " + str(t)
#        return t
#
#    def t_ACCHOST(self, t):
#        r'host'
#        print " ACCHOST token outside "
#        return t
#    def t_ACCDEVICE(self, t):
#        r'device'
#        print " ACCDEVICE token outside "
#        return t
#      
    def t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_ID(self, t):
        r'[a-zA-Z_][0-9a-zA-Z_]*'
        #t.type = AccLexer.get_reserved().get(t.value, 'ID') 
        t.type = 'ID'
        return t

    ### States Tokens for [ ] : { } + - * / ,
    t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_LBRACKET = '\['
    t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_RBRACKET = '\]'
    t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_COLON    = ':'
    t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_LBRACE   = '\{'
    t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_RBRACE   = '\}'
    t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_TIMES    = '\*'
    t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_PLUS     = '\+'
    t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_MINUS    = '\-'
    t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_DIVIDE   = '\/'
    t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_COMMA    = ','

    ## TODO: Review this regular expression - see c99 standard
    def t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCOPY_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_INT_CONST_DEC(self, t):
        r'\d+'
        return t

    def t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCOPY_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_RPAREN(self, t):
        r'\)'
        t.lexer.pop_state()
        return t
 
#
#    def t_ACCUPDATE_ACCHOST(self, t):
#        r'host'
#        print " TOKEN ACCHOST IN ACCUPDATE STATE "
#        return t
#
#    def t_ACCUPDATE_ACCDEVICE(self, t):
#        r'device'
#        print " TOKEN ACCDEVICE IN ACCUPDATE STATE "
#        return t
#C
       
    def t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCOPY_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_LPAREN(self, t):
        r'\('
        return t

    t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCOPY_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_ignore = ' \t'

    def t_ACCCOPYIN_ACCCOPYOUT_ACCCOPY_ACCPRIVATE_ACCFIRSTPRIVATE_ACCCREATE_ACCPCOPY_ACCPCOPYIN_ACCPCOPYOUT_ACCPCOPY_ACCPCREATE_ACCREDUCTION_OMPSHARED_OMPLASTPRIVATE_ACCPRESENT_error(self, t):
        print 'Lexical error in state "' + str(t.value[0]) + '" in line ' + str(t.lineno)
        t.lexer.skip(1)

    def t_directive_ID(self, t):
        r'[a-zA-Z_][0-9a-zA-Z_]*'
        t.type = AccLexer.get_reserved().get(t.value, 'ID')
        if (t.value == 'copyin'):
            t.type = 'ACCCOPYIN'
            t.lexer.push_state('ACCCOPYIN')
        elif (t.value == 'copyout'):
            t.type = 'ACCCOPYOUT'
            t.lexer.push_state('ACCCOPYOUT')
        elif (t.value == 'copy'):
            t.type = 'ACCCOPY'
            t.lexer.push_state('ACCCOPY')
        elif (t.value == 'private'):
            t.type = 'ACCPRIVATE'
            t.lexer.push_state('ACCPRIVATE') 
        elif (t.value == 'firstprivate'):
            t.type = 'ACCFIRSTPRIVATE'
            t.lexer.push_state('ACCFIRSTPRIVATE')
        elif (t.value == 'create'):
            t.type = 'ACCCREATE'
            t.lexer.push_state('ACCCREATE')
        elif (t.value == 'pcopyin'):
            t.type = 'ACCPCOPYIN'
            t.lexer.push_state('ACCPCOPYIN')
        elif (t.value == 'pcopyout'):
            t.type = 'ACCPCOPYOUT'
            t.lexer.push_state('ACCPCOPYOUT')
        elif (t.value == 'pcopy'):
            t.type = 'ACCPCOPY'
            t.lexer.push_state('ACCPCOPY')
	elif (t.value == 'present'):
	    t.type = 'ACCPRESENT'
	    t.lexer.push_state("ACCPRESENT")
        elif (t.value == 'pcreate'):
            t.type = 'ACCPCREATE'
            t.lexer.push_state('ACCPCREATE')
        #elif (t.value == 'update'):
        #    t.type = 'ACCUPDATE'
        #    t.lexer.push_state('ACCUPDATE')
        elif (t.value == 'reduction'):
            t.type = 'ACCREDUCTION'
            t.lexer.push_state('ACCREDUCTION')
        elif (t.value == 'shared'):
            t.type = 'OMPSHARED'
            t.lexer.push_state('OMPSHARED')
        elif (t.value == 'lastprivate'):
            t.type = 'OMPLASTPRIVATE'
            t.lexer.push_state('OMPLASTPRIVATE')
        return t

    def t_ACCIGNORE(self, t):
        r'_yacf_ignore_\n[^$]+_yacf_ignore_'
        t.type = 'ACCIGNORE'
        return t

    @staticmethod
    def get_states():
        states = (
		    ('ACCCOPYIN', 'exclusive'),
		    ('ACCCOPYOUT', 'exclusive'),
		    ('ACCCOPY', 'exclusive'),
		    ('ACCPRIVATE', 'exclusive'),
		    ('ACCFIRSTPRIVATE', 'exclusive'),
		    ('ACCCREATE', 'exclusive'),
		    ('ACCPCOPYIN', 'exclusive'),
		    ('ACCPCOPYOUT', 'exclusive'),
		    ('ACCPCOPY', 'exclusive'),
		    ('ACCPRESENT', 'exclusive'),
		    ('ACCPCREATE', 'exclusive'),
		   # ('ACCUPDATE', 'exclusive'),
		    ('ACCREDUCTION', 'exclusive'),
            ('OMPSHARED', 'exclusive'),
            ('OMPLASTPRIVATE', 'exclusive'),
	      )
        return states

