# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

#################################
# Ruyk
# Juan J.
# ------------------------------
# Adding support for OpenAcc 1.0
# Adding support for OpenMP  3.0
#################################

from Frontend.Acc.plyparser import PLYParser, Coord, ParseError
from Frontend.Acc.acc_lexer import AccLexer
from Frontend.Acc import acc_ast
# TODO: THIS NODE SHOULD NOT USE PRAGMA node at all
from Frontend.C99 import c99_ast as c_ast, c99_ast


class AccParser(PLYParser):
    # =============================
    def p_identifier_parenthesis_list(self, p):
       """ identifier_parenthesis_list : LPAREN identifier_list RPAREN """
       p[0] = p[2]
    def p_constant_parenthesis(self, p):
       """ constant_parenthesis : LPAREN constant_expression RPAREN """
       p[0] = p[2]

    # ==================================================	
    #            OpenACC Grammar
    # ==================================================	
    def p_pp_directive_2(self, p):
       """ pp_directive  : PPHASH acc_pragma_directive
       """
       p[0] = p[2]


    def p_pp_directive_1(self, p):
       """ pp_directive  : PPHASH omp_pragma_directive 
       """
       p[0] = p[2]


    def p_acc_pragma_directive_2(self, p):
       """ acc_pragma_directive : PRAGMA ACC_ACC acc_directive"""
       #~print " ACC DIRECTIVE "
       p[0] = [c_ast.Pragma(
            name='acc', 
            stmt = p[3],
            coord=self._coord(p.lineno(1)))]


    def p_external_declaration_5(self, p):
        """ external_declaration : ACCIGNORE """
        print 'En accignore ******************'
        p[0] =  [acc_ast.AccIgnore(name = "AccIgnore", code = p[1])]


    def p_statement_2(self, p):
        """ statement : ACCIGNORE
        """
        p[0] =  [acc_ast.AccIgnore(name = "AccIgnore", code = p[1])]

       

    # --acc yacf-frangollo ------------------------------
    def p_acc_region_type(self, p):
        """ acc_region_type : ACC_PARALLEL 
                            | ACC_KERNELS
                            | ACC_DATA
        """
        p[0] = p[1]


    def p_acc_directive_9(self, p):
        """ acc_directive : ACC_HOSTDATA ACC_USEDEVICE identifier_parenthesis_list statement
        """
        p[0] = acc_ast.AccHostData(name = p[1], varlist = p[3].params, stmt = c99_ast.Compound(block_items = [p[5],]) )

    def p_acc_directive_8(self, p):
        """ acc_directive : acc_region_type acc_clause_list statement 
        """
        stmt = p[3]
        # If the stmt contains a Pragma node, they came within a List from the Pycparser. 
        #   This breaks the transformations of the Frangollo backend if not properly handled
        if (type(stmt) == type([])):
           stmt = stmt[0]

        if p[1].lower() == 'parallel':
           p[0] = acc_ast.AccParallel(name = p[1], stmt = stmt, clauses = p[2] or [], coord=self._coord(p.lineno(1)))     
        elif p[1].lower() == 'kernels':
           p[0] = acc_ast.AccKernels(name = p[1], stmt = stmt, clauses = p[2] or [], coord=self._coord(p.lineno(1)))     
        elif p[1].lower() == 'data':
           p[0] = acc_ast.AccData(name = p[1], stmt = stmt, clauses = p[2] or [], coord=self._coord(p.lineno(1)))     
        else:
           print p[1]
           raise ParseError()

    def p_acc_directive_81(self, p):
        """ acc_directive : acc_region_type statement 
        """
        stmt = p[2]
        # If the stmt contains a Pragma node, they came within a List from the Pycparser. 
        #   This breaks the transformations of the Frangollo backend
        if (type(stmt) == type([])):
           stmt = stmt[0]
        if p[1].lower() == 'parallel':
           p[0] = acc_ast.AccParallel(name = p[1], stmt = stmt, clauses = list(), coord=self._coord(p.lineno(1)))     
        elif p[1].lower() == 'kernels':
           p[0] = acc_ast.AccKernels(name = p[1], stmt = stmt, clauses = list(), coord=self._coord(p.lineno(1)))     
        elif p[1].lower() == 'data':
           p[0] = acc_ast.AccData(name = p[1], stmt = stmt, clauses = list(), coord=self._coord(p.lineno(1)))     
        else:
           print p[1]
           raise ParseError()


    def p_acc_directive_11(self, p):
        """ acc_directive : workshare_directive """
        p[0] = p[1]

    def p_acc_size_expression(self, p):
        """ acc_size_expression : LBRACKET expression RBRACKET
                                | LBRACKET expression COLON expression RBRACKET
                                | LBRACKET expression RBRACKET LBRACE identifier RBRACE
        """
        if p[3] == ':':
            p[0] = acc_ast.AccSizeExpression(start = p[2], end = p[4], size = None, stride = None)
        else:
            if len(p) == 7: # Has a stride argument
               p[0] = acc_ast.AccSizeExpression(start = p[2], end = None, size = None, stride = p[5])
            elif len(p) > 4:
               p[0] = acc_ast.AccSizeExpression(start = p[2], end = None, size = p[4], stride = None)
            else:
               p[0] = acc_ast.AccSizeExpression(start = p[2], end = None, size = None, stride = None)


    def p_acc_size_expression_list(self, p):
        """ acc_size_expression_list : acc_size_expression  acc_size_expression 
                                     | acc_size_expression 
        """
        if len(p) > 2:
            p[0] = (p[1],p[2])
        else:
            p[0] = p[1]


    def p_acc_complex_identifier_list(self, p):
        """ acc_complex_identifier_list : identifier 
                                        | identifier acc_size_expression_list
                                        | acc_complex_identifier_list COMMA identifier
                                        | acc_complex_identifier_list COMMA identifier acc_size_expression_list
        """
#        acc_id = p[1]
        definitions = {}
        if len(p) == 3:
           definitions[p[1]] = p[2]
           p[0] = acc_ast.AccParamList([p[1]], definitions, p[1].coord)
        elif len(p) == 4:
           p[1].params.append(p[3])
           p[0] = p[1]
        elif len(p) == 5:
           definitions[p[3]] = p[4]
           p[1].params.append(p[3])
           p[1].definitions.update(definitions)
           p[0] = p[1]
        else:
            p[0] = acc_ast.AccParamList([p[1]], definitions, p[1].coord)

    def p_acc_clause_1(self, p):
        """ acc_clause : ACC_NAME LPAREN STRING_LITERAL RPAREN 
                       | acc_complex_clause  LPAREN acc_complex_identifier_list RPAREN 
                       | ACCREDUCTION LPAREN reduction_operator  COLON identifier_list RPAREN 
        """
        tmp = None
        additional = None
        if p[1] == 'name':
            # Remove semicolons fron STRING LITERAL
            tmp = acc_ast.AccName(name = 'NAME', label = p[3][1:-1], coord = self._coord(p.lineno(1)))    
        elif p[1] == 'reduction':
            tmp = acc_ast.AccClause(type = p[3] , name = 'REDUCTION', identifiers = p[5], coord = self._coord(p.lineno(1)))
        else:
            tmp = acc_ast.AccCopy(name = p[1].upper(), identifiers = p[3], present = (p[1][0] == 'p'), coord = self._coord(p.lineno(1)));

        p[0] = []
        p[0].append(tmp)
        if additional:
            p[0].extend(additional)

    def p_acc_clause_4(self, p):
        """ acc_clause : ACC_NUM_WORKERS constant_parenthesis 
                       | ACC_NUM_GANGS constant_parenthesis 
                       | ACC_VECTOR_LENGTH constant_parenthesis 
                       | ACC_GANG constant_parenthesis 
                       | ACC_WORKER constant_parenthesis 
                       | ACC_VECTOR constant_parenthesis 
		                 | ACC_GANG
                       | ACC_WORKER
                       | ACC_VECTOR
                       | ACC_SEQ
        """
        if len(p) == 3:
                # Only valid within the scope of a kernels
        	p[0] = [acc_ast.AccParallelOption(name=p[1], value=p[2])]
	else:
		# only valid within the scope of a parallel
        	p[0] = [acc_ast.AccParallelOption(name=p[1], value=[])]


    def p_acc_complex_clause(self, p):
        """ acc_complex_clause : ACCCOPYIN
                               | ACCCOPYOUT
                               | ACCCOPY
                               | ACCPCOPY
                               | ACCPCOPYIN
                               | ACCPCOPYOUT
                               | ACCCREATE
                               | ACCPCREATE
			       | ACCPRESENT
   			   | ACCPRIVATE
	                           | ACCFIRSTPRIVATE
        """
        p[0] = p[1]
       

    def p_acc_iteration_statement(self, p):
        """ acc_iteration_statement : PPHASH PRAGMA ACC_ACC acc_directive 
                                    | iteration_statement
        """
        if len(p) == 5:
           p[0] = p[4]
        else:
           p[0] = p[1]


    def p_workshare_directive_1(self, p):
       """ workshare_directive : ACC_LOOP acc_clause_list acc_iteration_statement """
       # Start of a parallel region within a for
       p[0] = acc_ast.AccLoop(name=p[1].lower(), clauses = p[2], stmt = p[3], coord = self._coord(p.lineno(1)))

    def p_workshare_directive_2(self, p):
       """ workshare_directive : ACC_LOOP  acc_iteration_statement """
       # Start of a parallel region within a for
       p[0] = acc_ast.AccLoop(name=p[1].lower(), clauses = [], stmt = p[2], coord = self._coord(p.lineno(1)))


    def p_acc_directive_200(self, p):
       """ acc_directive : acc_region_type ACC_LOOP acc_clause_list acc_iteration_statement
                     | acc_region_type ACC_LOOP acc_iteration_statement """
       if p[1].lower() == 'data':
            raise ParseError("#pragma data loop is not allowed")

       clauses = None
       stmt = p[3]
       if (len(p) == 5):
         clauses = p[3]
         stmt = p[4]
       else:
	 clauses = []
         stmt = p[3]

       regionDict = { 'parallel' : acc_ast.AccParallel, 'kernels' : acc_ast.AccKernels}
       non_copy = []
       copy = []
       # TODO: What to do with deviceptr?
       #import pdb
       #pdb.set_trace()
       if p[3] and not stmt == p[3]:
           non_copy = [ c for c in p[3] if not isinstance(c, acc_ast.AccCopy)]
           copy     = [ c for c in p[3] if     isinstance(c, acc_ast.AccCopy)]

       acc_loop = acc_ast.AccLoop(name=p[2].lower(), clauses = non_copy, stmt = stmt, coord = self._coord(p.lineno(1)))
       pragma = c_ast.Pragma(
            name='acc', 
            stmt = acc_loop,
            coord=self._coord(p.lineno(1)))
       p[0] = regionDict[p[1].lower()](name = p[1], stmt = c99_ast.Compound(block_items = [pragma, ]), clauses = copy , coord=self._coord(p.lineno(1)))     

   # def p_acc_directive_300(self,p):
   #    """ acc_directive : acc_region_type compound_statement """
   #    # Start of a parallel region
   #    p[0] = acc_ast.AccParallel(name='parallel', clauses = None, stmt = p[2], coord = self._coord(p.lineno(1)))

    def p_acc_directive_12(self, p):
       """ acc_directive : ACCUPDATE acc_update_destination LPAREN llc_complex_identifier_list RPAREN 
                     | ACC_WAIT LPAREN constant_expression RPAREN
                     | ACC_WAIT 
       """
       #~print " ACC UPDATE OR WAIT "
       if p[1].lower() == 'update':
           p[0] = acc_ast.AccUpdate(name = "update", destination = p[2], identifiers = p[4].params, coord = self._coord(p.lineno(1)))
       else:
           if len(p) > 2:
               p[0] = acc_ast.AccWait(name = "wait", value  = p[3], coord = self._coord(p.lineno(1)))
           else:
               p[0] = acc_ast.AccWait(name = "wait", value  = None, coord = self._coord(p.lineno(1)))



    def p_acc_update_destination(self, p):
       """ acc_update_destination : ACCHOST
                                  | ACCDEVICE
       """
       #~print " ACC UPDATE DESTINATION "
       p[0] = p[1].lower();


    def p_acc_clause_list(self, p):
       """ acc_clause_list : acc_clause_list acc_clause
			   | acc_clause_list COMMA acc_clause
                           | acc_clause 
       """
       if len(p) == 4:
           p[0] = p[1] + p[3]
       else:
           p[0] = p[1] if len(p) == 2 else p[1] + p[2]

#    def p_acc_clause_3(self, p):
#       """ acc_clause : PRIVATE LPAREN identifier_list RPAREN 
#                      | FIRSTPRIVATE LPAREN identifier_list RPAREN
#       """
#       p[0] = [acc_ast.AccClause(type = 'memory', name = str(p[1]).upper(), identifiers = p[3], coord = self._coord(p.lineno(1)))]
#
    def p_acc_clause_6(self, p):
       """ acc_clause : ACC_COLLAPSE LPAREN INT_CONST_DEC RPAREN """
       value = c_ast.Constant(value = p[3], type='int', coord = self._coord(p.lineno(1)))
       p[0] = [acc_ast.AccClause(type = str(p[1]) , name = str(p[1]).upper(), identifiers = value, coord = self._coord(p.lineno(1)))]

    def p_acc_clause_7(self, p):
        """ acc_clause : ACC_IF LPAREN expression RPAREN
        """
        p[0] = [acc_ast.AccClause(type = 'IF', name = str(p[1]).upper(), identifiers = p[3], coord = self._coord(p.lineno(1)))]

    def p_acc_clause_8(self, p):
        """ acc_clause : ACC_ASYNC
                       | ACC_ASYNC LPAREN constant_expression RPAREN
        """
        val = c_ast.Constant(value = "0", type="int")
        if len(p) > 2:
            val = p[3]
        p[0] = [acc_ast.AccAsync(name = str(p[1]).upper(), value = val, coord = self._coord(p.lineno(1)))]

    def p_acc_clause_9(self, p):
        """ acc_clause : ACC_INDEPENDENT
        """
	# Nothing to do here
	p[0] = []



    def p_reduction_operator(self, p):
       """ reduction_operator  : PLUS 
                               | MINUS
                               | TIMES
			       | ID  """
       p[0] = p[1]


    ###############################
    # End of OpenAcc Grammar
    ###############################

    # ==================================================	
    #            OpenMP Grammar
    # ==================================================	

    def p_omp_pragma_directive_1(self, p):
       """ omp_pragma_directive : PRAGMA OMP_OMP omp_directive"""
       p[0] = [c_ast.Pragma(
            name='omp', 
            stmt = p[3],
            coord=self._coord(p.lineno(1)))]

    def p_omp_pragma_directive_2(self, p):
       """ omp_pragma_directive : PRAGMA LLC llc_directive"""
       p[0] = [c_ast.Pragma(
            name='llc', 
            stmt = p[3],
            coord=self._coord(p.lineno(1)))]

    def p_llc_directive_1(self, p):
       """ llc_directive : NESTED_FOR PPHASH PRAGMA OMP_OMP omp_workshare_directive"""
       p[0] = acc_ast.llcNestedFor(name = 'NESTED FOR', loop = p[5], coord=self._coord(p.lineno(1)))

    def p_llc_directive_2(self, p):
       """ llc_directive : INTERCHANGE iteration_statement"""
       p[0] = acc_ast.llcInterchange(name = 'INTERCHANGE', loop = p[2], coord=self._coord(p.lineno(1)))

    def p_llc_directive_3(self, p):
       """ llc_directive : OMP_RESULT LPAREN llc_list RPAREN iteration_statement """
       p[0] = acc_ast.llcResult(name = 'RESULT', result_list = p[3], stmt = p[5], coord=self._coord(p.lineno(1)))

    def p_llc_directive_4(self, p):
       """ llc_directive : UNROLL LPAREN INT_CONST_DEC RPAREN iteration_statement"""
       p[0] = acc_ast.llcUnroll(name = 'UNROLL', factor = p[3], loop = p[5], coord=self._coord(p.lineno(1)))

    def p_llc_directive_5(self, p):
       """ llc_directive : UNSWITCH iteration_statement"""
       p[0] = acc_ast.llcUnswitch(name = 'UNSWITCH', loop = p[2], coord=self._coord(p.lineno(1)))
    
    def p_llc_directive_6(self, p):
       """ llc_directive : TILING LPAREN INT_CONST_DEC RPAREN iteration_statement"""
       p[0] = acc_ast.llcTiling(name = 'TILING', size = p[3], loop = p[5], coord=self._coord(p.lineno(1)))


    # --llc yacf-frangollo ------------------------------
    def p_llc_directive_8(self, p):
        """ llc_directive : REGION llc_clause statement """
        p[0] = acc_ast.llcRegion(name = 'REGION', stmt = p[3], clauses = p[2] or [], coord=self._coord(p.lineno(1)))     

    def p_llc_directive_9(self, p):
        """ llc_directive : TASK llc_clause postfix_expression """
        p[0] = acc_ast.llcTask(name = 'TASK', coord=self._coord(p.lineno(1)))     

    def p_llc_directive_10(self, p):
        """ llc_directive : CONCYCLE """
        p[0] = acc_ast.llcConcycle(name = 'CONCYCLE', coord=self._coord(p.lineno(1)))     

    def p_llc_directive_11(self, p):
        """ llc_directive : omp_workshare_directive """
        p[0] = p[1]

    def p_llc_size_expression(self, p):
        """ llc_size_expression : LBRACKET expression RBRACKET
                                | LBRACKET expression SEMI expression RBRACKET
                                | LBRACKET expression COLON expression RBRACKET
        """
        if p[3] == ':':
            p[0] = acc_ast.llcSizeExpression(start = p[2], end = p[4], size = None)
        else:
            if len(p) > 4:
               p[0] = acc_ast.llcSizeExpression(start = p[2], end = None, size = p[4])
            else:
               p[0] = acc_ast.llcSizeExpression(start = p[2], end = None, size = None)



    def p_llc_complex_identifier_list(self, p):
        """ llc_complex_identifier_list : identifier 
                                        | identifier llc_size_expression
                                        | llc_complex_identifier_list COMMA identifier
                                        | llc_complex_identifier_list COMMA identifier llc_size_expression
        """
        definitions = {}
        if len(p) == 3:
           definitions[p[1]] = p[2]
           p[0] = acc_ast.llcParamList([p[1]], definitions, p[1].coord)
        elif len(p) == 4:
           p[1].params.append(p[3])
           p[0] = p[1]
        elif len(p) == 5:
           definitions[p[3]] = p[4]
           p[1].params.append(p[3])
           p[1].definitions.update(definitions)
           p[0] = p[1]
        else:
            p[0] = acc_ast.llcParamList([p[1]], definitions, p[1].coord)


    def p_llc_clause_1(self, p):
        """ llc_clause : ACC_NAME LPAREN STRING_LITERAL RPAREN llc_clause 
                       | ACCCOPYIN  LPAREN llc_complex_identifier_list RPAREN llc_clause
                       | ACCCOPYOUT LPAREN llc_complex_identifier_list RPAREN llc_clause
        """
        tmp = None
        if p[1] == 'name':
            # Remove semicolons fron STRING LITERAL
            tmp = acc_ast.llcName(name = 'NAME', label = p[3][1:-1], coord = self._coord(p.lineno(1)))    
        if p[1] == 'copy_in':
            tmp = acc_ast.llcCopyin(name = 'COPY_IN', identifiers = p[3], coord = self._coord(p.lineno(1)))
        if p[1] == 'copy_out':
            tmp = acc_ast.llcCopyin(name = 'COPY_OUT', identifiers = p[3], coord = self._coord(p.lineno(1)))
        if p[1] == 'in':
            tmp = acc_ast.llcCopyin(name = 'IN', identifiers = p[3], coord = self._coord(p.lineno(1)))
        if p[1] == 'out':
            tmp = acc_ast.llcCopyin(name = 'OUT', identifiers = p[3], coord = self._coord(p.lineno(1)))
        if p[1] == 'inout':
            tmp = acc_ast.llcCopyin(name = 'INOUT', identifiers = p[3], coord = self._coord(p.lineno(1)))
        
        
        p[0] = []
        p[0].append(tmp)
        if p[4]:
            p[0].extend(p[5])

        
    def p_llc_clause_2(self, p):
        """ llc_clause : empty """
        p[0] = []


    def p_llc_list(self, p):
        """ llc_list : AND direct_declarator COMMA INT_CONST_DEC llc_lst """
        p[0] = [acc_ast.llcResultItem(declaration = p[2], size = c_ast.Constant(type = 'int', value = p[4]))] 
        if not p[5] is None:
           p[0].extend(p[5])

    def p_llc_lst_1(self, p):
        """ llc_lst : COMMA llc_list """
        p[0] = p[2]

    def p_llc_lst_2(self, p):
        """ llc_lst : empty """
        p[0] = None

    def p_llc_iteration_statement(self, p):
        """ llc_iteration_statement : PPHASH PRAGMA LLC llc_directive 
                                    | iteration_statement
        """
        if len(p) == 5:
           p[0] = p[4]
        else:
           p[0] = p[1]

    def p_statement_3(self, p):
        """ statement : llc_nc_result """
        p[0] = p[1]

    def p_llc_nc_result(self, p):
       """ llc_nc_result : PPHASH PRAGMA LLC OMP_NCRESULT LPAREN llc_nc_list RPAREN statement """
       p[0] = acc_ast.llcNcResult(name = 'NCRESULT', result_list = p[6], stmt = p[8], coord=self._coord(p.lineno(1)))

    def p_llc_nc_list_1(self, p):
        """ llc_nc_list : AND direct_declarator COMMA INT_CONST_DEC COMMA AND direct_declarator  llc_nc_lst """
        p[0] = [acc_ast.llcNcResultItem(declaration = p[2], size = c_ast.Constant(type = 'int', value = p[4])),]
        if not p[8] is None:
            p[0].extend(p[8])

    def p_llc_nc_lst_2(self, p):
        """ llc_nc_lst : COMMA llc_nc_list """
        p[0] = p[2]

    def p_llc_nc_lst_3(self, p):
        """ llc_nc_lst : empty """
        p[0] = None

    def p_omp_workshare_directive_1(self, p):
       """ omp_workshare_directive : OMP_FOR omp_clause_list llc_iteration_statement """
       # Start of a parallel region within a for
       p[0] = acc_ast.OmpFor(name='for', clauses = p[2], stmt = p[3], coord = self._coord(p.lineno(1)))

    def p_omp_workshare_directive_11(self, p):
       """ omp_workshare_directive : OMP_FOR not_omp_clause_list llc_iteration_statement """
       # Start of a parallel region within a for
       p[0] = acc_ast.OmpFor(name='for', clauses = p[2], stmt = p[3], coord = self._coord(p.lineno(1)))


    def p_omp_workshare_directive_2(self, p):
       """ omp_workshare_directive : OMP_FOR  llc_iteration_statement """
     # Start of a parallel region within a for 
       p[0] = acc_ast.OmpFor(name='for', clauses = [], stmt = p[2], coord = self._coord(p.lineno(1)))

    
    def p_omp_workshare_directive_3(self, p):
       """ omp_workshare_directive : OMP_SECTIONS omp_clause_list compound_statement """
       # Start of a parallel region within a for
       p[0] = p[3]


    def p_omp_directive_1(self, p):
       """ omp_directive : ACC_PARALLEL omp_clause_list compound_statement
       """
       p[0] =  acc_ast.OmpParallel(name='parallel', clauses = p[2], stmt = p[3], coord = self._coord(p.lineno(1)))


    def p_omp_directive_2(self, p):
       """ omp_directive : ACC_PARALLEL OMP_FOR omp_clause_list llc_iteration_statement
	             | ACC_PARALLEL OMP_FOR llc_iteration_statement """
       # Start of a parallel region within a for
       if (len(p) == 5):
          p[0] = acc_ast.OmpParallelFor(name='parallel for', clauses = p[3], stmt = p[4], coord = self._coord(p.lineno(1)))
       else:
          p[0] = acc_ast.OmpParallelFor(name='parallel for', clauses = None, stmt = p[3], coord = self._coord(p.lineno(1)))

    def p_omp_directive_3(self,p):
       """ omp_directive : ACC_PARALLEL compound_statement """
       # Start of a parallel region
       p[0] = acc_ast.OmpParallel(name='parallel', clauses = None, stmt = p[2], coord = self._coord(p.lineno(1)))

    def p_omp_directive_4(self, p):
       """ omp_directive : omp_workshare_directive """
       # We are on a parallel region
       p[0] = p[1]


    def p_omp_directive_5(self, p):
       """ omp_directive : OMP_THREADPRIVATE identifier_parenthesis_list """ 
       p[0] = acc_ast.OmpThreadPrivate(name = 'threadprivate', identifiers = p[2], coord = self._coord(p.lineno(1)))


    def p_omp_directive_6(self, p):
       """ omp_directive : omp_sync_directives omp_clause_list 
                         | omp_sync_directives
       """
       if (len(p) == 3):
          p[0] = acc_ast.OmpSyncronization(name = p[1], clauses = p[2], identifiers = None, coord = self._coord(p.lineno(1)))
       elif (len(p) == 2): 
          p[0] = acc_ast.OmpSyncronization(name = p[1], clauses = [], identifiers = None, coord = self._coord(p.lineno(1)))

    def p_omp_directive_7(self, p):
       """ omp_directive : TARGET ACCDEVICE LPAREN device_list RPAREN omp_clause_list
                         | TARGET ACCDEVICE LPAREN device_list RPAREN 
       """
       p[0] = acc_ast.OmpTargetDevice(name = "device", device = p[4], clauses = p[6] if len(p) > 6 else [], coord = self._coord(p.lineno(1)))

    def p_omp_directive_8(self, p):
       """ omp_directive : OMP_FLUSH  identifier_parenthesis_list 
       """
       p[0] = acc_ast.OmpFlush(name = "flush", identifiers = p[2], coord = self._coord(p.lineno(1)))


    def p_omp_directive_9(self, p):
       """ omp_directive : ACC_PARALLEL OMP_FOR omp_clause_list llc_iteration_statement pp_directive"""
      # Start of a parallel region within a for
       p[0] = acc_ast.OmpParallelFor(name='parallel for', clauses = p[3], stmt = p[4], coord = self._coord(p.lineno(1)))

    def p_omp_directive_10(self, p):
       """ omp_directive : ACC_PARALLEL OMP_FOR not_omp_clause_list llc_iteration_statement pp_directive"""
      # Start of a parallel region within a for
       p[0] = acc_ast.OmpParallelFor(name='parallel for', clauses = p[3], stmt = p[4], coord = self._coord(p.lineno(1)))


    def p_omp_sync_directives(self,p):
       """ omp_sync_directives : OMP_CRITICAL
                               | OMP_MASTER 
                               | OMP_ATOMIC
                               | OMP_FLUSH
                               | OMP_BARRIER
                               | OMP_SINGLE
       """
       p[0] = p[1]

    def p_omp_clause_list(self, p):
       """ omp_clause_list : omp_clause omp_clause_list
                           | omp_clause """
       if len(p) == 3:
            p[0] = p[1] + p[2]
       else:
            p[0] = p[1]

    def p_not_omp_clause_list(self, p):
        """ not_omp_clause_list : empty """
        p[0] = p[1]

    def p_omp_clause_1(self, p):
       """ omp_clause : OMPSHARED identifier_parenthesis_list 
                      | OMP_PRIVATE identifier_parenthesis_list
                      | ACCFIRSTPRIVATE identifier_parenthesis_list
                      | OMPLASTPRIVATE identifier_parenthesis_list
                      | OMP_NUM_THREADS constant_parenthesis
                      | ACC_IF constant_parenthesis
       """
       p[0] = [acc_ast.OmpClause(type = 'memory', name = str(p[1]).upper(), identifiers = p[2], coord = self._coord(p.lineno(1)))]


    def p_omp_clause_2(self, p):
       """ omp_clause : ACCREDUCTION LPAREN reduction_operator  COLON identifier_list RPAREN """
       print " ACc reduction clause "
       if len(p[3]) > 1:
		if not p[3] in ['max','min',]:
			raise ParseError(" Invalid reduction operator ");
       p[0] = [acc_ast.OmpClause(type = p[3] , name = 'REDUCTION', identifiers = p[5], coord = self._coord(p.lineno(1)))]

    def p_omp_clause_3(self, p):
       """ omp_clause : OMP_NOWAIT """
       p[0] = [acc_ast.OmpClause(type = str(p[1]) , name = str(p[1]).upper(), identifiers = None, coord = self._coord(p.lineno(1)))]



    def p_omp_clause_4(self, p):
       """ omp_clause : OMP_SCHEDULE LPAREN OMP_SCHEDULE_KIND RPAREN 
                  | OMP_SCHEDULE LPAREN OMP_SCHEDULE_KIND COMMA INT_CONST_DEC RPAREN 
       """
       p[0] = [acc_ast.OmpSchedule(type = str(p[1]) , name = str(p[1]).upper(), chunk_size = p[5] if len(p) > 5 else 0, kind = p[3], coord = self._coord(p.lineno(1)))]

    def p_omp_clause_5(self, p):
       """ omp_clause : ACCCOPYIN identifier_parenthesis_list """
       p[0] = [acc_ast.OmpClause(type = str(p[1]) , name = str(p[1]).upper(), identifiers = p[2], coord = self._coord(p.lineno(1)))]

    def p_omp_clause_6(self, p):
       """ omp_clause : ACCCOPYOUT identifier_parenthesis_list """
       p[0] = [acc_ast.OmpClause(type = str(p[1]) , name = str(p[1]).upper(), identifiers = p[2], coord = self._coord(p.lineno(1)))]


    def p_omp_clause_7(self, p):
       """ omp_clause : ACC_COLLAPSE LPAREN INT_CONST_DEC RPAREN """
       p[0] = [acc_ast.OmpClause(type = str(p[1]) , name = str(p[1]).upper(), identifiers = None, coord = self._coord(p.lineno(1)))]


    def p_omp_clause_8(self, p):
       """ omp_clause : OMP_DEFAULT LPAREN OMPSHARED RPAREN
                      | OMP_DEFAULT LPAREN OMP_NONE RPAREN
       """
       p[0] = [acc_ast.OmpClause(type = p[3] , name = str(p[1]).upper(), identifiers = None, coord = self._coord(p.lineno(1)))]

    def p_omp_clause_9(self, p):
       """ omp_clause : ACC_NAME LPAREN STRING_LITERAL RPAREN """
       p[0] = [acc_ast.llcName(name = 'NAME', label = p[3][1:-1], coord = self._coord(p.lineno(1)))]



    def p_omp_device_list(self, p):
       """ device_list  : CUDA
                        | FPGA
                        | CELL 
                        | MPI
                        | SMP
                        | HYBRID """
       p[0] = p[1]


# End of OpenMP
####################

