# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
    C99 AST: This class inherits from AstToIR
    .. moduleauthor:: Juan José Fumero Alfonso <jjfumero@gmail.com>
'''
# TODO: AstToIR in Common directory

from Frontend.C99 import c99_ast

from Frontend.InternalRepr import *

class NotValidParent(Exception):
    def __init__(self, node):
        self.node = node

    def __str__(self):
        if hasattr(self, 'parent'):
            return "Node " + str(self.node) + " has an invalid parent (" + str(type(self.node.parent)) + ")"
        else:
            return "Node " + str(self.node) + " has no parent attribute (Has been decorated?)"


class C99AstToIR(AstToIR):

    def __init__(self, Writer, ast):
        """ New instance of the class
               Implements a cache to avoid recreating previous trees
        """
        super(C99AstToIR, self).__init__(Writer, ast)
        #try:
        #    current = kwargs['ast']
        #except KeyError:
        #    raise NotImplemented
        current = ast
        if not current:
            # ast must be a named parameter
            raise NotImplemented
        
        if type(current) != c_ast.FileAST:
            # FIXME: Not sure this will work properly. 
            #      When the node is new, the parent link has not been called
            #      so this may be broken...
            act_ast = current
            while act_ast != c_ast.FileAST:
                act_ast = act_ast.parent
                if act_ast == None:
                    return None
                #if act_ast == None:
                #    raise NotImplemented
            current = act_ast
        # If not already stored, create the instance
        #if not new._instanceCache.has_key(id(current)):
        #    new._instanceCache[id(current)] = super(AstToIR, new).__new__(new, *args, **kwargs)
        #    print "Cache hit " + str(id(current))
        #return  new._instanceCache[id(current)]
        ast = current

    


