# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

"""Tools to manage the internal representation of the code


.. moduleauthor:: Ruyman Reyes Castro <rreyes@ull.es>

"""

#from C import  c_ast
#from C99 import  c_ast

import cStringIO

import copy as Copy

from Frontend.Common.IRNode import IRNodeVisitor, IRNode

from Frontend.SymbolTable import SymbolTable, SymbolTableBuilder

from Frontend.Shortcuts import getCurrentLanguageAst
from Tools.Tree import getFileNode
c_ast = getCurrentLanguageAst()

class CompoundAnnotation(IRNodeVisitor):
    """
        
    """
    def __init__(self,*args, **kwargs):
        print "*** DEPRECATION WARNING: CompoundAnnotation class is here only for backwards compatibility."
        print "*** It might produce collateral damage"
        super(CompoundAnnotation, self).__init__(*args, **kwargs)
        
        
    def visit_Compound(self, node):
        setattr(node, 'stmts', [])
        setattr(node, 'decls', [])
        for elem in node.block_items:
            if isinstance(elem, c_ast.Decl):
                node.decls.append(elem)
            else:
                node.stmts.append(elem)       
 

class AstToIR(object):
    """ Transform a C ast to the internal representation (Flyweight pattern)

    """

    _instanceCache = {}


    def __init__(self, Writer, ast):
        def fast_write(node):
            """Write the internal representation with the given Writer
                :param node: Node to write
            """
            writeIO = cStringIO.StringIO()
            cw = Writer(stream = writeIO)
            cw.visit(node)
            return writeIO.getvalue()

        # This need to be done at this point to use Writer inside a closure
        # so the __str__ uses a predefined writer
        # Overload __str__ method on Node class
        # Will apply to all instances !!
        IRNode.__str__ = fast_write
        self._symbolTable = None
        self._ast = ast
        # self._ast = getFileNode(ast)

    @staticmethod
    def deep_first_search(root, visited = None, preorder_process  = lambda x: None):
        """Given a starting vertex, root, do a depth-first search.

            :param root: Root to start search
            :param visited: List of already visited nodes
            :param preorder_process: Function to apply in all nodes
            
        """
        to_visit = [] 
        if visited is None: visited = set()
    
        to_visit.append(root) # Start with root
        while len(to_visit) != 0:
                  v = to_visit.pop()
                  if v not in visited and v:
                             visited.add(v)
                             preorder_process(v)
                             to_visit.extend(v.children())

    def link_all_parents(self, ast):
          """ Function to link the nodes of the AST in reverse order, using a parent attribute in each node 

                :param ast: Ast to relink
          """
          def link_parent(node):
                  for child in node.children():
                     if child and not (type(child) == type([])):
                        child.parent = node
                     elif type(child) == type([]):
                        import pdb
                        pdb.set_trace()
                        child.parent = node

          AstToIR.deep_first_search(root = ast, visited = None, preorder_process = link_parent)


    @staticmethod
    def link_subtree_parents(ast):
          """ Function to link the nodes of the AST in reverse order, using a parent attribute in each node 

                :param ast: Ast to relink
          """
          def link_parent(node):
                  for child in node.children():
			if child:
                             child.parent = node

          AstToIR.deep_first_search(root = ast, visited = None, preorder_process = link_parent)

    def annotate(self):
        """ Apply transformations needed to migrate from a c_ast to a InternalRepresentation.
             
                - Link all parents
                - Update symbol table
            
        """
        self.link_all_parents(self._ast)  
        self._symbolTable = SymbolTable()
        tsv = SymbolTableBuilder(symbol_table = self._symbolTable)
        tsv.visit(self._ast)
        # CompoundAnnotation().visit(self._ast)

    def update(self, node):
        """ Update parent links
            
            :param node: FileAST node to start conversion
        """
        self.annotate(node)  
        return node


    @property
    def symbolTable(self):
        """ Returns the symbol table for the current ast
            
            :param node: FileAST node to start conversion
        """
        if not self._symbolTable:
                self.annotate()
        return self._symbolTable

    @property
    def ast(self):
        """ Return the ast attached to this IR """
        return self._ast

    @staticmethod 
    def deepcopySubtree(subtree):
        new_var = subtree.__deepcopy__();
        AstToIR.link_subtree_parents(new_var)
        setattr(new_var, 'parent', subtree.parent)
        return new_var
    
