# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------


from Frontend.C99 import c99_lexer 
from ply.lex import TOKEN
class OmpLexer(object):

    @staticmethod
    def get_reserved():
        reserved = {
		'pragma'        : 'PRAGMA',
		# llc v2 syntax
		'llc'   	: 'LLC',
		# llc yacf-frangollo
		'context'	: 'REGION',
		'task'  	: 'TASK',
		'name'		: 'NAME',
		'concycle' 	: 'CONCYCLE',
		'in' 		: 'IN',
		'out'		: 'OUT',
		'inout'		: 'INOUT',
		# ----- Loop Transformations ------
		'nested_for'    : 'NESTED_FOR',
		'interchange'   : 'INTERCHANGE',
		'unroll'	: 'UNROLL',
		'unswitch'	: 'UNSWITCH',
		'tiling'	: 'TILING',
		# -------- omp ---------
		'omp'		: 'OMP_OMP',
		'parallel'	: 'OMP_PARALLEL',
		'sections'	: 'OMP_SECTIONS',
		# ------ clauses --------
#		'shared'	: 'OMPSHARED',
#		'private'	: 'OMP_PRIVATE',
		'reduction'	: 'OMP_REDUCTION',
		'nowait'	: 'OMP_NOWAIT',
		'threadprivate' : 'OMP_THREADPRIVATE',
		'copyin'	: 'OMP_COPYIN',
		'schedule'	: 'OMP_SCHEDULE',
		'omp_schedule_kind' : 'OMP_SCHEDULE_KIND',
		'master' 	: 'OMP_MASTER',
		'critical'	: 'OMP_CRITICAL',
		'atomic'	: 'OMP_ATOMIC',
		'barrier'	: 'OMP_BARRIER',
		'flush'		: 'OMP_FLUSH',
		'single'	: 'OMP_SINGLE',
#		'firstprivate'  : 'OMP_FIRSTPRIVATE',
		'collapse'	: 'OMP_COLLAPSE',
		'none'		: 'OMP_NONE',
		'result'	: 'OMP_RESULT',
		'nc_result'     : 'OMP_NCRESULT',
		'num_threads'   : 'OMP_NUM_THREADS',
#		'lastprivate'   : 'OMP_LASTPRIVATE',
		# ------- new clauses --------
		# From  "A proposal to extend the OpenMP Tasking Model for Heterogeneous Architectures" (2009)
		'device'	: 'DEVICE',
		'target'	: 'TARGET',
		'copy_in'	: 'OMP_COPY_IN',
		'copy_out'	: 'OMP_COPY_OUT',
		# ----- CELL syntax ---------
		'cell'		: 'CELL',
		# --------- CUDA -----------
		'cuda'		: 'CUDA',
		# --------- FPGA -----------
		'fpga'		: 'FPGA',
		# --------- MPI -----------
		'mpi'		: 'MPI',
		'hybrid'	: 'HYBRID',
		'smp' 		: 'SMP',
        # -------- QUICK HACK FOR CUDA/GMAC, CLEAN ME ------
     
                'default'       : 'OMP_DEFAULT',
                'for'		: 'OMP_FOR',
                'if'		: 'OMP_IF',
	    #'#'			: 'OMP_PPHASH',
#		'omp'		: 'OMP_OMP'

	}
        return reserved   
    #allreserved = reserved.copy()
    #allreserved.update(c99_lexer.C99Lexer.keyword_map)

    @staticmethod
    def get_tokens():
        tokens = list(OmpLexer.get_reserved().values()) + [
        'OMPSHARED', 'OMPPRIVATE', 'OMPFIRSTPRIVATE', 'OMPLASTPRIVATE'
        ]
        return tokens

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    #            Lexical states
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def t_OMPSHARED(self, t):
        r'shared'
        t.lexer.push_state('OMPSHARED')
        return t

    def t_OMPPRIVATE(self, t):
        r'private'
        t.lexer.push_state('OMPPRIVATE')
        return t

    def t_OMPFIRSTPRIVATE(self, t):
        r'firstprivate'
        t.lexer.push_state('OMPFIRSTPRIVATE')
        return t

    def t_OMPLASTPRIVATE(self, t):
        r'lastprivate'
        t.lexer.push_state('OMPLASTPRIVATE')
        return t

    def t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_ID(self, t):
        r'[a-zA-Z_][0-9a-zA-Z_]*'
        t.type = 'ID'
        return t

    ### States Tokens for [ ] : { } + - * / ,
    t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_LBRACKET = '\['
    t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_RBRACKET = '\]'
    t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_COLON    = ':'
    t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_LBRACE   = '\{'
    t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_RBRACE   = '\}'
    t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_TIMES    = '\*'
    t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_PLUS     = '\+'
    t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_MINUS    = '\-'
    t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_DIVIDE   = '\/'
    t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_COMMA    = ','

    ## TODO: Review this regular expression - see c99 standard
    def t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_INT_CONST_DEC(self, t):
        r'\d+'
        return t

    def t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_RPAREN(self, t):
        r'\)'
        t.lexer.pop_state()
        return t
        
    def t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_LPAREN(self, t):
        r'\('
        return t

    t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_ignore = ' \t'

    def t_OMPSHARED_OMPPRIVATE_OMPFIRSTPRIVATE_OMPLASTPRIVATE_error(self, t):
        print 'Lexical error in state "' + str(t.value[0]) + '" in line ' + str(t.lineno)
        t.lexer.skip(1)

    def t_directive_ID(self, t):
        r'[a-zA-Z_][0-9a-zA-Z_]*'
        t.type = OmpLexer.get_reserved().get(t.value, 'ID') 
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # States
        if (t.value == 'shared'):
            t.type = 'OMPSHARED'
            t.lexer.push_state('OMPSHARED')
        elif (t.value == 'private'):
            t.type = 'OMPPRIVATE'
            t.lexer.push_state('OMPPRIVATE')
        elif (t.value == 'firstprivate'):
            t.type = 'OMPFIRSTPRIVATE'
            t.lexer.push_state('OMPFIRSTPRIVATE')
        elif (t.value == 'lastprivate'):
            t.type = 'OMPLASTPRIVATE'
            t.lexer.push_state('OMPLASTPRIVATE')
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        return t

    @staticmethod
    def get_states():
        states = (
            ('OMPSHARED', 'exclusive'),
            ('OMPPRIVATE', 'exclusive'),
            ('OMPFIRSTPRIVATE', 'exclusive'),
            ('OMPLASTPRIVATE', 'exclusive'),
        )
        return states

if __name__ == '__main__':
    filename = '../testsC99/5_omp.c'
    text = open(filename).read()
    
    def errfoo(msg, a, b):
        printme(msg)
        sys.exit()
    
    def typelookup(namd):
        return False
    
    omplex = OmpLexer(errfoo, typelookup)
    omplex.build()
    omplex.input(text)
    
    while 1:
        tok = omplex.token()
        if not tok: 
			break
        print tok
