# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

from Tools.Dump import Dump

import config

#from yacf.Frontend.Parse import parse_source
from Frontend.FrontendFactory import FrontendFactory
from Frontend.Common.IRNode import *
from Frontend.SymbolTable import *
from Frontend.InternalRepr import AstToIR


TEST_PATH = config.FRONTEND_DIR + '/testsC99'
#CODE_PATH =  TEST_PATH + 'codes/'
#TREE_PATH =  TEST_PATH + 'trees/'

import unittest

from Backends.Common.tests.common import TestCase
from Backends.C.Writers.CWriter import CWriter

config.CURRENT_LANGUAGES = ['C99','Omp']

class TestSymbolTable(TestCase):

     def setUp(self):
          pass

     def test_symbolTable(self):
          """ Test symbol table """

          template_code = open(TEST_PATH + '/4_declarations.c', 'r').read()
          [ast, prepro_class, lexer_class, parser_class] = FrontendFactory().\
                        parse_with_language_list(template_code, 'declarations',
                        config.CURRENT_LANGUAGES)
          migrator = AstToIR(Writer = CWriter, ast = ast)
          migrator.annotate()
          st = migrator.symbolTable
          new_ast = migrator.ast

          current = str(st.lookUp(new_ast.ext[-1].body.block_items[4].lvalue))
          self.check_output(current, '{a: double, (20, 74)}')
          current = str(st.lookUp(new_ast.ext[-1].body.block_items[5].init.decls[0]))
          self.check_output(current, '{h: int, (42, 74)}')

     def test_all(self):
          """ Test all test codes """
          for test_code in [#'1_comments.c','2_longtypes.c',
                            '3_literals.c', '4_declarations.c',
                            #'5_omp.c','6_omp_default.c', 
                            '7_typedef.c', 
                            #'8_llc_nc_result.c', '9_clock.c'
                           ]:
                template_code = open(TEST_PATH + '/' + test_code, 'r').read()
                [ast, prepro_class, lexer_class, parser_class] = FrontendFactory().\
                        parse_with_language_list(template_code, test_code,
                        config.CURRENT_LANGUAGES)
                migrator = AstToIR(Writer = CWriter, ast = ast)
                migrator.annotate()


if __name__ == '__main__':
    test_suite = unittest.TestLoader().loadTestsFromTestCase(TestSymbolTable)
    unittest.TextTestRunner(verbosity=2).run(test_suite)
