'''
Created on Apr 15, 2012

@author: jlucasgl (jlucas.gl@gmail.com)

'''

from Frontend.Common.IRNode import IRNode

from Backends.Common.Mutators.AbstractMutator import AbstractMutator
from Backends.Common.Visitors.GenericVisitors import GenericFilterVisitor, AbstractReverseVisitor

from Frontend.Shortcuts import getCurrentLanguageAst
c_ast = getCurrentLanguageAst()

from Tools.Tree import getContainerAttribute

from MiddleEnd.Loop.Analysis import ParametrizeLoopTool
from Tools.Tree import ReplaceTool

class ParameterMustBeDefinedException(Exception):
    """ Parameter Must Be Defined Exception
    """
    def __init__(self, name = ""):
	self.name = name

    def __str__(self):
        return "The parameter '" + str(self.name) + "' must be defined in mutator function"

class CompundReverseFilter(AbstractReverseVisitor):
    """
    """
    def __init__(self):
        def condition(node):
            if type(node) == c_ast.Compound:
		    return True
            return False
        super(CompundReverseFilter, self).__init__(condition_func = condition)
        
class PragmaReverseFilter(AbstractReverseVisitor):
    """
    """
    def __init__(self):
        def condition(node):
            if type(node) == c_ast.Pragma:
                    return True
            return False
        super(PragmaReverseFilter, self).__init__(condition_func = condition)
    
class CompundFilter(GenericFilterVisitor):
    """
    """
    def __init__(self):
        def condition(node):
            if type(node) == c_ast.Compound:
                    return True
            return False
        super(CompundFilter, self).__init__(condition_func = condition)
    
class LoopFilter(GenericFilterVisitor):
    """ Returns the first node c_ast.For whose loop_variable name is
	the name of given c_ast.ID if given, else returns first one
    """
    def __init__(self, id = None):
        def condition(node):
            if type(node) == c_ast.For:
		if not id:
		    return True
		else:
		    loop_variable = \
			    ParametrizeLoopTool().get_loop_variable(node)
		    if id.name == loop_variable.name:
			return True
            return False
        super(LoopFilter, self).__init__(condition_func = condition)

class IDFilter(GenericFilterVisitor):
    """ Returns the first node c_ast.ID whose name is given value if
	exist, else returns first one
    """
    def __init__(self, value=None):
	""" @param value : Name with type str of ID to find
	"""
        def condition(node):
            # print "Finding", value
            if type(node) == c_ast.ID:
		if not value:
		    return True
		elif node.name == value:
		    return True
            return False
        super(IDFilter, self).__init__(condition_func = condition)

class IDMutator(AbstractMutator):
    """ Interchange old_id with new_id, both must be of type c_ast.ID
    """
    def __init__(self, old_id, new_id):
	""" """
	#print "IDMutator(", old_id, ":", type(old_id), ",", new_id, ":", type(new_id), ")"
        self.old_id = old_id
        self.new_id = new_id
        if (type(old_id) == type('')):
	    self.old_id = c_ast.ID(name = old_id)
	if (type(new_id) == type('')):
	    self.new_id = c_ast.ID(name = new_id)
	assert isinstance(self.old_id, c_ast.ID)
	assert isinstance(self.new_id, c_ast.ID)


    def filter(self, ast):
        """ """
        raise NotImplemented

    def filter_iterator(self, ast):
        """ Fast filter """
        return IDFilter(self.old_id.name).dfs_iter(ast)

    def fast_filter(self, ast):
        """ Fast filter , looking for compound statement """
        return IDFilter(self.old_id.name).dfs_iter(ast)

    def mutatorFunction(self, id_found):
	""" """
	assert isinstance(id_found, c_ast.ID)

	#print "id to find:", id_found
	
        if (id_found.name == self.old_id.name):

	    id_found_parent = id_found.parent
            cattr = getContainerAttribute(id_found)
	  
            #print " ID found in [", id_found.parent, "]"
            #print " new_id: ", self.new_id
            #print " id_found.parent:", id_found_parent
            #print " id_found.parent." + str(cattr) + ":", getattr(id_found_parent, cattr)

	    ReplaceTool(new_node = self.new_id,
                        old_node = id_found
                       ).apply(id_found_parent, cattr)
            
            #print " self.new_id.parent:", self.new_id.parent
            
	return id_found

class PragmaFilter(GenericFilterVisitor):
    """ Returns the first node c_ast.For whose loop_variable name is
        the name of given c_ast.ID if given, else returns first one
    """
    def __init__(self, id = None):
        def condition(node):
            if type(node) == c_ast.Pragma:
                return True
            return False
        super(PragmaFilter, self).__init__(condition_func = condition)
