# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
Created on Apr 28, 2011

@author: jlucasgl
'''

# from MiddleEnd.Loop.Analysis import ParametrizeLoopTool
from Backends.Common.Mutators.AbstractMutator import AbstractMutator

from Tools.Tree import ReplaceTool, getContainerAttribute, InsertTool
from Frontend.Shortcuts import getCurrentLanguageAst
from Backends.Common.Visitors.GenericVisitors import GenericFilterVisitor,\
    ArrayRefVisitor, AnyIDVisitor
from Tools.Tree import NodeNotFound
from MiddleEnd.DataAnalysis.LlcScope import ArrayIndexAnalysis
from Frontend.C99 import c99_ast
from MiddleEnd.Loop.Analysis import ParametrizeLoopTool
from Tools.Debug import DGraphDebug
from Frontend.InternalRepr import AstToIR

# self._aia.extractIndexVars(aref)
class InvariantRefFilter(GenericFilterVisitor):
    def _extractIndexVars(self, node):
        if not isinstance(node, c99_ast.ArrayRef):
            raise NotImplemented
        # For each var referenced inside subscript
        idVisitor = AnyIDVisitor()
        idList = [ elem for elem in idVisitor.iterate(node.subscript)]
        return idList
    
    def _checkInvariant(self, aref, forStmt):
        try:
            params = ParametrizeLoopTool().apply(forStmt)
            lvs = self._st.lookUp(params['loop_variable'])
            self._indexSymbols = [self._st.lookUp(iVar) for iVar in self._extractIndexVars(aref)]
            if lvs in self._indexSymbols:
                return False
            
            for var in self._indexSymbols:
                # Case there is a dependency bw a var inside index and the loop variable 
                #  transformation is not feasible
                if self._dGraph.checkDependency(self._dGraph[lvs], self._dGraph[var]):
                    # DGraphDebug().apply(self._dGraph)
                    print " Aref is loop dependent, cannot optimize"
                    return False
            else:
                # print " Statement is loop-independent, transform "
                return True
        except NodeNotFound:
            raise NotImplemented
    
    
    def __init__(self, st, dGraph):
        self._for = 0
        self._for_stmt = None
        self._st = st
        self._dGraph = dGraph
        # Create graph
        # self._aia = ArrayIndexAnalysis(dGraph, st)
        # Get index Variables
        # self._aia.visit(subtree)
        self._analyzed = set()
        self._relatedForStmt = dict()
        
        def condition_func(aref):
            # if isinstance(aref, c99_ast.ArrayRef) and (self._for == 1) and self._for_stmt:
            if isinstance(aref, c99_ast.ArrayRef)  and self._for_stmt:
                # check invariant of aref related to self._for_stmt
                if self._checkInvariant(aref, self._for_stmt) \
                        and not aref in self._analyzed \
                        and not isinstance(aref.parent, c99_ast.ArrayRef):
                    print " Aref: " + str(aref) + " is marked as potential invariant array access"
                    self._analyzed.add(aref)
                    self._relatedForStmt[aref.name] = self._for_stmt
                    return True
            return False
        super(InvariantRefFilter, self).__init__(condition_func = condition_func,
                                                    prev_brother = None,)

    def getForStmt(self, aref_name):
        return self._relatedForStmt[aref_name]
    
    def visit_For(self, node, prev, offset = 1, ignore = []):
        self._for += 1
        prev = self._for_stmt
        self._for_stmt = node
        rvalue =  self.generic_visit(node, offset, ignore)
        self._for_stmt = prev
        self._for -= 1
        return rvalue



class LoopInvariantMutator(AbstractMutator):
    def __init__(self, st, dGraph):
        self._st = st
        self._dGraph = dGraph
        self._ivf = InvariantRefFilter(self._st, self._dGraph)
        self._mutatorCount = 0
 
    def filter(self, subtree):
        """  """
        return self._ivf.apply(subtree)

    def filter_iterator(self, subtree):
        """ Search for invariant Ref accesses inside a particular subtree """
        return self._ivf.iterate(subtree)

    def fast_filter(self, subtree):
        return self._ivf.iterate(subtree)
    
    def _check_equivalent_subscript(self, subtree_a, subtree_b):
        # TODO: CHANGE THIS!!!!
        return (str(subtree_a) == str(subtree_b))

    
    def mutatorFunction(self, aref):
        self._mutatorCount += 1
        innerARef = aref
        invariantSymbol = self._st.lookUp(innerARef)
        invariantLoop = self._ivf.getForStmt(innerARef.name)
        
        print " Array access " + str(aref) + " is invariant to loop iterating with (" + str(invariantLoop.next) +")"
        
        
        # Create new private/register var before  invariantStmt
        # name, quals, storage, funcspec, type, init, bitsize
        # TODO: Add numbering to this variable
        privateVarName = "__invariant_tmp_" + str(self._mutatorCount)
        idNode = c99_ast.ID(name = privateVarName)
        
        # type __invariant_tmp; 
        newVar_node = c99_ast.Decl(name = privateVarName, quals = [], 
                                    storage = [], 
                                    funcspec = [], 
                                    bitsize = None, 
                                    type = invariantSymbol.type.type, 
                                    init = None, 
                                    parent = invariantLoop.parent)
        #  __invariant_tmp = array[position];
        update1_node = c99_ast.Assignment(op = '=', 
                                          lvalue = idNode,
                                          rvalue = AstToIR.deepcopySubtree(innerARef),  
                                          parent = invariantLoop.parent
                                          )
        # array[position] =  __invariant_tmp ;
        update2_node = c99_ast.Assignment(op = '=', 
                                          lvalue = AstToIR.deepcopySubtree(innerARef),
                                          rvalue = idNode,  
                                          parent = invariantLoop.parent
                                          )
        # Insert declaration
        InsertTool(subtree = newVar_node, position = "before", next = invariantLoop).apply(invariantLoop.parent, getContainerAttribute(invariantLoop))
        # Insert tmp update
        InsertTool(subtree = update1_node, position = "before", next = invariantLoop).apply(invariantLoop.parent, getContainerAttribute(invariantLoop))
        # Insert value update from tmp
        InsertTool(subtree = update2_node, position = "after", prev = invariantLoop).apply(invariantLoop.parent, getContainerAttribute(invariantLoop))
        # Replace all ocurrences of invariantStmt by newVar inside invariantLoop
        for aref in ArrayRefVisitor().iterate(invariantLoop.stmt):
            if self._st.lookUp(aref.name) == invariantSymbol and self._check_equivalent_subscript(innerARef.subscript, aref.subscript):
                ReplaceTool(new_node = idNode, old_node = aref).apply(aref.parent, getContainerAttribute(aref))

        print " -> Extracted an replaced by private var " + str(privateVarName)

