'''
Created on May 17, 2011

@author: jlucasgl (jlucas.gl@gmail.com)

'''

from MiddleEnd.Loop.Analysis import ParametrizeLoopTool
from Backends.Common.Mutators.AbstractMutator import AbstractMutator

from Tools.Tree import ReplaceTool, getContainerAttribute
from Frontend.Shortcuts import getCurrentLanguageAst
from Backends.Common.Visitors.GenericVisitors import GenericFilterVisitor
c_ast = getCurrentLanguageAst()

from Frontend.Omp import omp_ast

from MiddleEnd.Loop.Common import LoopFilter, IDFilter, IDMutator, CompundReverseFilter, ParameterMustBeDefinedException
from MiddleEnd.Loop.Mutators.LoopStripMining import get_new_index_name

class SizeRequiredNotFound(Exception):
    def __init__(self):
        pass

    def __str__(self):
        return "The class constructor expects an argument size"


class LoopStripMiningMutator(AbstractMutator):
    def __init__(self, size = None):
        self.size = size
        super(LoopStripMiningMutator, self).__init__()

    def filter(self, ast):
        """  """
        raise NotImplemented

    def filter_iterator(self, ast):
        """ Fast filter  """
        return NotImplemented

    def fast_filter(self, ast):
        """ Fast filter , looking for binary expressions """
        return LoopTilingFilter().dfs_iter(ast)

    def mutatorFunction(self, ast, size=None):

	# Test if ast points to a c_ast.For node
        if isinstance(ast, omp_ast.llcTiling):
	  loopTiling_directive_ast = ast
	  for_ast = LoopFilter().apply(loopTiling_directive_ast)
	  assert isinstance(for_ast, c_ast.For)
	else:
	  loopTiling_directive_ast = None
	  for_ast = ast
	  assert isinstance(for_ast, c_ast.For)

        # Parametrize
        parametrizer = ParametrizeLoopTool()
        loop_parameters = parametrizer.apply(for_ast,)

        print "*** Applying STRIPMINING over loop", \
	      loop_parameters['loop_variable'].name, "***"

        #--------------------------------------------------------------
        # from Tools.Debug import DotDebugTool
        # DotDebugTool(highlight = [for_ast,]).apply(llc_node.parent.parent)
        #--------------------------------------------------------------

        # Get sizes of the tiling loop blocks
	try:
	    self.size = loopTiling_directive_ast.size
	except AttributeError:
	    if size: self.size = size
	    elif self.size is None:
		    raise ParameterMustBeDefinedException('size')
        
        #==============================================================
        # 1. Get parameters to modify AST
        #==============================================================
	loop_variable = loop_parameters['loop_variable']
        new_index_name =  get_new_index_name(loop_variable)

	init_node = loop_parameters['init_node']
	cond_node = loop_parameters['cond_node']
	last_iteration = loop_parameters['last_iteration']
	next_for = for_ast.next

	size_constant = c_ast.Constant(type = 'int', value = str(self.size))

        #==============================================================
        # 2. Build external for loop from original loop changing its
        #    condition node and its loop variable
        #==============================================================
        #    2.1 Replace cond_node with new_for_condition_node
        #==============================================================
        cond_node = for_ast.cond
        new_for_condition_node = c_ast.BinaryOp(op = '<',
                                            left = c_ast.ID(name = new_index_name),
                                            right = size_constant
                                        )
	ReplaceTool(new_node = new_for_condition_node,
                        old_node = cond_node
                       ).apply(for_ast, getContainerAttribute(cond_node))

        #==============================================================
        #    2.2 Update ID in loop init_node
        #==============================================================
        IDMutator(old_id = loop_variable,
	          new_id = c_ast.ID(name = new_index_name)
	          ).fast_apply_all(init_node)          
	# Test if IDMutator do its job right
        test_id_1 = IDFilter().apply(init_node)
        assert isinstance(test_id_1, c_ast.ID)
        assert (test_id_1.name == new_index_name)

        #==============================================================
        #    2.3 Update ID in loop stride
        #==============================================================
        IDMutator(old_id = loop_variable,
                  new_id = c_ast.ID(name = new_index_name)
		  ).fast_apply_all(next_for)        
	# Test if IDMutator do its job right
        test_id_2 = IDFilter().apply(next_for)
        assert isinstance(test_id_2, c_ast.ID)
        assert(test_id_2.name == new_index_name)
        
        #==============================================================
        # 3. Build new internal for loop
        #==============================================================
        #    3.1 Build the internal for loop with new init node
        #        and step node
        #==============================================================
        cond_node = c_ast.BinaryOp(op = '<',
                                        left = loop_variable,
                                        right = last_iteration
                                  )

        init_node = c_ast.Assignment(op = '=',
                                        lvalue = loop_variable,
                                        rvalue = c_ast.ID(name = new_index_name)
                                    )

        #==============================================================
        #    3.2 Generate new internal for loop
        #==============================================================

        internal_loop_parameters = {}
        internal_loop_parameters['init_node'] = init_node
        internal_loop_parameters['cond_node'] = cond_node
        internal_loop_parameters['loop_variable'] = loop_variable
        internal_loop_parameters['stride'] = size_constant

        parametrizer_internal = ParametrizeLoopTool()
        parametrizer_internal.update(internal_loop_parameters)
        internal_for_ast = parametrizer_internal.buildLoop()        
        
        #print internal_for_ast
        ReplaceTool(new_node = for_ast.stmt,
                    old_node = None
                   ).apply(internal_for_ast, 'stmt')
	
        setattr(init_node, 'parent', internal_for_ast)
        setattr(cond_node, 'parent', internal_for_ast)

        #DotDebugTool(highlight = []).apply(internal_for_ast)

        #================================================================
        #    3.3 Attach original for loop compound to internal for loop
        #================================================================
        new_internal_compound = c_ast.Compound(block_items = [internal_for_ast,])
	setattr(internal_for_ast, 'parent', new_internal_compound)

	ReplaceTool(new_node = new_internal_compound,
                    old_node = for_ast.stmt
                   ).apply(for_ast, getContainerAttribute(for_ast.stmt))

        #==============================================================
        # 4. Replace and update the AST
        #==============================================================
        #    4.1 Replace the forLoop with the tiling compound statement
        #==============================================================
        
        # Get llc node        
        if loopTiling_directive_ast:
	    llc_node = loopTiling_directive_ast.parent
	    ReplaceTool(new_node = for_ast, 
                    old_node = llc_node
                    ).apply(llc_node.parent, 
                            getContainerAttribute(llc_node)
                           )

        #==============================================================
        #    4.2 Finally, build declaration of new loop variable
        #==============================================================
        x_variable_decl  =  c_ast.Decl(name = new_index_name,
                                        quals = [],
                                        storage = [],
                                        funcspec = [],
                                        type = c_ast.TypeDecl(declname = new_index_name,
                                                        quals = [], 
                                                        type = c_ast.IdentifierType(names = ['int'])
                                                    ),
                                        init = None,
                                        bitsize = None
                                    )

        over_for = CompundReverseFilter().apply(for_ast)
                                    
        setattr(over_for, "block_items", [x_variable_decl, ] +
		    over_for.block_items)

        #for_ast.parent.block_items = [x_variable_decl, ] + for_ast.parent.block_items

        #--------------------------------------------------------------
        # from Tools.Debug import DotDebugTool
        # DotDebugTool(highlight = [for_ast, x_variable_decl]).apply(llc_node.parent.parent)
        #--------------------------------------------------------------
	return ast

