# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
Created on May 17, 2011

@author: jlucasgl (jlucas.gl@gmail.com)

'''

from MiddleEnd.Loop.Analysis import ParametrizeLoopTool
from Backends.Common.Mutators.AbstractMutator import AbstractMutator

from Tools.Tree import ReplaceTool, getContainerAttribute
from Frontend.Shortcuts import getCurrentLanguageAst
from Backends.Common.Visitors.GenericVisitors import GenericFilterVisitor
c_ast = getCurrentLanguageAst()

from Frontend.Omp import omp_ast

from MiddleEnd.Loop.Common import LoopFilter, IDFilter, IDMutator, CompundReverseFilter, ParameterMustBeDefinedException

class SizeRequiredNotFound(Exception):
    def __init__(self):
        pass

    def __str__(self):
        return "The class constructor expects an argument size"


def get_new_index_name(loop_variable=''):
    """ Returns a new index variable string as name """
    return "__" + str(loop_variable) + "__"

def get_old_index_name(new_loop_variable=''):
    """ Returns a new index variable string as name """
    return str(new_loop_variable).replace('__', '')

class LoopStripMiningMutator(AbstractMutator):
    def __init__(self, size = None):
        self.size = size
        super(LoopStripMiningMutator, self).__init__()

    def filter(self, ast):
        """  """
        raise NotImplemented

    def filter_iterator(self, ast):
        """ Fast filter  """
        return NotImplemented

    def fast_filter(self, ast):
        """ Fast filter , looking for binary expressions """
        return LoopTilingFilter().dfs_iter(ast)
    
    def mutatorFunction(self, ast, size=None):

	# Test if ast points to a c_ast.For node
        if isinstance(ast, omp_ast.llcTiling):
	  loopTiling_directive_ast = ast
	  for_ast = LoopFilter().apply(loopTiling_directive_ast)
	  assert isinstance(for_ast, c_ast.For)
	else:
	  loopTiling_directive_ast = None
	  for_ast = ast
	  assert isinstance(for_ast, c_ast.For)

        # Parametrize
        parametrizer = ParametrizeLoopTool()
        loop_parameters = parametrizer.apply(for_ast,)

        print "*** Applying STRIPMINING over loop", \
	      loop_parameters['loop_variable'].name, "***"

        #--------------------------------------------------------------
        # from Tools.Debug import DotDebugTool
        # DotDebugTool(highlight = [for_ast,]).apply(llc_node.parent.parent)
        #--------------------------------------------------------------

        # Get sizes of the tiling loop blocks
	try:
	    self.size = loopTiling_directive_ast.size
	except AttributeError:
	    if size: self.size = size
	    elif self.size is None:
		    raise ParameterMustBeDefinedException('size')
        
        #==============================================================
        # Modify parameters with the tiling
        #==============================================================
        # 1. 
        #==============================================================
	loop_variable = loop_parameters['loop_variable']
        x_var_name = get_new_index_name(loop_variable)
        x_id_ast = c_ast.ID(name = x_var_name)

	init_node = loop_parameters['init_node']
	cond_node = loop_parameters['cond_node']
	last_iteration = loop_parameters['last_iteration']

        size_constant = c_ast.Constant(type = 'int', value = str(self.size))

        #==============================================================
        # 2. Build external for loop changing its next node and index
        #==============================================================

        next_for = for_ast.next
        new_next_for = c_ast.Assignment(op = '+=',
                                       lvalue = x_id_ast,
                                       rvalue = size_constant
                                      )
        setattr(next_for.parent, getContainerAttribute(next_for), new_next_for)

	IDMutator(old_id = loop_variable, new_id = c_ast.ID(name = x_var_name)
		 ).fast_apply_all(init_node)
        
	IDMutator(old_id = loop_variable, new_id = c_ast.ID(name = x_var_name)
		 ).fast_apply_all(cond_node)
        
        #==============================================================
        # 3. Build new internal for loop
        #==============================================================
        #    3.1 Build the ternary operator parameters for inner loop
        #        last iteration
        #==============================================================
        next_iteration = c_ast.BinaryOp(op = '+',
                                            left = x_id_ast,
                                            right = size_constant
                                    )

        ternary_condition = c_ast.BinaryOp(op = '<',
                                            left = next_iteration,
                                            right = last_iteration
                                        )

        #==============================================================
        #    3.2 Build the internal for loop with new loop_variable as
        #        index
        #==============================================================
        cond_node = c_ast.BinaryOp(op = '<',
                                        left = loop_variable,
                                        right = c_ast.TernaryOp(cond = 			ternary_condition, 
						iftrue = next_iteration,
                                                iffalse = last_iteration
                                                )
                                )

        init_node = c_ast.Assignment(op = '=',
                                        lvalue = loop_variable,
                                        rvalue = x_id_ast
                                    )

        #iterator = c_ast.UnaryOp(op='p++', expr = x_id_ast)

        internal_loop_parameters = {}
        internal_loop_parameters['init_node'] = init_node
        internal_loop_parameters['cond_node'] = cond_node
        internal_loop_parameters['loop_variable'] = loop_variable
        internal_loop_parameters['stride'] = c_ast.Constant(type = 'int', value = str(1))

        parametrizer_internal = ParametrizeLoopTool()
        parametrizer_internal.update(internal_loop_parameters)
        internal_for_ast = parametrizer_internal.buildLoop()        
        
        #print internal_for_ast
        internal_for_ast.stmt = for_ast.stmt

        #DotDebugTool(highlight = []).apply(internal_for_ast)

        #================================================================
        #    3.3 Attach internal for loop to original for loop
        #================================================================
        new_internal_compound = c_ast.Compound(block_items = [internal_for_ast,])
	setattr(internal_for_ast, 'parent', new_internal_compound)
        
        setattr(for_ast, getContainerAttribute(for_ast.stmt), new_internal_compound)        
	setattr(new_internal_compound, 'parent', for_ast)

        #==============================================================
        # 4. Replace and update the AST
        #==============================================================
        #    4.1 Replace the forLoop with the tiling compound statement
        #==============================================================
        
        # Get llc node        
        if loopTiling_directive_ast:
	    llc_node = loopTiling_directive_ast.parent
	    ReplaceTool(new_node = for_ast, 
                    old_node = llc_node
                    ).apply(llc_node.parent, 
                            getContainerAttribute(llc_node)
                           )

        #==============================================================
        #    4.2 Finally, build declaration of new loop variable
        #==============================================================
        x_variable_decl  =  c_ast.Decl(name = x_var_name,
                                        quals = [],
                                        storage = [],
                                        funcspec = [],
                                        type = c_ast.TypeDecl(declname = x_var_name, 
                                                        quals = [], 
                                                        type = c_ast.IdentifierType(names = ['int'])
                                                    ),
                                        init = None,
                                        bitsize = None
                                    )

        over_for = CompundReverseFilter().apply(for_ast)
                                    
        setattr(over_for, "block_items", [x_variable_decl, ] +
		    over_for.block_items)

        #for_ast.parent.block_items = [x_variable_decl, ] + for_ast.parent.block_items

        #--------------------------------------------------------------
        # from Tools.Debug import DotDebugTool
        # DotDebugTool(highlight = [for_ast, x_variable_decl]).apply(llc_node.parent.parent)
        #--------------------------------------------------------------
	return ast

