# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
Created on Jun 16, 2012

@author: jlucasgl (jlucas.gl@gmail.com)

'''

from MiddleEnd.Loop.Analysis import ParametrizeLoopTool
from Backends.Common.Mutators.AbstractMutator import AbstractMutator

from Tools.Tree import ReplaceTool, getContainerAttribute
from Frontend.Shortcuts import getCurrentLanguageAst
from Backends.Common.Visitors.GenericVisitors import GenericFilterVisitor
c_ast = getCurrentLanguageAst()

from Frontend.Omp import omp_ast

from MiddleEnd.Loop.Common import LoopFilter, IDFilter, \
      ParameterMustBeDefinedException
from MiddleEnd.Loop.Mutators.LoopInterchange \
      import LoopInterchangeMutator
      
from MiddleEnd.Loop.Mutators.LoopStripMining import get_new_index_name
#from MiddleEnd.Loop.Mutators.LoopStrip2 import LoopStripMiningMutator
from MiddleEnd.Loop.Mutators.LoopStripMining import LoopStripMiningMutator
                                                    


from Tools.Tree import NodeNotFound

class LoopTilingFilter(GenericFilterVisitor):
    """ Returns the first node with the given attribute
    """
    def __init__(self):
        def condition(node):
            if type(node) == omp_ast.llcTiling:
                return True
            return False
        super(LoopTilingFilter, self).__init__(condition_func = condition)

class LoopTilingMutator(AbstractMutator):
    def __init__(self, size = None, indexes=[]):
        self.size = size
        self.indexes = indexes
        super(LoopTilingMutator, self).__init__()

    def filter(self, ast):
        """  """
        raise NotImplemented

    def filter_iterator(self, ast):
        """ Fast filter  """
        return NotImplemented

    def fast_filter(self, ast):
        """ Fast filter , looking for binary expressions """
        return LoopFilter().dfs_iter(ast)

    def process_index(self, indexes=[]):
	""" Process to fit indexes """
	if (len(indexes) <= 1):
	    pass
	elif (len(indexes) == 2):
	    # If both tiling mutators are applied to the same loop then
	    if (str(indexes[0]) == str(indexes[1])):
		# Second index name need to be regenerated
		indexes[1] = c_ast.ID(name = get_new_index_name(indexes[0]))
		# As we construct new loops from inner to outter,
		# size list need to be reverse
		self.size.reverse()
	else:
            pass
	    #raise NotImplemented
	return indexes
    
    def mutatorFunction(self, ast, size=None, indexes=[]):

	#import copy
	#new_ast = copy.deepcopy(ast)
    
	# Test if ast points to a c_ast.For node
	assert isinstance(ast, c_ast.For)

        # Set list of indexes
        if len(indexes) != 0:
	    self.indexes = indexes
	elif len(self.indexes) == 0:
	    parametrizer = ParametrizeLoopTool()
	    loop_parameters = parametrizer.apply(ast,)
	    loop_ast = ast
	    while loop_ast:
		loop_parameters = parametrizer.apply(loop_ast,)
		indexes += [loop_parameters['loop_variable']]
		try:
		    loop_ast = LoopFilter().apply(loop_ast.stmt)
		    #print "loop_ast:\n" + str(loop_ast)
		except NodeNotFound:
		    loop_ast = None

        # Get sizes of the stripminng loop blocks
	if size: self.size = size
	elif self.size is None:
	    raise ParameterMustBeDefinedException('size')
		    
	indexes_procesed = self.process_index(indexes)	
	str_indexes = [x.name for x in indexes_procesed]
	print "*** Applying TILING over loops", str_indexes, "***"

	assert(len(str_indexes) == len(self.size))
	
        #----------------------------------------------------------------
        # from Tools.Debug import DotDebugTo	print "* Applying stripminng over loop ", indexes[-1]ol
        # DotDebugTool(highlight = [for_ast,]).apply(llc_node.parent.parent)
        #----------------------------------------------------------------

	#print "range:", range(len(indexes))

	#================================================================
	# 1. Apply strip-mining to n-1 loops
	#================================================================
	loop = ast
	for i in range(len(indexes)-1):

	#================================================================
	#   1.1 Filter interest loop by its index
	#================================================================
	    loop = LoopFilter(indexes[i]).apply(ast)
	    #print "loop:\n" + str(loop)

	#================================================================
	#   1.2 Apply stripminng mutator with a given size
	#================================================================
	    if type(self.size) == type([]):
	      	stripminng_ast = LoopStripMiningMutator().mutatorFunction(loop, self.size[i])
	    else:
		stripminng_ast = LoopStripMiningMutator().mutatorFunction(loop, self.size)
	    #print "stripminng_ast:\n" + str(stripminng_ast)

	#================================================================
	#   1.3 Apply interchange mutator over indexes (i, i+1)
	#================================================================
	    interchange_ast = LoopInterchangeMutator()\
		.mutatorFunction(loop, indexes[i].name,indexes[i+1].name)
	    #print "interchange_ast:\n" + str(interchange_ast)
	    #raw_input("Press ENTER to continue")
	
	#print ast
	#raw_input("Press ENTER to continue")
	
	#================================================================
	# 2. Apply strip-mining to last loop
	#================================================================
	loop = LoopFilter(indexes[-1]).apply(ast)
	#print "loop:\n" + str(loop), type(self.size), self.size
	if type(self.size) == type([]):
	    stripminng_ast = LoopStripMiningMutator().mutatorFunction(loop, self.size[-1])
	else:
	    stripminng_ast = LoopStripMiningMutator().mutatorFunction(loop, self.size)

	#print "stripminng_ast:\n" + str(stripminng_ast)
	#raw_input("Press ENTER to continue")

	
	#================================================================
	# 3. Finally reorder loops
	#================================================================
	interchange_ast = loop

	#print range(len(indexes)-2, -1, -1)
	#print indexes[0+1].name, indexes[0].name
	
	for j in range(len(indexes)-2, -1, -1):

	    # Apply loop interchange mutator over interchange_ast
	    #print "2. Applying interchange over ", indexes[j+1].name, indexes[j].name

	    interchange_ast = LoopInterchangeMutator().mutatorFunction(ast, indexes[j+1].name, 
                                                                            indexes[j].name)
	    #print "interchange_ast:\n" + str(interchange_ast)
	    #raw_input("Press ENTER to continue")

        
	return ast

