# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

from Backends.DotBackend.Writers.DotWriter import DotWriter, DGraphWriter
import subprocess
import os
from cStringIO import StringIO

import config
import sys


class DotDebugTool(object):
        """Display the tree with xdot, highlighting the given nodes 

            :param highlight: Nodes to highlight, default: None

            .. warning::
                    This tool requires xdot in *config.WORKDIR*, and uses a temporary
                    storage for the dot file
        """
        def __init__(self, highlight = None):
            """Instantiates the debug tool
                 
                :param highlight: Nodes to highlight, default: None
            """
            self.tmpfile = '/tmp/dotfile.dot'
            self.MAX_LINES = 20000

            if type(highlight) != type([]) and highlight:
                self.highlight = [highlight]
            else:
                self.highlight = highlight

        def apply(self,node):
            """Display the given node (or list of nodes)
                 
                :param node: Node or list of nodes to begin display
            """
            if type(node) == type([]):
                for elem in node:
                    self.debug_node(elem)
            else:
                self.debug_node(node)

        def debug_node(self, node):
            """Display the given node (or list of nodes)
                 
                :param node: Node to display
            """
            DotWriter(filename = self.tmpfile, highlight = self.highlight).visit(node)
            size =  len(open(self.tmpfile).readlines())
            if not (size > self.MAX_LINES):
                try:
                  p = subprocess.Popen("python " + config.WORKDIR + "/utils/xdot.py " + "/tmp/dotfile.dot", shell=True)
                  sts = os.waitpid(p.pid, 0)[1]
                except:
                     print "DotDebugTool:::: Cannot launch xdot"
            else:
                print "DotDebugTool:::: AST Too big to show"



class DGraphDebug(DotDebugTool):
        def debug_node(self, depGraph):
            """Display the given node (or list of nodes)
                 
                :param node: Node to display
            """
            DGraphWriter(filename = self.tmpfile, highlight = self.highlight).apply(depGraph)
            size =  len(open(self.tmpfile).readlines())
            if not (size > self.MAX_LINES):
                try:
                  p = subprocess.Popen("python " + config.WORKDIR + "/utils/xdot.py " + "/tmp/dotfile.dot", shell=True)
                  sts = os.waitpid(p.pid, 0)[1]
                except:
                  print "DGraphDebug:::: Cannot launch xdot"
            else:
                print "DotDebugTool:::: AST Too big to show"
