'''
Created on Jul 8, 2012

@author: jlucasgl (jlucas.gl@gmail.com)

'''
from Frontend.Shortcuts import getCurrentLanguageAst
c_ast = getCurrentLanguageAst()

#from Backends.C99.Writers.OmpWriter import OmpWriter
#from Backends.C99.Writers.OmpWriter import CWriter
from Backends.C99.Writers.AccWriter import AccWriter

from sys import argv, exit

import config


def genTs(original_source, filename):
        ###################### First Layer  : File parsing
        # Parse file
        print "Parsing " + filename + " .... ",
        config.CURRENT_LANGUAGES = ['C99','Acc']
        # original_source = " ".join(open(filename, 'r').readlines())
        from Frontend.FrontendFactory import FrontendFactory
        [ast, prepro_class, lexer_class, parser_class] = FrontendFactory().parse_with_language_list(original_source, filename, config.CURRENT_LANGUAGES)
        print " OK "
        print " Migrating to Internal Representation ...."
        from Frontend.C99.C99InternalRepr import C99AstToIR
        # Transform the C99 ast into the intermediate representation
        migrator = C99AstToIR(Writer = AccWriter, ast = ast)
        print "*************************"
        migrator.annotate()
        new_ast = migrator.ast
        print "*************************"
        print " OK "
        # Get the symbol table
        from Frontend.SymbolTable import *
        print " Build symbol table ...."
        st = migrator.symbolTable
        print "*************************"
        return [new_ast, st]


output_file = None

accgangworker_test = """
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>


typedef  struct timeval  CLOCK_TYPE;


int main ( int argc, char *argv[] );
void r8_mxm ( int l, int m, int n );
double r8_uniform_01 ( int *seed );
int n_threads = 0;

/******************************************************************************/

int main ( int argc, char *argv[] )

/******************************************************************************/
/*
  Purpose:

    MAIN is the main program for MXM.

  Licensing:

    This code is distributed under the GNU LGPL license.

  Modified:

    19 April 2009

  Author:

    John Burkardt
*/
{
  int id;
  int l;
  int m;
  int n;

  /*printf ( "\n" );
  printf ( "MXM\n" );
  printf ( "  C/OpenMP version.\n" );
  printf ( "\n" );
  printf ( "  Matrix multiplication tests.\n" );

  printf ( "\n" );
  printf ( "  Number of processors available = %d\n", omp_get_num_procs ( ) );
  printf ( "  Number of threads =              %d\n", omp_get_max_threads ( ) );*/

  if (argc < 2) {
     /* printf ("Usage ./mxm_omp size\n     size: Size of the matrix (l,m and n)\n"); */
     return -1;
  }
  l = atoi(argv[1]);
  m = l;
  n = l;

  r8_mxm ( l, m, n );

  /* 
  printf ( "\n" );
  printf ( "MXM:\n" );
  printf ( "  Normal end of execution.\n" );
  */
  
  return 0;
}
/******************************************************************************/

void r8_mxm ( int l, int m, int n )

/******************************************************************************/
/*
  Purpose:

    R8_MXM carries out a matrix-matrix multiplication in R8 arithmetic.

  Discussion:

    A(LxN) = B(LxM) * C(MxN).

  Licensing:

    This code is distributed under the GNU LGPL license.

  Modified:

    13 February 2008

  Author:

    John Burkardt

  Parameters:

    Input, int L, M, N, the dimensions that specify the sizes of the
    A, B, and C matrices.
*/
{
  double *a;
  double *b;
  double *c;
  int i;
  int j;
  int k;
  int ops;
  double rate;
  int seed;
  double time_begin;
  double time_elapsed;
  double time_stop;
/*
  Allocate the matrices.
*/
  a = ( double * ) malloc ( l * n * sizeof ( double ) );
  b = ( double * ) malloc ( l * m * sizeof ( double ) );
  c = ( double * ) malloc ( m * n * sizeof ( double ) );
/*
  Assign values to the B and C matrices.
*/
  seed = 123456789;

  for ( k = 0; k < l * m; k++ )
  {
    b[k] = r8_uniform_01 ( &seed );
  }

  for ( k = 0; k < m * n; k++ )
  {
    c[k] = r8_uniform_01 ( &seed );
  }
/*
  Compute A = B * C.
*/
  //omp_set_num_threads (n_threads);

  CLOCK_TYPE chrono;
  gettimeofday ((&chrono), NULL);
  
#pragma acc kernels copyin(b [l * m],c [m * n], l, m, n) copy(a [l * n])
{
  #pragma acc loop private(j) gang(10)
  for ( j = 0; j < n; j++)
  {
    #pragma acc loop private(i) worker(20)
    for ( i = 0; i < l; i++ )
    {
      a[i+j*l] = 0.0;
      for ( k = 0; k < m; k++ )
      {
        a[i+j*l] = a[i+j*l] + b[i+k*l] * c[k+j*m];
      }
    }
  }
}

  CLOCK_TYPE ch2;
  gettimeofday ((&ch2), NULL);
  time_elapsed = (double )(ch2.tv_sec  - chrono.tv_sec ) + ((double )(ch2.tv_usec  - chrono.tv_usec ) / 1.0e6);
  /* printf  ("Time: %g s\n"  ,time_elapsed  ) ; */


/*
  Report.
*/
  //ops = l * n * ( 2 * m );
  //time_elapsed = time_stop - time_begin;
  //rate = ( double ) ( ops ) / time_elapsed / 1000000.0;

  /*printf ( "\n" );
  printf ( "R8_MXM matrix multiplication timing.\n" );
  printf ( "  A(LxN) = B(LxM) * C(MxN).\n" );
  printf ( "  L = %d\n", l );
  printf ( "  M = %d\n", m );
  printf ( "  N = %d\n", n );
  printf ( "  Floating point OPS roughly %d\n", ops );
  printf ( "  Elapsed time dT = %f\n", time_elapsed );
  printf ( "  Rate = MegaOPS/dT = %f\n", rate );*/

  free ( a );
  free ( b );
  free ( c );

  return;
}
/******************************************************************************/

double r8_uniform_01 ( int *seed )

/******************************************************************************/
/*
  Purpose:

    R8_UNIFORM_01 is a unit pseudorandom R8.

  Discussion:

    This routine implements the recursion

      seed = 16807 * seed mod ( 2**31 - 1 )
      unif = seed / ( 2**31 - 1 )

    The integer arithmetic never requires more than 32 bits,
    including a sign bit.

  Licensing:

    This code is distributed under the GNU LGPL license.

  Modified:

    11 August 2004

  Author:

    John Burkardt

  Reference:

    Paul Bratley, Bennett Fox, Linus Schrage,
    A Guide to Simulation,
    Springer Verlag, pages 201-202, 1983.

    Bennett Fox,
    Algorithm 647:
    Implementation and Relative Efficiency of Quasirandom
    Sequence Generators,
    ACM Transactions on Mathematical Software,
    Volume 12, Number 4, pages 362-376, 1986.

  Parameters:

    Input/output, int *SEED, a seed for the random number generator.

    Output, double R8_UNIFORM_01, a new pseudorandom variate, strictly between
    0 and 1.
*/
{
  int k;
  double r;

  k = *seed / 127773;

  *seed = 16807 * ( *seed - k * 127773 ) - k * 2836;

  if ( *seed < 0 )
  {
    *seed = *seed + 2147483647;
  }

  r = ( double ) ( *seed ) * 4.656612875E-10;

  return r;
}
"""

if len(argv) > 1:
    filename  = argv[1]
    if len(argv) > 2 :
        output_file = argv[2]
    original_source = " ".join(open(filename, 'r').readlines())
else:
    filename  = 'accgangworker_test'
    original_source = accgangworker_test

[new_ast, st] = genTs(original_source, filename)


from MiddleEnd.Acc.AccLoopFilters import AccConstructFilter, AccLoopGangFilter, AccLoopWorkerFilter, \
					      AccGangFilter, AccWorkerFilter
from MiddleEnd.Loop.Analysis import ParametrizeLoopTool

from Frontend.Acc import acc_ast
from MiddleEnd.Loop.Common import LoopFilter

#Filter region
region_ast = AccConstructFilter().apply(new_ast)

print region_ast



# Filter loop gang
accloop_gang = AccLoopGangFilter().apply(region_ast)
assert isinstance(accloop_gang, acc_ast.AccLoop)

# Search c99 loop into accloop
loop_gang = LoopFilter().apply(accloop_gang.stmt)
gang_loop_parameters = ParametrizeLoopTool().apply(loop_gang,)

# Extract parameters of loop_gang
gang_loop_variable = gang_loop_parameters['loop_variable']
gang_value = AccGangFilter().apply(accloop_gang).value

print "Gang of ", gang_value, ". Loop", gang_loop_variable

# Filter acloop worker
accloop_worker = AccLoopWorkerFilter().apply(accloop_gang)
assert isinstance(accloop_worker, acc_ast.AccLoop)

# Search c99 loop into accloop
loop_worker = LoopFilter().apply(accloop_worker.stmt)
worker_loop_parameters = ParametrizeLoopTool().apply(loop_worker,)

# Extract parameters of loop_worker
worker_loop_variable = worker_loop_parameters['loop_variable']
worker_value = AccWorkerFilter().apply(accloop_worker).value

print "Worker of ", worker_value, ". Loop", worker_loop_variable

raw_input("Press ENTER to continue")

from MiddleEnd.Loop.Mutators.LoopTiling \
      import LoopTilingMutator

# Apply tiling mutatorFunction with a given tiling size as parameter
new_sm_ast = LoopTilingMutator().mutatorFunction(loop_gang,
				    size=[gang_value, worker_value],
				    indexes=[gang_loop_variable,
					     worker_loop_variable])

############################################3
# Write file

# Call pretty printer over the file
if output_file:
    v = AccWriter(filename = output_file)
    v.visit(new_ast)
    del v  # Ensure file closing
    import os
    if os.system("indent -kr " + output_file) != 0:
        print " You need to install the indent tool to pretty print ouput files "
else:
    #new_ast.show(attrnames = True)
    print "************"
    print str(new_ast)

