# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

# from Visitors.clone_visitor import CWriter, OmpWriter
from Backends.C.Writers.OmpWriter import OmpWriter

from sys import argv, exit

output_file = None

if len(argv) > 1:
	filename  = argv[1]
	if len(argv) > 2 :
		output_file = argv[2]

else:
	print ">>> File not found!"
	exit()



###################### First Layer  : File parsing

original_source = " ".join(open(filename, 'r').readlines())


# Parse file
CURRENT_LANGUAGES = ['C99', 'Acc',]
from Frontend.FrontendFactory import FrontendFactory
[ast_, prepro_class, lexer_class, parser_class,] = FrontendFactory().parse_with_language_list(original_source, filename, CURRENT_LANGUAGES,[])
# ast_.show()
ast = ast_


print " OK "


print " Migrating to Internal Representation ....", 


from Frontend.InternalRepr import AstToIR

# Transform the C ast into the internal representation
migrator = AstToIR(Writer = OmpWriter, ast = ast)
migrator.annotate()
new_ast = migrator.ast

print " OK "

# Get the symbol table
from Backends.Common.Visitors.GenericVisitors import GenericFilterVisitor

from Frontend.C import c_ast
from Frontend.Common.IRNode import IRNodeVisitor

from Frontend.SymbolTable import *


print " Build symbol table ....", 

st = SymbolTable()
tsv = SymbolTableBuilder(symbol_table = st)
tsv.visit(new_ast)


from Backends.DotBackend.Writers.DotWriter import DotWriter

# Print the AST
v = DotWriter(filename = output_file, highlight = [ast])
v.visit(new_ast)

# Ensure file is closed and saved (calling destructor)
del v


