# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
 :date: Feb 17, 2011
 :author: rreyes
 
'''

from sys import argv
from Backends.C.Files.FileTypes import C_FileType
import optparse
import argparse
import config
import os

#from termcolor import colored
class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'

    def disable(self):
        self.HEADER = ''
        self.OKBLUE = ''
        self.OKGREEN = ''
        self.WARNING = ''
        self.FAIL = ''
        self.ENDC = ''

def colored(string, color):
    _color = ''
    if color == 'blue':
        _color = bcolors.OKBLUE
    if color == 'green':
        _color = bcolors.OKGREEN
    elif color == 'red':
        _color = bcolors.FAIL
    return _color + string + bcolors.ENDC


import sys
sys.setcheckinterval(1000000000)

profile = False

aparser = argparse.ArgumentParser(description='OpenAcc to Frangollo Driver')
aparser.add_argument('-I', action="append", dest="IPATH", help = "Path to additional includes")
aparser.add_argument('-L', action="store", dest="LPATH", help = "Path to additional libraries (not used)")
aparser.add_argument('-c', action="store", dest="CNAME", help = "Name of the cluster which we are generating code")
aparser.add_argument('-x', action="store", dest="XMLFILE", help = "Path and filename of the XML description file")
aparser.add_argument('-t', action="store", dest="target", help = "Target can be: <SnuCL | OpenCL CUDA>")
aparser.add_argument('source_name', action="store")
aparser.add_argument('destination_path', action="store")
aparser.add_argument('-o', action="append", dest="OTHER", help = "Other files to be included in the compilation (won't be transformed)")


options = aparser.parse_args()

from Backends.Frangollo.Mutators.Common import FrangolloMutatorError
from Backends.Frangollo.Runner import FrangolloTransformer
from Tools import SourceStorage
#from Tools.DescriptionFile.XMLParser import XMLManager

os.environ['_OPENACC']='1'

# print aparser
if os.environ.has_key('YACF_CFLAGS'):
    print os.environ['YACF_CFLAGS']

# -----------------------------------
# Includes
from Frontend.C99.c99_prepro import INCLUDEPATH as c99_INCLUDEPATH
dirs=c99_INCLUDEPATH
if options.IPATH:
    for f in options.IPATH:
	    for i in f.split(','):
        	if i != '':
	            dirs.append(i)
# -----------------------------------

filename = options.source_name
output = options.destination_path


xmlm = None
if options.XMLFILE:
    xmlm = XMLManager(options.XMLFILE)
# print " XML Option " + str(xmlm)


###################### First Layer  : File parsing

# Parse file
    
# from Frontend.Parse import parse_source, remove_includes

print colored("Parsing " + filename + " .... ", 'blue'),
# print  
original_source = " ".join(open(filename, 'r').readlines())


# CHANGE TO USE THIS
# http://docs.python.org/library/configparser.html
if config.GNU_SUPPORT:
    config.CURRENT_LANGUAGES = ['C99','GNU', 'Acc',]
else:
    config.CURRENT_LANGUAGES = ['C99', 'Acc',]
from Frontend.FrontendFactory import FrontendFactory
[ast_, prepro_class, lexer_class, parser_class,] = FrontendFactory().parse_with_language_list(original_source, filename, config.CURRENT_LANGUAGES,dirs)
# ast_.show()
ast = ast_
print colored(" OK ", 'blue')

print colored(" Migrating to Internal Representation ....", 'blue'),
from Frontend.C99.C99InternalRepr import C99AstToIR
from Backends.C99.Writers.AccWriter import AccWriter
# Transform the C ast into the internal representation
migrator = C99AstToIR(Writer = AccWriter, ast = ast_)
migrator.annotate()
new_ast = migrator.ast

print colored(" OK ", 'blue')

###################### Second Layer  : Transformation tools
print colored("Mutating ...\n",'green'),


stripped_filename = filename.split('/')[-1].split('.')[0]

try:
    end_ast = None
    if profile:
        import cProfile
        cProfile.run('end_ast = FrangolloTransformer.apply(new_ast, stripped_filename, output, xmlm = xmlm, other_files = options.OTHER)', "prof")
    else:
        end_ast = FrangolloTransformer.apply(new_ast, stripped_filename, output, xmlm = xmlm, other_files = options.OTHER)

except FrangolloMutatorError as fme:
    print colored(" Error while mutating tree ", 'red')
    print fme


################### 
# Recreate serial version in a separate storage


# print "Normal file"
st = SourceStorage.Storage('', output + "/C/")
st.addFile(stripped_filename, C_FileType())
st.append(stripped_filename, C_FileType(), open(filename,'r').read())

# Delete SourceStorage ensures file saving
del st

if end_ast:
    print colored("OK", 'green')
else:
    print colored(" No mutation applied ", 'red')
    import sys
    sys.exit(-1)

