# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
YaCF - Outline example driver

 :date: Nov 25, 2010
 :author: rreyes
 
'''

from Backends.C.Writers.CWriter import CWriter


from sys import argv, exit
from Tools import SourceStorage
from MiddleEnd.Loop.Analysis import OmpParametrize
from Backends.C.Writers.OmpWriter import OmpWriter
from Tools.Tree import getContainerAttribute, ReplaceTool
from Frontend.C import c_ast
from Backends.Common.Files.FileTypes import Makefile_FileType
from Backends.Cuda.Files.FileTypes import Cuda_FileType
from Backends.Cuda.Kernelize import cuda_outline_loop_stmts
from Backends.Cuda.Platform import build_cuda_system_parameters

import config

output = None

if len(argv) != 2:
    print " Incorrect number of parameters "
    print "Usage: c2outline.py original_source"
    print "Example: bin/c2outline.py examples/pi.c"
    exit()

if len(argv) > 1:
    filename  = argv[1]
    if len(argv) > 2 :
        output = argv[2]
    else:
        output = "Default"

else:
    print ">>> File not found!"
    exit()



###################### First Layer  : File parsing

# Parse file

def genTs(original_source, filename):
        ###################### First Layer  : File parsing
        # Parse file
        print "Parsing " + filename + " .... ",
        config.CURRENT_LANGUAGES = ['C99',]
        # original_source = " ".join(open(filename, 'r').readlines())
        from Frontend.FrontendFactory import FrontendFactory
        [ast, prepro_class, lexer_class, parser_class] = FrontendFactory().parse_with_language_list(original_source, filename, config.CURRENT_LANGUAGES)
        print " OK "
        print " Migrating to Internal Representation ....", 
        from Frontend.C99.C99InternalRepr import C99AstToIR
        # Transform the C99 ast into the intermediate representation
        migrator = C99AstToIR(Writer = CWriter, ast = ast)
        print "*************************"
        migrator.annotate()
        new_ast = migrator.ast
        print "*************************"
        print " OK "
        # Get the symbol table
        from Frontend.SymbolTable import *
        print " Build symbol table ....", 
        st = migrator.symbolTable
        print "*************************"
        return [new_ast, st]


original_source = " ".join(open(filename, 'r').readlines())
config.CURRENT_LANGUAGES = ['C99','Omp']
[new_ast, st] = genTs(original_source, filename)

print " OK "

###################### Second Layer  : Transformation tools
print "Mutating ...",

st = SourceStorage.Storage('Test','C')
stripped_filename = filename.split('/')[-1].split('.')[0]

from Backends.C.Files.FileTypes import C_FileType

st.addFile(stripped_filename, C_FileType())
st.append(stripped_filename, C_FileType(), str(new_ast))

# del st


from MiddleEnd.Functions.Mutators.Outliner import block_is_SESE, OutlineMutator

# Block statement to outline (Fixed for now, later use OmpParallelForFilter)
#                  FileAst CompoundStmt   OmpPragma OmpParallelFor 

new_ast.show()

objetive = new_ast.ext[-1].body.block_items[-1].stmt
omp_node = new_ast.ext[-1].body.block_items[-1]


#### Outline the loop contents

def outline_loop_stmts(ojetive):   
    assert block_is_SESE(objetive)
    om = OutlineMutator()
    om.apply(objetive)
    kernel_template = """
    %for typedef in typedeflist:
        ${typedef};
    %endfor
    
    extern "C" {
    %for funcs in funccalllist:
        ${funcs};
    %endfor
    
    ${outlinedProto};
    };
    
    ${outlinedFunction};
    
    """
    
    from mako.template import Template
    kernel_source = Template(kernel_template).render_unicode(typedeflist = om.getTypedefList(),
                                                             funccalllist = om.getFuncCallList(),
                                                             outlinedProto = om.getOutlinedPrototype(),
                                                             outlinedFunction = om.getOutlinedFunction())
    return kernel_source

kernel_source = outline_loop_stmts(objetive)


print str(kernel_source)
    
#### Kernelize the loop

#===============================================================================
# 
# 
# 
# st = SourceStorage.Storage('Test','Cuda')
# 
# 
# loop_parameters = OmpParametrize().apply(objetive.parent,)
# system_parameters = build_cuda_system_parameters(loop_parameters,)
#===============================================================================

#===============================================================================
# 
#                           
# print "#########################################"                      
# print "Extracted Loop Parameters"
# 
# for elem in loop_parameters:
#    if type(loop_parameters[elem]) == type([]):
#        print str(elem) + " : " + str([str(e) for e in loop_parameters[elem]])
#    else:
#        print str(elem) + " : " + str(loop_parameters[elem])
# 
# print "#########################################"
# print "Inferred system Parameters"
# 
# for elem in system_parameters:
#    if type(system_parameters[elem]) == type([]):
#        print str(elem) + " : " + str([str(e) for e in system_parameters[elem]])
#    else:
#        print str(elem) + " : " + str(system_parameters[elem])
#===============================================================================


# print "(1) Kernel(s) have been written"

#### Final steps

st.addFile('main', C_FileType())
st.append('main', C_FileType(), str(new_ast))


del st   # Ensure file writing

#===============================================================================

# loop_parameters = OmpParametrize().apply(objetive.parent,)
# system_parameters = build_opencl_system_parameters(loop_parameters,)
# st = SourceStorage.Storage('Test','OpenCL')
# kernel_source = opencl_outline_loop_stmts(objetive)
# st.addFile('kernel',  SourceStorage.OPENCL_KERNEL_SOURCE)
# st.append('kernel', SourceStorage.OPENCL_KERNEL_SOURCE, kernel_source)
#===============================================================================


if __name__ == '__main__':
    pass



#===============================================================================
    # print "*************************************************"
    # print "Result"
    # print objetive.parent
    # print "*************************************************"
    # print "Outlined function"
    # print om.getOutlinedFunction()
    # print "*************************************************"
    # print "Outlined Prototype"
    # print om.getOutlinedPrototype()
    # print "*************************************************"
    # print "Required Typedefs"
    # print om.getTypedefList()
    # print "*************************************************"
    # print "Functions called inside kernel"
    # print om.getFuncCallList()
    # print "*************************************************"
    #===============================================================================

