# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

# from Visitors.clone_visitor import CWriter, OmpWriter
from Backends.C99.Writers.OmpWriter import CWriter
from sys import argv, exit

import config


def genTs(original_source, filename):
        ###################### First Layer  : File parsing
        # Parse file
        print "Parsing " + filename + " .... ",
        config.CURRENT_LANGUAGES = ['C99','Acc']
        # original_source = " ".join(open(filename, 'r').readlines())
        from Frontend.FrontendFactory import FrontendFactory
        [ast, prepro_class, lexer_class, parser_class] = FrontendFactory().parse_with_language_list(original_source, filename, config.CURRENT_LANGUAGES)
        print " OK "
        print " Migrating to Internal Representation ....", 
        from Frontend.C99.C99InternalRepr import C99AstToIR
        # Transform the C99 ast into the intermediate representation
        migrator = C99AstToIR(Writer = CWriter, ast = ast)
        print "*************************"
        migrator.annotate()
        new_ast = migrator.ast
        print "*************************"
        print " OK "
        # Get the symbol table
        from Frontend.SymbolTable import *
        print " Build symbol table ....", 
        st = migrator.symbolTable
        print "*************************"
        return [new_ast, st]


output_file = None


struct_test = """
  struct complex {
      double creal;
      double cimag;
  };


typedef struct {
	double creal;
	double cimag;
} complejo;

int main (int b) {

 struct  complex c;
 complejo h;
 int a;

  a = 3;
  c.cimag = 1.0;
  c.creal = 1.0;
}
"""


[new_ast, st] = genTs(struct_test, filename = 'struct_test')

print "****> TS for Struct Test "
print str(st)
print "****> Lookup check "
print str(st.lookUp(new_ast.ext[-1].body.block_items[-1].lvalue))
print str(st.lookUp(new_ast.ext[-1].body.block_items[-2].lvalue))
print str(st.lookUp(new_ast.ext[-1].body.block_items[-3].lvalue))

print " ******* FUNCTION ***** "

array_test = """
double * array;

void testfunc (double array[10][10]) {
  int i, j;


  for (i = 0; i < 10; i++)
    for (j = 0; i < 10; i++) 
     array[i][j] = -1.0;

}

int main () {
  double array[10][20];
  int i, j;
  array[i][j] = 3.0;
  {
  for (i = 1; i < 10; i++) {
    for (j = i; j < 10; j++) 
     array[i][j] = -1.0;
    
  }
  }

  testfunc(array);

}

void testfunc (double array[10][10]) {
  int i, j;


  for (i = 0; i < 10; i++)
    for (j = 0; i < 10; i++) 
     array[i][j] = -1.0;

}

"""

[new_ast, st] = genTs(array_test, filename = 'array_test')


print "****> TS for Array Test "
print str(st)

print "****> Check validation "
first_array_access = new_ast.ext[-2].body.block_items[3].lvalue
print str(st.lookUp(first_array_access))

print "****> Check validation "
parameter_array_access = new_ast.ext[1].body.block_items[-1].stmt.stmt.lvalue
print str(st.lookUp(parameter_array_access))

print "****> Check duplicated Id "
loop_var = new_ast.ext[-1].body.block_items[-1].init.lvalue
print str(st.lookUp(loop_var))



structarray_test = """

struct complex {
   double c;
   double t;
} array[10][20];

int main () {
   array[10][20] = 3;
}
"""

[new_ast, st] = genTs(structarray_test, filename = 'structarray_test')


print "****> TS for Struct - Array Test "
print str(st)


print str(st.lookUp(new_ast.ext[-1].body.block_items[-1].lvalue))



performance_test = """

int a,b,c,d,e,f,g;
double h,i,j,k,l;
int * w, * x, *y;

int main() {
    {
    int h, i, j, k, l;
    w[3] = i;
    i = 0;
    j = 0;
    k = 0; 
    
    a = 0;
    b = 0;
    c = 0; 
    
    d = 0;
    e = 0;
    f = 0; 
    }
}


"""
[new_ast, st] = genTs(performance_test, filename = 'performance_test')

print "****> TS for Struct - Performance Test "
print str(st)


symbols = []
for i in range(5,15):
    symbols.append (st.lookUp(new_ast.ext[-1].body.block_items[0].block_items[i].lvalue))

print " List " 
print [str(elem) for elem in symbols]



structarray_test = """
typedef int tutu;


int main () {
   tutu i;
   float t;
   t = 3;
   i = 1;   
   
}
"""

[new_ast, st] = genTs(structarray_test, filename = 'structarray_test')


print "****> TS for Typedef Test "
print str(st)

symbol = st.lookUp(new_ast.ext[-1].body.block_items[-1].lvalue)
tsymbol = st.lookUpTypeDecl(symbol.type)
print str(tsymbol)
tsymbol.type.show()
print "Btype " + str(tsymbol.btype)


symbol = st.lookUp(new_ast.ext[-1].body.block_items[-2].lvalue)
symbol.node.show()
print str(symbol)

# Test for structs
struct_complex_test = """
  struct complex {
      double creal;
      double cimag;
  };

  int main () {
      struct complex c;
  }
"""
[new_ast, st] = genTs(struct_complex_test, filename = 'struct_complex_test')

print "****> TS for Struct Complex Test "
print str(st)
    
print "****> Lookup check "
symbol = st.lookUp(new_ast.ext[-1].body.block_items[0])
tsymbol = st.lookUpTypeDecl(symbol.type)
print str(tsymbol)
tsymbol.type.show()
print "Btype " + str(tsymbol.btype)



techrep_test = """
char * foo;
int func () {
   int i,j;
   {
      char * foo;
      int i;
      printf("%d",i);
   }
}

int main() {
   int i;
   foo = NULL;
   func();
   printf("%d",i);
   if (foo)  printf("%s", foo);
}
"""

[new_ast, st] = genTs(techrep_test, filename = 'techrep_test')

print "****> TS for TechRep Test "
print str(st)
 

