# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

'''
    Created 24-06-2011
    @autor jlucasgl
'''

from sys import argv, exc_info, exit
from Tools.DescriptionFile.XMLParser import XMLManager


def showstructure(xmlm):

    # Get tree and root node from xml
    # 
    xmltree = xmlm.xmltree
    root = xmltree.getroot()
    
    # Show in screen the whole parsed xml file 
    # 
    print xmlm
    print

    # Show XMLManager('.xml').nex_cluster() behavior 
    #
    '''print "First 4 calls to XMLManager('.xml').next_cluster() give us:"
    print xmlm.next_cluster()
    print xmlm.next_cluster()
    print xmlm.next_cluster()
    print xmlm.next_cluster()
    xmlm.reset_iter_cluster()
    print'''

    # Structure of the whole XML file 
    #
    print "The following clusters are defined in the <" + xmlm.xmlfilename + ">:"

    while True:
        print "Cluster " + xmlm.cluster.get("name") + "."
        while (xmlm.next_region() is not None):
            print "\tRegion " + xmlm.region.get("name")
            while (xmlm.next_element() is not None):
                print "\t\tElement " + xmlm.element.get("class") + " " + xmlm.element.get("name")
                while (xmlm.next_transform() is not None):
                    print "\t\t\tMutator " + xmlm.transform.get("name") + " (" + xmlm.transform.text + ")"
        if (xmlm.next_cluster() is None):
            break
    print

    # First transform in a given cluster 
    #   - find the cluster and explore the tree
    '''otrocluster = xmlm.find_cluster("otrocluster")
    xmlm.next_region()
    xmlm.next_element()
    transform = xmlm.next_transform()
    print "First transform in Cluster 'otrocluster': " + transform.get("name") + " (" + transform.text + ")"
    print'''

    print "The following mutators are used in the <" + xmlm.xmlfilename + ">:"
    for transform in xmltree.findall("//mutator"):
        print "Mutator " + transform.get("name") + " (" + str(transform.text) + ")"
    print



from Frontend.Shortcuts import getCurrentLanguageAst
c_ast = getCurrentLanguageAst()


# from Visitors.clone_visitor import CWriter, OmpWriter
from Backends.C99.Writers.OmpWriter import CWriter

from sys import argv, exit

import config


def genTs(original_source, filename):
        ###################### First Layer  : File parsing
        # Parse file
        print "Parsing " + filename + " .... ",
        config.CURRENT_LANGUAGES = ['C99','Omp']
        # original_source = " ".join(open(filename, 'r').readlines())
        from Frontend.FrontendFactory import FrontendFactory
        [ast, prepro_class, lexer_class, parser_class] = FrontendFactory().parse_with_language_list(original_source, filename, config.CURRENT_LANGUAGES)
        print " OK "
        print " Migrating to Internal Representation ...."
        from Frontend.C99.C99InternalRepr import C99AstToIR
        # Transform the C99 ast into the intermediate representation
        migrator = C99AstToIR(Writer = CWriter, ast = ast)
        print "*************************"
        migrator.annotate()
        new_ast = migrator.ast
        print "*************************"
        print " OK "
        # Get the symbol table
        # from Frontend.SymbolTable import *
        print " Build symbol table ...."
        st = migrator.symbolTable
        print "*************************"
        return [new_ast, st]



#===============================================================================
# Parse XML <xmlfilename>


xmlfilename = config.WORKDIR + 'examples/xml/test_03.xml'
xmlm = XMLManager(xmlfilename)
#===============================================================================

#===============================================================================
# Prints some info about the XML structure
showstructure(xmlm)
#===============================================================================

#===============================================================================
# Read original source and generate simbol table and AST
output_file = None

if len(argv) > 1:
    filename  = argv[1]
    if len(argv) > 2 :
        output_file = argv[2]
else:
    print ">>> File not found!"
    exit()

original_source = " ".join(open(filename, 'r').readlines())

[new_ast, st] = genTs(original_source, filename)
#===============================================================================

# print "====================================================="
# print str(new_ast)
# print "====================================================="


from Tools.DescriptionFile.XMLTranslationMutator import XMLTranslationMutator

#=================================================================================
# Apply XMLTranslationMutator
#=================================================================================
XMLTranslationMutator(xmlm).apply_all(new_ast)



#=================================================================================
# Write file

# Call pretty printer over the file
if output_file:
    v = CWriter(filename = output_file)
    v.visit(new_ast)
    del v  # Ensure file closing
    import os
    if os.system("indent -kr " + output_file) != 0:
        print " You need to install the indent tool to pretty print ouput files "
else:
    # new_ast.show(attrnames = True)
    print "************"
    print "====================================================="
    print str(new_ast)
    print "====================================================="
#=================================================================================

