# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------------
# Copyright (c) 2012, Francisco de Sande González. 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this 
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, 
#   this list of conditions and the following disclaimer in the documentation 
#   and/or other materials provided with the distribution.
# * Neither the name of Francisco de Sande nor the names of its contributors may 
#   be used to endorse or promote products derived from this software without 
#   specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
# --------------------------------------------------------------------------------

# from Visitors.clone_visitor import CWriter, OmpWriter
from Backends.C99.Writers.OmpWriter import OmpWriter
from Frontend.Common.IRNode import IRNodeVisitor
from Frontend.SymbolTable import getSymbolTableFromNode
from Tools.Tree import getFileNode

from MiddleEnd.DataAnalysis.LlcScope import DGraphBuilder, BlockDepGraphBuilder
# Get the symbol table
from Frontend.SymbolTable import SymbolTable
from Frontend.C99.C99InternalRepr import C99AstToIR
from Frontend.FrontendFactory import FrontendFactory

from sys import argv, exit

import config


def genTs(original_source, filename):
    ###################### First Layer  : File parsing
    # Parse file
    print "Parsing " + filename + " .... ",
    config.CURRENT_LANGUAGES = ['C99','Omp']
    # original_source = " ".join(open(filename, 'r').readlines())
    from Frontend.FrontendFactory import FrontendFactory
    ast =FrontendFactory().parse_with_language_list(original_source, filename, config.CURRENT_LANGUAGES)[0]
    print " OK "
    print " Migrating to Internal Representation ....", 
    # Transform the C99 ast into the intermediate representation
    migrator = C99AstToIR(Writer = OmpWriter, ast = ast)
    print "*************************"
    migrator.annotate()
    ast = migrator.ast
    print "*************************"
    print " OK "
    print " Build symbol table ....", 
    _st = migrator.symbolTable
    print "*************************"
    return [ast, _st]


output_file = None

pi_test = '''
// #include <stdio.h>
#define N_ELEM 5120000
// Main
int main ()
{
  int i;
  int n = N_ELEM;
  float pi, sum;
  // Initial value to a global var
  #pragma llc context name("compute_pi") copy_out(pi, sum) copy_in(sum, n)
 {
  // This could be run in the device 
  // Create Var statement is added for this variable 
  float h = 1.0 / (float) n;   // local var to the device
  sum = h + 1;   
  #pragma llc for shared(h, n) reduction(+ : sum)
  for (i = 0; i <= n; i ++)
    {
      float x = h * ((float) i - 0.5);
      sum += 4.0 / (1.0 + x * x);
    }
   pi = h * sum;
  }
  
  printf("Pi %f \\n", pi);

}

'''
 # printf ("Pi %f \n", pi);

[new_ast, st] = genTs(pi_test, filename = 'pi_test')

loop = new_ast.ext[-1].body.block_items[-2] # .ext[-1].body.block_items[-3]

from MiddleEnd.DataAnalysis.LlcScope import DNode,DGraph


dGraph = DGraph()
st = getSymbolTableFromNode(getFileNode(loop))
DGraphBuilder(dGraph, st).visit(loop)


sumSymbol = st.lookUp(new_ast.ext[-1].body.block_items[-2].stmt.stmt.block_items[-1].rvalue.right)
piSymbol = st.lookUp(new_ast.ext[-1].body.block_items[-2].stmt.stmt.block_items[-1].lvalue)


print "Test dependency bw piSymbol (" + str(piSymbol.name) + ") and sumSymbol (" + str(sumSymbol.name) + ")"
if dGraph.checkDependency(dGraph[piSymbol], dGraph[sumSymbol]):
    print " It works "
else:
    print " FAIL "
 

print "Show dependency Graph "
from Tools.Debug import DGraphDebug
DGraphDebug().apply(dGraph)


#################### Build BLOCK structure


llcContext_subtree = new_ast.ext[-1].body.block_items[-2].stmt.stmt



#print [str(elem[0]) + "," for elem in bdgb.blockRead]
dGraph = DGraph()
bdgb = BlockDepGraphBuilder(dGraph, getSymbolTableFromNode(llcContext_subtree))
bdgb.visit(llcContext_subtree)
[predStmt, requiredVars] = bdgb.computeDependencyMatrix()

# Show dependencies among stmt
for i in range(0,len(llcContext_subtree.block_items)):
    print "S" + str(i) + " = " + str(predStmt[i])

for i in range(0,len(llcContext_subtree.block_items)):
    print "S" + str(i) + " = " + str([str(elem) for elem in requiredVars[i]])

#print str(new_ast)




