# -*- coding: utf-8 -*-
# Config file for llcomp

# Current work dir
import os
WORKDIR=os.environ['PYTHONPATH']

# Location of cuda files
CUDA_INSTALL_DIR="/usr/local/cuda/"

###############################################
#            Frontend variables               #
###############################################
LEX_OPTIMIZE=False
YACC_OPTIMIZE=False
YACC_DEBUG =True
GNU_SUPPORT=False
### STS_ONLY
# If True, some syntax rules are relaxed. 
#  This enables partial snippets to be parsed e.g without
#  having the type definitions. 
# This is only valid if the destination of the transformation is another
#  source file that will be parsed by a full compiler, as some errors
#  won't be checked.
STS_ONLY=True   
###############################################
#           Preprocessor variables            #
###############################################
SHOW_FILE_AFTER_PREPRO=False
SHOW_LINES=False
COMMENT_INCLUDES = True
MULTIPLE_FILE_PREPRO=False
INSERT_LOCAL_INCLUDES=True
################################################

# Enable/Disable internal hacks (like overrides, macros or highly experimental features)
# Not recommended for sensitive people
INTERNAL_HACKS=False
# Generate Frangollo calls with extra debug information
FRANGOLLO_DEBUG=False
# Generation of Host code, this code will be active if no GPU is avail. 
#### This is not completely tested, default is False
GENERATE_HOST_CODE=False

# Use the new JJ processor
MULTIPLE_FILE_PREPRO=False

# ACC directives should ignore scope and create the registersVars only for variables directly registered on this directive, and ignore the previous variables
IGNORE_SCOPE=False 

# Additional nvidia compiler options
nv_opts = ""
# List of kernel optimizations to apply, currently supported "invariant"
kernel_optimizations = []

MKFILE=None

# Import local settings from user
from config_local import *
INSTALLED_BACKENDS = [ 'Cuda', 'DotBackend', 'C', 'MPI', 'Frangollo' ]

if not MKFILE:
   MK_FILE=WORKDIR + '/makefile.mk'

FRONTEND_DIR = WORKDIR + '/Frontend'
FRONTEND_MODULE = "Frontend"

# C includes for templates
# TODO: Automatically detect include dirs
INCLUDE_DIR = WORKDIR + '/Backends/Frangollo/Templates/include/'
# Fake stdlib
FAKE_LIBC = WORKDIR + '/Backends/Cuda/Templates/include/fake_libc_include'

# This variable is used by each driver to set the current used languages across all modules
CURRENT_LANGUAGES = ['C99', 'C']

PREPROCESSOR="gcc -E"


DEFAULT_INCLUDES='#include <stdio.h>\n #include <stdlib.h> \n #include <stdint.h> \n #include <assert.h> \n #include <sys/time.h> \n #include <time.h> \n #include <omp.h> \n '

#### CUDA Backend config
CUDA_KERNEL_HEADER_TEMPLATE_FILE= WORKDIR + '/Backends/Cuda/Templates/kernel_header.tmpl.c'

#### OpenCL Backend config
CL_KERNEL_HEADER_TEMPLATE_FILE= WORKDIR + '/Backends/OpenCl/Templates/kernel_header.tmpl.c'


