
#include "llc.h"


# line 9 "/home/ray/Proyecto/yaCF-MPI-src/examples/MPI_examples/NBODY/NBODY-LLCOMP/nbody_llcomp.c" 
#include "nbody.h"
Body *init_bodies (unsigned int num_bodies, int init_type) {
  int i;
  double n = num_bodies;
  Body *bodies = (Body *) malloc (num_bodies * sizeof (Body));
  
  for (i = 0; i < num_bodies; i++) {
    switch (init_type) {
      case INIT_LINEAR : bodies[i].mass = 1.0;
      bodies[i].pos[X] = i / n;
      bodies[i].pos[Y] = i / n;
      bodies[i].vel[X] = 0.0;
      bodies[i].vel[Y] = 0.0;
      break;
      case INIT_SPIRAL : bodies[i].mass = (n - i) / n;
      bodies[i].pos[X] = (1 + i / n) * cos (2 * M_PI * i / n) / 2;
      bodies[i].pos[Y] = (1 + i / n) * sin (2 * M_PI * i / n) / 2;
      bodies[i].vel[X] = 0.0;
      bodies[i].vel[Y] = 0.0;
      break;
    }
  }
  return bodies;
}



# line 45 "/home/ray/Proyecto/yaCF-MPI-src/examples/MPI_examples/NBODY/NBODY-LLCOMP/nbody_llcomp.c" 
int check_simulation (Body *bodies, int num_bodies) {
  int i, check_ok;
  double momentum [2] = {
    0.0, 0.0
  };
  
  for (i = 0; i < num_bodies; i++) {
    momentum[X] += bodies[i].mass * bodies[i].vel[X];
    momentum[Y] += bodies[i].mass * bodies[i].vel[Y];
  }
  check_ok = ((abs (momentum[X]) < EPSILON) && (abs (momentum[Y]) < EPSILON));
  if (!check_ok)
    printf ("Warning: total momentum = (%3.3f, %3.3f)\n", momentum[X], momentum[Y]);
  
  return check_ok;
}



# line 61 "/home/ray/Proyecto/yaCF-MPI-src/examples/MPI_examples/NBODY/NBODY-LLCOMP/nbody_llcomp.c" 

# line 1 "LLC code (MAIN)" 

	/* LLC Autogenerated main */

int main (int argc, char **argv) {
	int llc_return_code = 0;

	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, (int*)&LLC_GLOBAL_NAME);
	MPI_Comm_size(MPI_COMM_WORLD, (int*)&LLC_GLOBAL_NUMPROCESSORS);
	LLC_NAME = LLC_GLOBAL_NAME;
	LLC_NUMPROCESSORS = LLC_GLOBAL_NUMPROCESSORS;
	llc_GlobalGroup = MPI_COMM_WORLD;
	llc_CurrentGroup = &llc_GlobalGroup;

	llc_return_code = llc_main (argc, argv);

	MPI_Finalize();
	return llc_return_code;
}




# line 61 "/home/ray/Proyecto/yaCF-MPI-src/examples/MPI_examples/NBODY/NBODY-LLCOMP/nbody_llcomp.c" 
int llc_main (int argc, char **argv) {
  struct timeval start_time, end_time;
  double time;
  int num_bodies, num_steps;
  int i, j, k;
  double dt = 1.0, dv [2];
  double r [2], dist, force_len, force_ij [2], tot_force_i [2];
  Body *bodies;
  const char Usage [] = "Usage: nbody <num bodies> <num_steps>\n";
  
  if (argc < 2) {
    fprintf (stderr, Usage);
    LLC_EXIT (1);
  }
  num_bodies = atoi (argv[1]);
  num_steps = atoi (argv[2]);
  bodies = init_bodies (num_bodies, INIT_SPIRAL);
  check_simulation (bodies, num_bodies);
  gettimeofday (&start_time, 0);
#define Calc_Force_ij() \
  r[X] = bodies[j].pos[X] - bodies[i].pos[X]; \
  r[Y] = bodies[j].pos[Y] - bodies[i].pos[Y]; \
  dist = r[X] * r[X] + r[Y] * r[Y]; \
  force_len = GRAV_CONST * bodies[i].mass * bodies[j].mass / (dist * sqrt(dist)); \
  force_ij[X] = force_len * r[X]; \
  force_ij[Y] = force_len * r[Y]
#define Step_Body_i() \
  dv[X] = dt * tot_force_i[X] / bodies[i].mass; \
  dv[Y] = dt * tot_force_i[Y] / bodies[i].mass; \
  bodies[i].pos[X] += dt * ( bodies[i].vel[X] + dv[X] / 2 ); \
  bodies[i].pos[Y] += dt * ( bodies[i].vel[Y] + dv[Y] / 2 ); \
  bodies[i].vel[X] += dt * dv[X]; \
  bodies[i].vel[Y] += dt * dv[Y]
  for (k = 0; k < num_steps; k++) {
    
# line 1 "LLC code (/src/parser/llc_src/forall/forall_init_unit.src)" 
			/*  START CODE FROM '/home/ray/Proyecto/yaCF-MPI-doc/src/llCoMP_Antonio/llCoMP/src/parser/llc_src/forall/forall_init_unit.src'  */

		{	
		  /* i => Variable del bucle
		   * 0 => Valor inicial de la variable del bucle
		   * ((num_bodies) - 1) => Valor final de la variable del bucle
		   * 1 => Peso(s)
			 * LLC_NO_WEIGHTS => Flags de asignacin equitativa por pesos (LLC_WEIGHTS/LLC_NO_WEIGHTS)
			 */
		  if(LLC_NUMPROCESSORS == 1) {
		    llc_case = LLC_SEQUENTIAL;
		  }
		  else {
				if ((llc_nT = ((((num_bodies) - 1)) - (0) + 1)) >= LLC_NUMPROCESSORS) {
		  	 	register int llc_i, llc_j, llc_k;
					int llc_rnT, llc_w, llc_mW;

		  		llc_case = LLC_BLOCK;

					if ((LLC_NO_WEIGHTS) == LLC_WEIGHTS) {

								/* begin: LLC_WEIGHTS */
			    /* Obtain medium weight (only weight > 0) */
		  	    for(llc_mW = 0, i = 0; i <= ((num_bodies) - 1); i++) {
					  	if ((1) > 0) {
								llc_mW += (1);
							}
						}
						llc_mW /= LLC_NUMPROCESSORS;

			   		LLC_CHECK_MALLOC (llc_F, (LLC_NUMPROCESSORS + 1), int);
						llc_F[0] = 0;
						llc_rnT = llc_nT - LLC_NUMPROCESSORS;
			
						/* Weighted task assignement */
						i = 0;
		  	 		for(llc_i = 0; llc_i < LLC_NUMPROCESSORS; llc_i++) {
							llc_F[llc_i + 1] = llc_F[llc_i] + 1;
							llc_w = 1;
							i++;
							while ((llc_rnT > 0) && (llc_w < (llc_mW + LLC_WEIGHT_TOL))) {
								llc_F[llc_i + 1]++;
								llc_w += 1;
								i++;
								llc_rnT--;
							}
						}

						llc_i = 0;			/* Unassigned tasks: Cyclic assignment */
						while (llc_rnT > 0) {
							llc_F[(llc_i % LLC_NUMPROCESSORS) + 1]++;
							llc_rnT--;
							llc_i++;
						}
			 		}  			/* end: LLC_WEIHGTS */
					else {
									/* begin: LLC_NO_WEIGHTS */
						register int llc_i;
						/* Medium number of tasks */
						int llc_meanT = llc_nT / LLC_NUMPROCESSORS;
						/* Number of unassigned tasks */
						int llc_rnT = llc_nT % LLC_NUMPROCESSORS;
			   		
						LLC_CHECK_MALLOC (llc_F, (LLC_NUMPROCESSORS + 1), int);
						
						llc_F[0] = 0;
						for (llc_i = 1; llc_i < (LLC_NUMPROCESSORS + 1); llc_i++) { 
							llc_F[llc_i] = llc_F[llc_i-1] + llc_meanT + (((llc_i - 1) < llc_rnT) ? 1 : 0);
						}
					} 			/* end: LLC_NO_WEIGHTS */
					
				} /* LLC_BLOCK */
		  	else if(llc_nT > 0) {
					llc_case = LLC_GROUPS;
					
					if ((LLC_NO_WEIGHTS) == LLC_WEIGHTS) {
		  	 		register int llc_i, llc_j, llc_k;
		 	 			int *llc_ptr;
						int *llc_W, llc_nW;
		  			int *llc_P, llc_nP;
		  			int *llc_G;
		 
								/* begin: LLC_WEIGHTS */


		    	/* ------------------------------------------------------------
		      	 Weights processing: Threads with non-possitive weight are
			       discarded
		  	     ------------------------------------------------------------ */

						LLC_CHECK_MALLOC (llc_ptr, 3 * llc_nT, int);
						llc_W = llc_ptr;
						llc_P = llc_ptr + llc_nT;
						llc_G = llc_ptr + (2 * llc_nT);

				  	llc_k = llc_nP = LLC_NUMPROCESSORS - llc_nT + 1;
		    		for(llc_nW = 0, llc_i = 0, i = (0); i <= (((num_bodies) - 1)); i++, llc_i++) {
			      	if((llc_W[llc_i] = (1)) > 0) {
		  	      	llc_nW += llc_W[llc_i];
		    	    	if(llc_W[llc_i] < llc_k) llc_k = llc_W[llc_i];
		      		}
			      	else
		  	      	llc_W[llc_i] = 0;
		    		}

			    	/* Weights normalization */
			
		  	  	if(llc_k < llc_nP)
		    	  	for(llc_i = 0; llc_i < llc_nT; llc_i++) {
		      		 	llc_W[llc_i] *= llc_nP;
			        	llc_W[llc_i] /= llc_k;
		  	    	}

			    	/* ------------------------------------------------------------
		  	     	Processors assigment policy: BLOCK
		    	   	At least one processor by thread
		      	 	------------------------------------------------------------ */

			   		for(llc_i = 0; llc_i < llc_nT; llc_i++) {
		  	    	llc_P[llc_i] = 1;

			      	llc_j = llc_i;
		  	    	while((llc_j > 0) && (llc_W[llc_i] > llc_W[llc_G[llc_j - 1]])) {
		    	    	llc_G[llc_j] = llc_G[llc_j - 1];
		      	  	llc_j--;
			      	}
		  	    	llc_G[llc_j] = llc_i;
		    		}

			    	llc_nP = llc_nT;
		  	  	while(llc_nP++ < LLC_NUMPROCESSORS) {
		    	  	llc_W[llc_G[0]] = (llc_W[llc_G[0]] * llc_P[llc_G[0]]) /
		      	                (llc_P[llc_G[0]] + 1);
		      		llc_P[llc_G[0]]++;
			      	llc_k = llc_G[0];
		  	    	llc_j = 1;
		    	  	while((llc_j < llc_nT) && (llc_W[llc_G[llc_j]] > llc_W[llc_G[llc_j-1]])) {
		      	  	llc_G[llc_j-1] = llc_G[llc_j];
			        	llc_j++;
		  	    	}
		    	  	llc_G[llc_j-1] = llc_k;
			    	}

		  	  	/* ------------------------------------------------------------
		    	  	 Groups creation
			    	   ------------------------------------------------------------ */

						LLC_CHECK_MALLOC (llc_F, (llc_nT + 1), int);
		  	  	
						llc_F[0] = 0;
		    		for(llc_i = 1; llc_i <= llc_nT; llc_i++)
			      	llc_F[llc_i] = llc_F[llc_i-1] + llc_P[llc_i-1];
		  	                                                    		

		  	  	/* ------------------------------------------------------------
								Auxiliar variables deallocation
			    	   ------------------------------------------------------------ */

			    	free(llc_ptr);
			 		}  			/* end: LLC_WEIHGTS */
					else {
									/* begin: LLC_NO_WEIGHTS */

						register int llc_i;
						/* Medium number of processors */
						int llc_meanP = LLC_NUMPROCESSORS / llc_nT;
						/* Number of unassigned processors */
						int llc_rnP = LLC_NUMPROCESSORS % llc_nT;
			   		
						LLC_CHECK_MALLOC (llc_F, (llc_nT + 1), int);
						
						llc_F[0] = 0;
						for (llc_i = 1; llc_i < (llc_nT + 1); llc_i++) { 	
							llc_F[llc_i] = llc_F[llc_i-1] + llc_meanP + (((llc_i - 1) < llc_rnP) ? 1 : 0);
			
						}
			
					} 			/* end: LLC_NO_WEIGHTS */

		  	  /* Get the group name */
		  	  llc_grp = 0;
		 	  	while(LLC_NAME >= llc_F[llc_grp+1]) llc_grp++;
					
				} /* LLC_GROUPS */
		  	else llc_case = LLC_ERROR_CASE;
			}
		}




			/*  END CODE FROM '/home/ray/Proyecto/yaCF-MPI-doc/src/llCoMP_Antonio/llCoMP/src/parser/llc_src/forall/forall_init_unit.src'  */


# line 136 "/home/ray/Proyecto/yaCF-MPI-src/examples/MPI_examples/NBODY/NBODY-LLCOMP/nbody_llcomp.c" 

# line 1 "LLC code (/src/parser/llc_src/forall/forall_exec_unit.src)" 
			/*  START CODE FROM '/home/ray/Proyecto/yaCF-MPI-doc/src/llCoMP_Antonio/llCoMP/src/parser/llc_src/forall/forall_exec_unit.src'  */

		{
			/* i => Variable del bucle
			 * 0 => Valor inicial de la variable del bucle
			 * ((num_bodies) - 1) => Valor final de la variable del bucle
			 * dolar3 => Bloque del bucle
			 */
			
		  int llc_case_save = llc_case;

		  switch(llc_case) {
		    case LLC_SEQUENTIAL:
		      for(i = (0); i <= (((num_bodies) - 1)); (i)++)
		         {
      tot_force_i[X] = 0.0;
      tot_force_i[Y] = 0.0;
      for (j = 0; j < num_bodies; j++) {
        if (j == i)
          continue;
        
        Calc_Force_ij ();
        tot_force_i[X] += force_ij[X];
        tot_force_i[Y] += force_ij[Y];
      }
      Step_Body_i ();
    };
		    break;
		    case LLC_BLOCK:
					{
					  int llc_nP;
						register int llc_i;
					  int llc_nT_save;
					  int llc_grp_save;
						llc_var_reg_data llc_nc_result_list_save;

		 	     	llc_nP = LLC_NUMPROCESSORS;
						llc_grp_save = LLC_NAME;
			      LLC_NUMPROCESSORS = 1;
		  	    LLC_NAME = 0;
		/*				llc_nc_result_list_save_copy (llc_nc_result_list_save); */

						i = (0) + llc_F[llc_grp_save];
						for(llc_i = 0; llc_i < LLC_PROCSGRP(llc_grp_save); llc_i++, i++) {
							 {
      tot_force_i[X] = 0.0;
      tot_force_i[Y] = 0.0;
      for (j = 0; j < num_bodies; j++) {
        if (j == i)
          continue;
        
        Calc_Force_ij ();
        tot_force_i[X] += force_ij[X];
        tot_force_i[Y] += force_ij[Y];
      }
      Step_Body_i ();
    };
						}
						
		/*	      if (llc_nc_result_list.num_var > 0) {
		        	llc_nc_result_destroy_var_list();
						}
		        llc_nc_result_list = llc_nc_result_list_save; */
		  	   	LLC_NAME = llc_grp_save;
		    	 	LLC_NUMPROCESSORS = llc_nP;
					}
			    break;

		    case LLC_GROUPS:
					{	
						MPI_Comm *llc_AuxGroup, llc_SubGroup;
					  int llc_nT_save;
					  int llc_grp_save;
					  int *llc_F_save;
						llc_var_reg_data llc_nc_result_list_save;


			      LLC_NAME -= llc_F[llc_grp];
		  	    LLC_NUMPROCESSORS = llc_F[llc_grp+1] - llc_F[llc_grp];
			      MPI_Comm_split(*llc_CurrentGroup, llc_grp, LLC_NAME, &llc_SubGroup);
		  	    llc_AuxGroup = llc_CurrentGroup;
		    	  llc_CurrentGroup = &llc_SubGroup;
		      	llc_nT_save = llc_nT;
			      llc_grp_save = llc_grp;
		  	    llc_F_save = llc_F;
		/*				llc_nc_result_list_save_copy (llc_nc_result_list_save); */
		        

		    	  i = (0) + llc_grp;
						 {
      tot_force_i[X] = 0.0;
      tot_force_i[Y] = 0.0;
      for (j = 0; j < num_bodies; j++) {
        if (j == i)
          continue;
        
        Calc_Force_ij ();
        tot_force_i[X] += force_ij[X];
        tot_force_i[Y] += force_ij[Y];
      }
      Step_Body_i ();
    };

		/*        if (llc_nc_result_list.num_var > 0) {
		        	llc_nc_result_destroy_var_list();
						}
		        llc_nc_result_list = llc_nc_result_list_save; */
			      llc_F = llc_F_save;
		  	    llc_grp = llc_grp_save;
		    	  llc_nT = llc_nT_save;
			      LLC_NUMPROCESSORS = llc_F[llc_nT];
		  	    LLC_NAME += llc_F[llc_grp];
		    	  llc_CurrentGroup = llc_AuxGroup;
		      	MPI_Comm_free(&llc_SubGroup);
					}
			    break;
		  	}
		  llc_case = llc_case_save;
		}



			/*  END CODE FROM '/home/ray/Proyecto/yaCF-MPI-doc/src/llCoMP_Antonio/llCoMP/src/parser/llc_src/forall/forall_exec_unit.src'  */


# line 136 "/home/ray/Proyecto/yaCF-MPI-src/examples/MPI_examples/NBODY/NBODY-LLCOMP/nbody_llcomp.c" 

# line 1 "LLC code (/src/parser/llc_src/forall/forall_exit.src)" 
			/*  START CODE FROM '/home/ray/Proyecto/yaCF-MPI-doc/src/llCoMP_Antonio/llCoMP/src/parser/llc_src/forall/forall_exit.src'  */

		if ((llc_case == LLC_GROUPS) || (llc_case == LLC_BLOCK)) {
		  free(llc_F);
		  llc_F = NULL;

			if (llc_nc_result_list.num_var > 0) {
			  llc_nc_result_destroy_var_list ();
			}
		}



			/*  END CODE FROM '/home/ray/Proyecto/yaCF-MPI-doc/src/llCoMP_Antonio/llCoMP/src/parser/llc_src/forall/forall_exit.src'  */


# line 136 "/home/ray/Proyecto/yaCF-MPI-src/examples/MPI_examples/NBODY/NBODY-LLCOMP/nbody_llcomp.c" 

# line 136 "/home/ray/Proyecto/yaCF-MPI-src/examples/MPI_examples/NBODY/NBODY-LLCOMP/nbody_llcomp.c" 

  }
  gettimeofday (&end_time, 0);
  time = end_time.tv_sec - start_time.tv_sec;
  time = time + (end_time.tv_usec - start_time.tv_usec) / 1e6;
  check_simulation (bodies, num_bodies);
  printf ("%d:%f\n", LLC_NAME, time);
  return (0);
}





/*
 * vim:ts=2:sw=2:
 */

