#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <omp.h>

#define DEFAULT_N 1000000000

int
main (int argc, char **argv)
{
  double local, pi_omp = 0.0, w;
  struct timeval start_time, end_time;
  double exe_time_omp;
  long i, N;
  int miid, nprocs;

  N = (argc > 1) ? atol (argv[1]) : DEFAULT_N;
  w = 1.0 / N;

  /* OMP EXECUTION */

  gettimeofday (&start_time, 0);

/* #pragma omp target device(mpi) */
#pragma omp parallel private(i, local)
  {
#pragma omp single
    pi_omp = 0.0;

#pragma omp for reduction (+: pi_omp)
    for (i = 0; i < N; i++)
      {
	local = (i + 0.5) * w;
	pi_omp = pi_omp + 4.0 / (1.0 + local * local);
      }
  }

  gettimeofday (&end_time, 0);
  exe_time_omp = end_time.tv_sec - start_time.tv_sec;
  exe_time_omp = exe_time_omp + (end_time.tv_usec - start_time.tv_usec) / 1e6;
  pi_omp *= w;

  /* END OMP EXECUTION */

    #pragma omp parallel private(miid, nprocs)
    {
    nprocs = omp_get_num_threads();
    miid = omp_get_thread_num();
    printf ("%d:%f:%f\n", miid, exe_time_omp, pi_omp);
    }

  return 0;
}
