#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <omp.h>
#include <stdlib.h>

#include <stdio.h>
#ifdef _OPENACC
#include <openacc.h>
#endif

/*
 * Default problem limits.
 */
#ifndef TOL
#define TOL 1e-14
#define FDTOL 1e-12 /* See fdtd2d for justification. */
#endif
#ifndef T_MAX
#define T_MAX 50
#endif



#define uint8_t int


int main() {

  int i = 0;
  int j = 0;
  unsigned int datasize = 100;
  int n = (int)(sqrt((int)datasize/sizeof(double)));
//   double t_end = 0;
//   double t_start = 0;
  int flag = 0;

  double *A = NULL;
  double *x = NULL;
  double *y = NULL;
  double *ya = NULL;
  double *tmp = NULL;
  double tmpscalar = 0;

  A = (double *)malloc(sizeof(double)*n*n);
  x = (double *)malloc(sizeof(double)*n);
  y = (double *)malloc(sizeof(double)*n);
  ya = (double *)malloc(sizeof(double)*n);
  tmp = (double *)malloc(sizeof(double)*n);
  
  if(A==NULL||x==NULL||y==NULL||ya==NULL||tmp==NULL){
    /* Something went wrong in the memory allocation here, fail gracefully */
    return(-10000);
  }

  for (i = 0; i < n; i++){
    x[i] = rand() / (1.0 + RAND_MAX);
    y[i] = 0;
    for (j = 0; j < n; j++){
      A[i*n+j] = rand() / (1.0 + RAND_MAX);
    }
  }

  for (i = 0; i < n; i++){
    tmpscalar = 0;
    for (j = 0; j < n; j++){
      tmpscalar += A[i*n+j] * x[j];
    }
    tmp[i] = tmpscalar;
  }


  for (j = 0; j < n; j++){
    tmpscalar = 0;
    for (i = 0; i < n; i++){
      tmpscalar += A[i*n+j] * tmp[i];
    }
    y[j] = tmpscalar;
    ya[j] = -1.0;
  }

  tmpscalar = 0;
  
  #pragma acc data copyin(A[0:n*n], x[0:n]), create(tmp[0:n]), copy(ya[0:n])
  {

    #pragma acc parallel loop private(tmpscalar) reduction(+:tmpscalar)
    for (i = 0; i < n; i++){
      tmpscalar = 0;
      #pragma acc loop reduction(+:tmpscalar)
      for (j = 0; j < n; j++){
          tmpscalar += A[i*n+j] * x[j];
      }
      tmp[i] = tmpscalar;
    }
    
    #pragma acc parallel loop private(tmpscalar) reduction(+:tmpscalar)
    for (j = 0; j < n; j++){
      tmpscalar = 0;
      #pragma acc loop reduction(+:tmpscalar)
      for (i = 0; i < n; i++){
        tmpscalar += A[i*n+j] * tmp[i];
      }
      ya[j] = tmpscalar;
    }

  } /* end data */

  for (j = 0; j < n; j++){
    if (  fabs(y[j] - ya[j])/y[j] > TOL ){
      printf("in %d y = %f != ya = %f\n", j, y[j], ya[j]);
      flag = 1;
    }
  }
  
  /* Free malloc'd memory to prevent leaks */
  free(A);
  free(x);
  free(y);
  free(ya);
  free(tmp);

  if (flag != 0) return 1;
  
  printf("OK\n");  
  return 0;

}